@@ -2,7 +2,7 @@
  *
  * Copyright 1999,2000,2001 BrightStation PLC
  * Copyright 2001,2002 Ananova Ltd
- * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010 Olly Betts
+ * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010,2011,2013,2014,2015,2016 Olly Betts
  * Copyright 2007,2009 Lemur Consulting Ltd
  * Copyright 2011, Action Without Borders
  *
@@ -27,21 +27,22 @@
 
 #include "xapian/document.h"
 #include "xapian/error.h"
-#include "xapian/errorhandler.h"
 #include "xapian/expanddecider.h"
+#include "xapian/matchspy.h"
 #include "xapian/termiterator.h"
 #include "xapian/weight.h"
 
 #include "vectortermlist.h"
 
-#include "database.h"
+#include "backends/database.h"
 #include "debuglog.h"
-#include "esetinternal.h"
-#include "expandweight.h"
-#include "multimatch.h"
-#include "omenquireinternal.h"
+#include "expand/esetinternal.h"
+#include "expand/expandweight.h"
+#include "matcher/multimatch.h"
+#include "omassert.h"
+#include "api/omenquireinternal.h"
 #include "str.h"
-#include "weightinternal.h"
+#include "weight/weightinternal.h"
 
 #include <algorithm>
 #include "autoptr.h"
@@ -52,6 +53,8 @@
 using namespace std;
 
 using Xapian::Internal::ExpandWeight;
+using Xapian::Internal::Bo1EWeight;
+using Xapian::Internal::TradEWeight;
 
 namespace Xapian {
 
@@ -155,10 +158,6 @@ MSet::MSet() : internal(new MSet::Internal)
 {
 }
 
-MSet::MSet(MSet::Internal * internal_) : internal(internal_)
-{
-}
-
 MSet::~MSet()
 {
 }
@@ -167,63 +166,38 @@ MSet::MSet(const MSet & other) : internal(other.internal)
 {
 }
 
-void
+MSet &
 MSet::operator=(const MSet &other)
 {
     internal = other.internal;
+    return *this;
 }
 
 void
-MSet::fetch(const MSetIterator & beginiter, const MSetIterator & enditer) const
+MSet::fetch_(Xapian::doccount first, Xapian::doccount last) const
 {
-    LOGCALL_VOID(API, "Xapian::MSet::fetch", beginiter | enditer);
+    LOGCALL_VOID(API, "Xapian::MSet::fetch_", first | last);
     Assert(internal.get() != 0);
-    if (beginiter != enditer)
-	internal->fetch_items(beginiter.index, enditer.index - 1);
+    internal->fetch_items(first, last);
 }
 
-void
-MSet::fetch(const MSetIterator & beginiter) const
+int
+MSet::convert_to_percent(double wt) const
 {
-    LOGCALL_VOID(API, "Xapian::MSet::fetch", beginiter);
-    Assert(internal.get() != 0);
-    internal->fetch_items(beginiter.index, beginiter.index);
-}
-
-void
-MSet::fetch() const
-{
-    LOGCALL_VOID(API, "Xapian::MSet::fetch", NO_ARGS);
-    Assert(internal.get() != 0);
-    if (!internal->items.empty())
-	internal->fetch_items(0, internal->items.size() - 1);
-}
-
-percent
-MSet::convert_to_percent(Xapian::weight wt) const
-{
-    LOGCALL(API, Xapian::percent, "Xapian::MSet::convert_to_percent", wt);
+    LOGCALL(API, int, "Xapian::MSet::convert_to_percent", wt);
     Assert(internal.get() != 0);
     RETURN(internal->convert_to_percent_internal(wt));
 }
 
-percent
-MSet::convert_to_percent(const MSetIterator & it) const
-{
-    LOGCALL(API, Xapian::percent, "Xapian::MSet::convert_to_percent", it);
-    Assert(internal.get() != 0);
-    RETURN(internal->convert_to_percent_internal(it.get_weight()));
-}
-
 Xapian::doccount
 MSet::get_termfreq(const string &tname) const
 {
     LOGCALL(API, Xapian::doccount, "Xapian::MSet::get_termfreq", tname);
-    map<string, Internal::TermFreqAndWeight>::const_iterator i;
     Assert(internal.get() != 0);
-    i = internal->termfreqandwts.find(tname);
-    if (i != internal->termfreqandwts.end()) {
-	RETURN(i->second.termfreq);
+    if (usual(internal->stats)) {
+	Xapian::doccount termfreq;
+	if (internal->stats->get_stats(tname, termfreq))
+	    RETURN(termfreq);
     }
     if (internal->enquire.get() == 0) {
 	throw InvalidOperationError("Can't get termfreq from an MSet which is not derived from a query.");
@@ -231,18 +205,21 @@ MSet::get_termfreq(const string &tname) const
     RETURN(internal->enquire->get_termfreq(tname));
 }
 
-Xapian::weight
+double
 MSet::get_termweight(const string &tname) const
 {
-    LOGCALL(API, Xapian::weight, "Xapian::MSet::get_termweight", tname);
-    map<string, Internal::TermFreqAndWeight>::const_iterator i;
+    LOGCALL(API, double, "Xapian::MSet::get_termweight", tname);
     Assert(internal.get() != 0);
-    i = internal->termfreqandwts.find(tname);
-    if (i == internal->termfreqandwts.end()) {
-	throw InvalidArgumentError("Term weight of `" + tname +
-				     "' not available.");
+    if (!internal->stats) {
+	throw InvalidOperationError("Can't get termweight from an MSet which is not derived from a query.");
     }
-    RETURN(i->second.termweight);
+    double termweight;
+    if (!internal->stats->get_termweight(tname, termweight)) {
+	string msg = tname;
+	msg += ": termweight not available";
+	throw InvalidArgumentError(msg);
+    }
+    RETURN(termweight);
 }
 
 Xapian::doccount
@@ -294,20 +271,34 @@ MSet::get_uncollapsed_matches_upper_bound() const
     return internal->uncollapsed_upper_bound;
 }
 
-Xapian::weight
+double
 MSet::get_max_possible() const
 {
     Assert(internal.get() != 0);
     return internal->max_possible;
 }
 
-Xapian::weight
+double
 MSet::get_max_attained() const
 {
     Assert(internal.get() != 0);
     return internal->max_attained;
 }
 
+string
+MSet::snippet(const string & text,
+	      size_t length,
+	      const Xapian::Stem & stemmer,
+	      unsigned flags,
+	      const string & hi_start,
+	      const string & hi_end,
+	      const string & omit) const
+{
+    Assert(internal.get() != 0);
+    return internal->snippet(text, length, stemmer, flags,
+			     hi_start, hi_end, omit);
+}
+
 Xapian::doccount
 MSet::size() const
 {
@@ -315,48 +306,6 @@ MSet::size() const
     return internal->items.size();
 }
 
-bool
-MSet::empty() const
-{
-    Assert(internal.get() != 0);
-    return internal->items.empty();
-}
-
-void
-MSet::swap(MSet & other)
-{
-    std::swap(internal, other.internal);
-}
-
-MSetIterator
-MSet::begin() const
-{
-    return MSetIterator(0, *this);
-}
-
-MSetIterator
-MSet::end() const
-{
-    Assert(internal.get() != 0);
-    return MSetIterator(internal->items.size(), *this);
-}
-
-MSetIterator
-MSet::operator[](Xapian::doccount i) const
-{
-    // Don't test 0 <= i - that gives a compiler warning if i is unsigned
-    Assert(0 < (i + 1) && i < size());
-    return MSetIterator(i, *this);
-}
-
-MSetIterator
-MSet::back() const
-{
-    Assert(!empty());
-    Assert(internal.get() != 0);
-    return MSetIterator(internal->items.size() - 1, *this);
-}
-
 string
 MSet::get_description() const
 {
@@ -364,15 +313,15 @@ MSet::get_description() const
     return "Xapian::MSet(" + internal->get_description() + ")";
 }
 
-percent
-MSet::Internal::convert_to_percent_internal(Xapian::weight wt) const
+int
+MSet::Internal::convert_to_percent_internal(double wt) const
 {
-    LOGCALL(MATCH, Xapian::percent, "Xapian::MSet::Internal::convert_to_percent_internal", wt);
+    LOGCALL(MATCH, int, "Xapian::MSet::Internal::convert_to_percent_internal", wt);
     if (percent_factor == 0) RETURN(100);
 
     // Excess precision on x86 can result in a difference here.
     double v = wt * percent_factor + 100.0 * DBL_EPSILON;
-    Xapian::percent pcent = static_cast<Xapian::percent>(v);
+    int pcent = static_cast<int>(v);
     LOGLINE(MATCH, "wt = " << wt << ", max_possible = " << max_possible <<
 		   " =>  pcent = " << pcent);
     if (pcent > 100) pcent = 100;
@@ -406,9 +355,7 @@ MSet::Internal::get_doc_by_index(Xapian::doccount index) const
 	}
     }
 
-    // Don't cache unless fetch() was called by the API user.
-    enquire->request_doc(items[index - firstitem]);
-    RETURN(enquire->read_doc(items[index - firstitem]));
+    RETURN(enquire->get_document(items[index - firstitem]));
 }
 
 void
@@ -418,6 +365,9 @@ MSet::Internal::fetch_items(Xapian::doccount first, Xapian::doccount last) const
     if (enquire.get() == 0) {
 	throw InvalidOperationError("Can't fetch documents from an MSet which is not derived from a query.");
     }
+    if (items.empty()) return;
+    if (last > items.size() - 1)
+	last = items.size() - 1;
     for (Xapian::doccount i = first; i <= last; ++i) {
 	map<Xapian::doccount, Document>::const_iterator doc;
 	doc = indexeddocs.find(i);
@@ -469,156 +419,72 @@ MSet::Internal::read_docs() const
     requested_docs.clear();
 }
 
-// Methods for Xapian::ESet
-
-ESet::ESet() : internal(new Internal) { }
-
-ESet::~ESet()
-{
-}
-
-ESet::ESet(const ESet & other) : internal(other.internal)
-{
-}
-
-void
-ESet::operator=(const ESet &other)
-{
-    internal = other.internal;
-}
-
-Xapian::termcount
-ESet::get_ebound() const
-{
-    return internal->ebound;
-}
-
-Xapian::termcount
-ESet::size() const
-{
-    return internal->items.size();
-}
-
-bool
-ESet::empty() const
-{
-    return internal->items.empty();
-}
-
-void
-ESet::swap(ESet & other)
-{
-    std::swap(internal, other.internal);
-}
-
-ESetIterator
-ESet::begin() const
-{
-    return ESetIterator(0, *this);
-}
-
-ESetIterator
-ESet::end() const
-{
-    Assert(internal.get() != 0);
-    return ESetIterator(internal->items.size(), *this);
-}
-
-ESetIterator
-ESet::operator[](Xapian::termcount i) const
-{
-    // Don't test 0 <= i - that gives a compiler warning if i is unsigned
-    Assert(0 < (i + 1) && i < size());
-    return ESetIterator(i, *this);
-}
-
-ESetIterator
-ESet::back() const
-{
-    Assert(!empty());
-    Assert(internal.get() != 0);
-    return ESetIterator(internal->items.size() - 1, *this);
-}
-
-string
-ESet::get_description() const
-{
-    Assert(internal.get() != 0);
-    return "Xapian::ESet(" + internal->get_description() + ")";
-}
-
-// Xapian::ESetIterator
-
-const string &
-ESetIterator::operator *() const
-{
-    return eset.internal->items[index].term;
-}
-
-Xapian::weight
-ESetIterator::get_weight() const
-{
-    return eset.internal->items[index].wt;
-}
-
-string
-ESetIterator::get_description() const
-{
-    return "Xapian::ESetIterator(" + str(index) + ")";
-}
-
 // MSetIterator
 
 Xapian::docid
 MSetIterator::operator *() const
 {
+    Assert(mset.internal.get());
+    Xapian::doccount size = mset.internal->items.size();
+    Xapian::doccount index = size - off_from_end;
+    AssertRel(index,<,size);
     return mset.internal->items[index].did;
 }
 
 Document
 MSetIterator::get_document() const
 {
+    Assert(mset.internal.get());
+    Xapian::doccount size = mset.internal->items.size();
+    Xapian::doccount index = size - off_from_end;
+    AssertRel(index,<,size);
     return mset.internal->get_doc_by_index(index);
 }
 
-Xapian::weight
+double
 MSetIterator::get_weight() const
 {
+    Assert(mset.internal.get());
+    Xapian::doccount size = mset.internal->items.size();
+    Xapian::doccount index = size - off_from_end;
+    AssertRel(index,<,size);
     return mset.internal->items[index].wt;
 }
 
 std::string
 MSetIterator::get_collapse_key() const
 {
+    Assert(mset.internal.get());
+    Xapian::doccount size = mset.internal->items.size();
+    Xapian::doccount index = size - off_from_end;
+    AssertRel(index,<,size);
     return mset.internal->items[index].collapse_key;
 }
 
 Xapian::doccount
 MSetIterator::get_collapse_count() const
 {
+    Assert(mset.internal.get());
+    Xapian::doccount size = mset.internal->items.size();
+    Xapian::doccount index = size - off_from_end;
+    AssertRel(index,<,size);
     return mset.internal->items[index].collapse_count;
 }
 
-Xapian::percent
-MSetIterator::get_percent() const
-{
-    LOGCALL(API, Xapian::percent, "MSetIterator::get_percent", NO_ARGS);
-    RETURN(mset.internal->convert_to_percent_internal(get_weight()));
-}
-
 string
 MSetIterator::get_description() const
 {
-    return "Xapian::MSetIterator(" + str(index) + ")";
+    return "Xapian::MSetIterator(" + str(mset.size() - off_from_end) + ")";
 }
 
 // Methods for Xapian::Enquire::Internal
 
-Enquire::Internal::Internal(const Database &db_, ErrorHandler * errorhandler_)
+Enquire::Internal::Internal(const Database &db_)
   : db(db_), query(), collapse_key(Xapian::BAD_VALUENO), collapse_max(0),
     order(Enquire::ASCENDING), percent_cutoff(0), weight_cutoff(0),
     sort_key(Xapian::BAD_VALUENO), sort_by(REL), sort_value_forward(true),
-    sorter(0), errorhandler(errorhandler_), weight(0)
+    sorter(), time_limit(0.0), weight(0),
+    eweightname("trad"), expand_k(1.0)
 {
     if (db.internal.empty()) {
 	throw InvalidArgumentError("Can't make an Enquire object from an uninitialised Database object.");
@@ -647,10 +513,9 @@ Enquire::Internal::get_query() const
 MSet
 Enquire::Internal::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 			    Xapian::doccount check_at_least, const RSet *rset,
-			    const MatchDecider *mdecider,
-			    const MatchDecider *matchspy_legacy) const
+			    const MatchDecider *mdecider) const
 {
-    LOGCALL(MATCH, MSet, "Enquire::Internal::get_mset", first | maxitems | check_at_least | rset | mdecider | matchspy_legacy);
+    LOGCALL(MATCH, MSet, "Enquire::Internal::get_mset", first | maxitems | check_at_least | rset | mdecider);
 
     if (percent_cutoff && (sort_by == VAL || sort_by == VAL_REL)) {
 	throw Xapian::UnimplementedError("Use of a percentage cutoff while sorting primary by value isn't currently supported");
@@ -669,18 +534,18 @@ Enquire::Internal::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 	check_at_least = max(check_at_least, maxitems);
     }
 
-    Xapian::Weight::Internal stats;
-    ::MultiMatch match(db, query.internal.get(), qlen, rset,
+    AutoPtr<Xapian::Weight::Internal> stats(new Xapian::Weight::Internal);
+    ::MultiMatch match(db, query, qlen, rset,
 		       collapse_max, collapse_key,
 		       percent_cutoff, weight_cutoff,
 		       order, sort_key, sort_by, sort_value_forward,
-		       errorhandler, stats, weight, spies,
-		       (sorter != NULL),
-		       (mdecider != NULL || matchspy_legacy != NULL));
+		       time_limit, *(stats.get()), weight, spies,
+		       (sorter.get() != NULL),
+		       (mdecider != NULL));
     // Run query and put results into supplied Xapian::MSet object.
     MSet retval;
     match.get_mset(first, maxitems, check_at_least, retval,
-		   stats, mdecider, matchspy_legacy, sorter);
+		   *(stats.get()), mdecider, sorter.get());
     if (first_orig != first && retval.internal.get()) {
 	retval.internal->firstitem = first_orig;
     }
@@ -693,16 +558,23 @@ Enquire::Internal::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
     // networked case.
     retval.internal->enquire = this;
 
+    if (!retval.internal->stats) {
+	retval.internal->stats = stats.release();
+    }
+
     RETURN(retval);
 }
 
 ESet
 Enquire::Internal::get_eset(Xapian::termcount maxitems,
-                    const RSet & rset, int flags, double k,
-		    const ExpandDecider * edecider, Xapian::weight min_wt) const
+			    const RSet & rset, int flags,
+			    const ExpandDecider * edecider_,
+			    double min_wt) const
 {
-    LOGCALL(MATCH, ESet, "Enquire::Internal::get_eset", maxitems | rset | flags | k | edecider | min_wt);
+    LOGCALL(MATCH, ESet, "Enquire::Internal::get_eset", maxitems | rset | flags | edecider_ | min_wt);
 
+    using Xapian::Internal::opt_intrusive_ptr;
+    opt_intrusive_ptr<const ExpandDecider> edecider(edecider_);
     if (maxitems == 0 || rset.empty()) {
 	// Either we were asked for no results, or wouldn't produce any
 	// because no documents were marked as relevant.
@@ -711,33 +583,32 @@ Enquire::Internal::get_eset(Xapian::termcount maxitems,
 
     LOGVALUE(MATCH, rset.size());
 
-    /* The AutoPtrs will clean up any dynamically allocated
-     * expand deciders automatically.
-     */
-    AutoPtr<ExpandDecider> decider_noquery;
-    AutoPtr<ExpandDecider> decider_andnoquery;
-
     if (!query.empty() && !(flags & Enquire::INCLUDE_QUERY_TERMS)) {
-	AutoPtr<ExpandDecider> temp1(
-	    new ExpandDeciderFilterTerms(query.get_terms_begin(),
-					 query.get_terms_end()));
-        decider_noquery = temp1;
-
-	if (edecider) {
-	    AutoPtr<ExpandDecider> temp2(
-		new ExpandDeciderAnd(decider_noquery.get(), edecider));
-	    decider_andnoquery = temp2;
-	    edecider = decider_andnoquery.get();
+	opt_intrusive_ptr<const ExpandDecider> decider_noquery(
+	    (new ExpandDeciderFilterTerms(query.get_terms_begin(),
+					  query.get_terms_end()))->release());
+	if (edecider.get()) {
+	    edecider = (new ExpandDeciderAnd(decider_noquery.get(),
+					     edecider.get()))->release();
 	} else {
-	    edecider = decider_noquery.get();
+	    edecider = decider_noquery;
 	}
     }
 
     bool use_exact_termfreq(flags & Enquire::USE_EXACT_TERMFREQ);
-    ExpandWeight eweight(db, rset.size(), use_exact_termfreq, k);
-
     Xapian::ESet eset;
-    eset.internal->expand(maxitems, db, rset, edecider, eweight, min_wt);
+    eset.internal = new Xapian::ESet::Internal;
+
+    if (eweightname == "bo1") {
+	Bo1EWeight bo1eweight(db, rset.size(), use_exact_termfreq);
+	eset.internal->expand(maxitems, db, rset, edecider.get(), bo1eweight,
+			      min_wt);
+    } else {
+	TradEWeight tradeweight(db, rset.size(), use_exact_termfreq, expand_k);
+	eset.internal->expand(maxitems, db, rset, edecider.get(), tradeweight,
+			      min_wt);
+    }
+
     RETURN(eset);
 }
 
@@ -762,19 +633,17 @@ TermIterator
 Enquire::Internal::get_matching_terms(Xapian::docid did) const
 {
     if (query.empty())
-	throw Xapian::InvalidArgumentError("get_matching_terms with empty query");
-	//return TermIterator();
+	return TermIterator();
 
     // The ordered list of terms in the query.
     TermIterator qt = query.get_terms_begin();
-    TermIterator qt_end = query.get_terms_end();
 
     // copy the list of query terms into a map for faster access.
     // FIXME: a hash would be faster than a map, if this becomes
     // a problem.
     map<string, unsigned int> tmap;
     unsigned int index = 1;
-    for ( ; qt != qt_end; ++qt) {
+    for ( ; qt != query.get_terms_end(); ++qt) {
 	if (tmap.find(*qt) == tmap.end())
 	    tmap[*qt] = index++;
     }
@@ -825,42 +694,44 @@ Enquire::Internal::get_description() const
 void
 Enquire::Internal::request_doc(const Xapian::Internal::MSetItem &item) const
 {
-    try {
-	unsigned int multiplier = db.internal.size();
+    unsigned int multiplier = db.internal.size();
 
-	Xapian::docid realdid = (item.did - 1) / multiplier + 1;
-	Xapian::doccount dbnumber = (item.did - 1) % multiplier;
+    Xapian::docid realdid = (item.did - 1) / multiplier + 1;
+    Xapian::doccount dbnumber = (item.did - 1) % multiplier;
 
-	db.internal[dbnumber]->request_document(realdid);
-    } catch (Error & e) {
-	if (errorhandler) (*errorhandler)(e);
-	throw;
-    }
+    db.internal[dbnumber]->request_document(realdid);
 }
 
 Document
 Enquire::Internal::read_doc(const Xapian::Internal::MSetItem &item) const
 {
-    try {
-	unsigned int multiplier = db.internal.size();
+    unsigned int multiplier = db.internal.size();
 
-	Xapian::docid realdid = (item.did - 1) / multiplier + 1;
-	Xapian::doccount dbnumber = (item.did - 1) % multiplier;
+    Xapian::docid realdid = (item.did - 1) / multiplier + 1;
+    Xapian::doccount dbnumber = (item.did - 1) % multiplier;
 
-	Xapian::Document::Internal *doc;
-	doc = db.internal[dbnumber]->collect_document(realdid);
-	return Document(doc);
-    } catch (Error & e) {
-	if (errorhandler) (*errorhandler)(e);
-	throw;
-    }
+    Xapian::Document::Internal *doc;
+    doc = db.internal[dbnumber]->collect_document(realdid);
+    return Document(doc);
+}
+
+Document
+Enquire::Internal::get_document(const Xapian::Internal::MSetItem &item) const
+{
+    unsigned int multiplier = db.internal.size();
+
+    Xapian::docid realdid = (item.did - 1) / multiplier + 1;
+    Xapian::doccount dbnumber = (item.did - 1) % multiplier;
+
+    // We know the doc exists, so open lazily.
+    return Document(db.internal[dbnumber]->open_document(realdid, true));
 }
 
 // Methods of Xapian::Enquire
 
 Enquire::Enquire(const Enquire & other) : internal(other.internal)
 {
-    LOGCALL_VOID(API, "Xapian::Enquire::Enquire", other);
+    LOGCALL_CTOR(API, "Enquire", other);
 }
 
 void
@@ -870,15 +741,21 @@ Enquire::operator=(const Enquire & other)
     internal = other.internal;
 }
 
-Enquire::Enquire(const Database &databases, ErrorHandler * errorhandler)
-    : internal(new Internal(databases, errorhandler))
+Enquire::Enquire(const Database &databases)
+    : internal(new Internal(databases))
 {
-    LOGCALL_VOID(API, "Xapian::Enquire::Enquire", databases);
+    LOGCALL_CTOR(API, "Enquire", databases);
+}
+
+Enquire::Enquire(const Database &databases, ErrorHandler *)
+    : internal(new Internal(databases))
+{
+    LOGCALL_CTOR(API, "Enquire", databases | Literal("errorhandler"));
 }
 
 Enquire::~Enquire()
 {
-    LOGCALL_VOID(API, "Xapian::Enquire::~Enquire", NO_ARGS);
+    LOGCALL_DTOR(API, "Enquire");
 }
 
 void
@@ -903,7 +780,7 @@ Enquire::add_matchspy(MatchSpy * spy) {
 
 void
 Enquire::clear_matchspies() {
-    LOGCALL(API, const Xapian::Query &, "Xapian::Enquire::clear_matchspies", NO_ARGS);
+    LOGCALL_VOID(API, "Xapian::Enquire::clear_matchspies", NO_ARGS);
     internal->spies.clear();
 }
 
@@ -917,6 +794,19 @@ Enquire::set_weighting_scheme(const Weight &weight_)
     delete wt;
 }
 
+void
+Enquire::set_expansion_scheme(const std::string &eweightname_, double expand_k_) const
+{
+     LOGCALL_VOID(API, "Xapian::Enquire::set_expansion_scheme", eweightname_ | expand_k_);
+
+     if (eweightname_ != "bo1" && eweightname_ != "trad") {
+	 throw InvalidArgumentError("Invalid name for query expansion scheme.");
+     }
+
+     internal->eweightname = eweightname_;
+     internal->expand_k = expand_k_;
+}
+
 void
 Enquire::set_collapse_key(Xapian::valueno collapse_key, Xapian::doccount collapse_max)
 {
@@ -932,7 +822,7 @@ Enquire::set_docid_order(Enquire::docid_order order)
 }
 
 void
-Enquire::set_cutoff(Xapian::percent percent_cutoff, Xapian::weight weight_cutoff)
+Enquire::set_cutoff(int percent_cutoff, double weight_cutoff)
 {
     internal->percent_cutoff = percent_cutoff;
     internal->weight_cutoff = weight_cutoff;
@@ -1001,21 +891,10 @@ Enquire::set_sort_by_relevance_then_key(KeyMaker * sorter, bool ascending)
     internal->sort_value_forward = ascending;
 }
 
-MSet
-Enquire::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
-		  Xapian::doccount check_at_least, const RSet *rset,
-		  const MatchDecider *mdecider,
-		  const MatchDecider *matchspy) const
+void
+Enquire::set_time_limit(double time_limit)
 {
-    LOGCALL(API, Xapian::MSet, "Xapian::Enquire::get_mset", first | maxitems | check_at_least | rset | mdecider | matchspy);
-
-    try {
-	RETURN(internal->get_mset(first, maxitems, check_at_least, rset,
-				  mdecider, matchspy));
-    } catch (Error & e) {
-	if (internal->errorhandler) (*internal->errorhandler)(e);
-	throw;
-    }
+    internal->time_limit = time_limit;
 }
 
 MSet
@@ -1024,66 +903,29 @@ Enquire::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 		  const MatchDecider *mdecider) const
 {
     LOGCALL(API, Xapian::MSet, "Xapian::Enquire::get_mset", first | maxitems | check_at_least | rset | mdecider);
-
-    try {
-	RETURN(internal->get_mset(first, maxitems, check_at_least, rset,
-				  mdecider, NULL));
-    } catch (Error & e) {
-	if (internal->errorhandler) (*internal->errorhandler)(e);
-	throw;
-    }
+    RETURN(internal->get_mset(first, maxitems, check_at_least, rset, mdecider));
 }
 
 ESet
 Enquire::get_eset(Xapian::termcount maxitems, const RSet & rset, int flags,
-		  double k, const ExpandDecider * edecider) const
+		  const ExpandDecider * edecider, double min_wt) const
 {
-    LOGCALL(API, Xapian::ESet, "Xapian::Enquire::get_eset", maxitems | rset | flags | k | edecider);
-
-    try {
-	RETURN(internal->get_eset(maxitems, rset, flags, k, edecider, 0));
-    } catch (Error & e) {
-	if (internal->errorhandler) (*internal->errorhandler)(e);
-	throw;
-    }
-}
-
-ESet
-Enquire::get_eset(Xapian::termcount maxitems, const RSet & rset, int flags,
-		  double k, const ExpandDecider * edecider, Xapian::weight min_wt) const
-{
-    LOGCALL(API, Xapian::ESet, "Xapian::Enquire::get_eset", maxitems | rset | flags | k | edecider | min_wt);
-
-    try {
-	RETURN(internal->get_eset(maxitems, rset, flags, k, edecider, min_wt));
-    } catch (Error & e) {
-	if (internal->errorhandler) (*internal->errorhandler)(e);
-	throw;
-    }
+    LOGCALL(API, Xapian::ESet, "Xapian::Enquire::get_eset", maxitems | rset | flags | edecider | min_wt);
+    RETURN(internal->get_eset(maxitems, rset, flags, edecider, min_wt));
 }
 
 TermIterator
 Enquire::get_matching_terms_begin(const MSetIterator &it) const
 {
-    LOGCALL(API, Xapian::TermIterator, "Xapian::Enquire::get_matching_terms", it);
-    try {
-	RETURN(internal->get_matching_terms(it));
-    } catch (Error & e) {
-	if (internal->errorhandler) (*internal->errorhandler)(e);
-	throw;
-    }
+    LOGCALL(API, Xapian::TermIterator, "Xapian::Enquire::get_matching_terms_begin", it);
+    RETURN(internal->get_matching_terms(it));
 }
 
 TermIterator
 Enquire::get_matching_terms_begin(Xapian::docid did) const
 {
-    LOGCALL(API, Xapian::TermIterator, "Xapian::Enquire::get_matching_terms", did);
-    try {
-	RETURN(internal->get_matching_terms(did));
-    } catch (Error & e) {
-	if (internal->errorhandler) (*internal->errorhandler)(e);
-	throw;
-    }
+    LOGCALL(API, Xapian::TermIterator, "Xapian::Enquire::get_matching_terms_begin", did);
+    RETURN(internal->get_matching_terms(did));
 }
 
 string
