@@ -23,19 +23,17 @@
 #include <config.h>
 
 #include "safeerrno.h"
-#ifdef __WIN32__
-# include "msvc_posix_wrapper.h"
-#endif
 
 #include <xapian/error.h>
 
 #include "chert_btreebase.h"
 #include "errno_to_string.h"
+#include "fd.h"
 #include "io_utils.h"
 #include "omassert.h"
 #include "pack.h"
+#include "posixy_wrapper.h"
 #include "str.h"
-#include "utils.h"
 
 #include <algorithm>
 #include <climits>
@@ -170,12 +168,7 @@ ChertTable_base::read(const string & name, char ch, bool read_bitmap,
 		      string &err_msg)
 {
     string basename = name + "base" + ch;
-#ifdef __WIN32__
-    int h = msvc_posix_open(basename.c_str(), O_RDONLY | O_BINARY);
-#else
-    int h = open(basename.c_str(), O_RDONLY | O_BINARY);
-#endif
-
+    FD h(posixy_open(basename.c_str(), O_RDONLY | O_CLOEXEC));
     if (h == -1) {
 	err_msg += "Couldn't open ";
 	err_msg += basename;
@@ -184,12 +177,11 @@ ChertTable_base::read(const string & name, char ch, bool read_bitmap,
 	err_msg += "\n";
 	return false;
     }
-    fdcloser closefd(h);
 
     char buf[REASONABLE_BASE_SIZE];
 
     const char *start = buf;
-    const char *end = buf + io_read(h, buf, REASONABLE_BASE_SIZE, 0);
+    const char *end = buf + io_read(h, buf, REASONABLE_BASE_SIZE);
 
     DO_UNPACK_UINT_ERRCHECK(&start, end, revision);
     uint4 format;
@@ -216,7 +208,7 @@ ChertTable_base::read(const string & name, char ch, bool read_bitmap,
     if (have_fakeroot && !sequential) {
 	sequential = true; // FIXME : work out why we need this...
 	/*
-	err_msg += "Corrupt base file, `" + basename + "':\n"
+	err_msg += "Corrupt base file, '" + basename + "':\n"
 		"sequential must be set whenever have_fakeroot is set.\n" +
 		"sequential=" + (sequential?"true":"false") +
 		", have_fakeroot=" + (have_fakeroot?"true":"false") + "\n";
@@ -310,11 +302,7 @@ ChertTable_base::write_to_file(const string &filename,
     }
     pack_uint(buf, revision);  // REVISION3
 
-#ifdef __WIN32__
-    int h = msvc_posix_open(filename.c_str(), O_WRONLY | O_CREAT | O_TRUNC | O_BINARY);
-#else
-    int h = open(filename.c_str(), O_WRONLY | O_CREAT | O_TRUNC | O_BINARY, 0666);
-#endif
+    FD h(posixy_open(filename.c_str(), O_WRONLY | O_CREAT | O_TRUNC | O_CLOEXEC, 0666));
     if (h < 0) {
 	string message("Couldn't open base ");
 	message += filename;
@@ -322,7 +310,6 @@ ChertTable_base::write_to_file(const string &filename,
 	errno_to_string(errno, message);
 	throw Xapian::DatabaseOpeningError(message);
     }
-    fdcloser closefd(h);
 
     if (changes_fd >= 0) {
 	string changes_buf;
@@ -373,6 +360,24 @@ ChertTable_base::free_block(uint4 n)
 	    bit_map_low = i;
 }
 
+/* mark_block(B, n) causes block n to be marked allocated in the bit map.
+   B->bit_map_low is the lowest byte in the bit map known to have a free bit
+   set. Searching starts from there when looking for a free block.
+*/
+
+void
+ChertTable_base::mark_block(uint4 n)
+{
+    uint4 i = n / CHAR_BIT;
+    int bit = 0x1 << n % CHAR_BIT;
+    while (i >= bit_map_size)
+	extend_bit_map();
+    bit_map[i] |= bit;
+
+    if (bit_map_low == i && bit_map[i] == 0xff)
+	++bit_map_low;
+}
+
 /* extend(B) increases the size of the two bit maps in an obvious way.
    The bitmap file grows and shrinks as the DB file grows and shrinks in
    internal usage. But the DB file itself does not reduce in size, no matter
@@ -472,23 +477,20 @@ ChertTable_base::block_free_now(uint4 n)
 void
 ChertTable_base::calculate_last_block()
 {
-    if (bit_map_size == 0) {
-	last_block = 0;
-	return;
-    }
     int i = bit_map_size - 1;
-    while (bit_map[i] == 0 && i > 0) {
+    while (i >= 0 && bit_map[i] == 0) {
 	i--;
     }
     bit_map_size = i + 1;
 
-    int x = bit_map[i];
-
     /* Check for when there are no blocks */
-    if (x == 0) {
+    if (bit_map_size == 0) {
 	last_block = 0;
 	return;
     }
+
+    int x = bit_map[i];
+
     uint4 n = (i + 1) * CHAR_BIT - 1;
     int d = 0x1 << (CHAR_BIT - 1);
     while ((x & d) == 0) { d >>= 1; n--; }
@@ -496,24 +498,13 @@ ChertTable_base::calculate_last_block()
     last_block = n;
 }
 
-bool
-ChertTable_base::is_empty() const
-{
-    for (uint4 i = 0; i < bit_map_size; i++) {
-	if (bit_map[i] != 0) {
-	    return false;
-	}
-    }
-    return true;
-}
-
 void
 ChertTable_base::clear_bit_map()
 {
     memset(bit_map, 0, bit_map_size);
 }
 
-// We've commited, so "bitmap at start" needs to be reset to the current bitmap.
+// We've committed, so "bitmap at start" needs to be reset to the current bitmap.
 void
 ChertTable_base::commit()
 {
