@@ -3,7 +3,7 @@
  * Copyright 1999,2000,2001 BrightStation PLC
  * Copyright 2001 Hein Ragas
  * Copyright 2002 Ananova Ltd
- * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010,2011,2015 Olly Betts
+ * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010,2011,2012,2013,2014,2015 Olly Betts
  * Copyright 2006,2008 Lemur Consulting Ltd
  * Copyright 2009,2010 Richard Boulton
  * Copyright 2009 Kan-Ru Chen
@@ -29,10 +29,11 @@
 
 #include "chert_database.h"
 
-#include <xapian/error.h>
-#include <xapian/valueiterator.h>
+#include "xapian/constants.h"
+#include "xapian/error.h"
+#include "xapian/valueiterator.h"
 
-#include "contiguousalldocspostlist.h"
+#include "backends/contiguousalldocspostlist.h"
 #include "chert_alldocsmodifiedpostlist.h"
 #include "chert_alldocspostlist.h"
 #include "chert_alltermslist.h"
@@ -49,22 +50,18 @@
 #include "chert_valuelist.h"
 #include "chert_values.h"
 #include "debuglog.h"
+#include "fd.h"
 #include "io_utils.h"
-#include "net/length.h"
 #include "pack.h"
-#include "remoteconnection.h"
+#include "posixy_wrapper.h"
+#include "net/remoteconnection.h"
 #include "replicate_utils.h"
-#include "replication.h"
+#include "api/replication.h"
 #include "replicationprotocol.h"
-#include "serialise.h"
+#include "net/length.h"
 #include "str.h"
 #include "stringutils.h"
-#include "utils.h"
-#include "valuestats.h"
-
-#ifdef __WIN32__
-# include "msvc_posix_wrapper.h"
-#endif
+#include "backends/valuestats.h"
 
 #include "safeerrno.h"
 #include "safesysstat.h"
@@ -77,12 +74,13 @@
 
 using namespace std;
 using namespace Xapian;
+using Xapian::Internal::intrusive_ptr;
 
 // The maximum safe term length is determined by the postlist.  There we
 // store the term using pack_string_preserving_sort() which takes the
 // length of the string plus an extra byte (assuming the string doesn't
 // contain any zero bytes), followed by the docid with encoded with
-// pack_uint_preserving_sort() which takes up to 5 bytes.
+// C_pack_uint_preserving_sort() which takes up to 5 bytes.
 //
 // The Btree manager's key length limit is 252 bytes so the maximum safe term
 // length is 252 - 1 - 5 = 246 bytes.  We use 245 rather than 246 for
@@ -104,10 +102,10 @@ const int MAX_OPEN_RETRIES = 100;
  * determining the current and next revision numbers, and stores handles
  * to the tables.
  */
-ChertDatabase::ChertDatabase(const string &chert_dir, int action,
+ChertDatabase::ChertDatabase(const string &chert_dir, int flags,
 			     unsigned int block_size)
 	: db_dir(chert_dir),
-	  readonly(action == XAPIAN_DB_READONLY),
+	  readonly(flags == Xapian::DB_READONLY_),
 	  version_file(db_dir),
 	  postlist_table(db_dir, readonly),
 	  position_table(db_dir, readonly),
@@ -119,41 +117,42 @@ ChertDatabase::ChertDatabase(const string &chert_dir, int action,
 	  lock(db_dir),
 	  max_changesets(0)
 {
-    LOGCALL_CTOR(DB, "ChertDatabase", chert_dir | action | block_size);
+    LOGCALL_CTOR(DB, "ChertDatabase", chert_dir | flags | block_size);
 
-    if (action == XAPIAN_DB_READONLY) {
+    if (readonly) {
 	open_tables_consistent();
 	return;
     }
 
+    int action = flags & Xapian::DB_ACTION_MASK_;
     if (action != Xapian::DB_OPEN && !database_exists()) {
 
 	// Create the directory for the database, if it doesn't exist
 	// already.
 	bool fail = false;
 	struct stat statbuf;
-	if (stat(db_dir, &statbuf) == 0) {
+	if (stat(db_dir.c_str(), &statbuf) == 0) {
 	    if (!S_ISDIR(statbuf.st_mode)) fail = true;
-	} else if (errno != ENOENT || mkdir(db_dir, 0755) == -1) {
+	} else if (errno != ENOENT || mkdir(db_dir.c_str(), 0755) == -1) {
 	    fail = true;
 	}
 	if (fail) {
-	    throw Xapian::DatabaseCreateError("Cannot create directory `" +
+	    throw Xapian::DatabaseCreateError("Cannot create directory '" +
 					      db_dir + "'", errno);
 	}
-	get_database_write_lock(true);
+	get_database_write_lock(flags, true);
 
 	create_and_open_tables(block_size);
 	return;
     }
 
     if (action == Xapian::DB_CREATE) {
-	throw Xapian::DatabaseCreateError("Can't create new database at `" +
+	throw Xapian::DatabaseCreateError("Can't create new database at '" +
 					  db_dir + "': a database already exists and I was told "
 					  "not to overwrite it");
     }
 
-    get_database_write_lock(false);
+    get_database_write_lock(flags, false);
     // if we're overwriting, pretend the db doesn't exist
     if (action == Xapian::DB_CREATE_OR_OVERWRITE) {
 	create_and_open_tables(block_size);
@@ -176,7 +175,7 @@ ChertDatabase::ChertDatabase(const string &chert_dir, int action,
 
 ChertDatabase::~ChertDatabase()
 {
-    LOGCALL_DTOR(DB, "~ChertDatabase");
+    LOGCALL_DTOR(DB, "ChertDatabase");
 }
 
 bool
@@ -213,10 +212,10 @@ ChertDatabase::create_and_open_tables(unsigned int block_size)
     stats.zero();
 }
 
-void
+bool
 ChertDatabase::open_tables_consistent()
 {
-    LOGCALL_VOID(DB, "ChertDatabase::open_tables_consistent", NO_ARGS);
+    LOGCALL(DB, bool, "ChertDatabase::open_tables_consistent", NO_ARGS);
     // Open record_table first, since it's the last to be written to,
     // and hence if a revision is available in it, it should be available
     // in all the other tables (unless they've moved on already).
@@ -236,7 +235,7 @@ ChertDatabase::open_tables_consistent()
     if (cur_rev && cur_rev == revision) {
 	// We're reopening a database and the revision hasn't changed so we
 	// don't need to do anything.
-	return;
+	RETURN(false);
     }
 
     // Set the block_size for optional tables as they may not currently exist.
@@ -287,6 +286,7 @@ ChertDatabase::open_tables_consistent()
     }
 
     stats.read(postlist_table);
+    return true;
 }
 
 void
@@ -339,14 +339,7 @@ ChertDatabase::get_changeset_revisions(const string & path,
 				       chert_revision_number_t * startrev,
 				       chert_revision_number_t * endrev) const
 {
-    int changes_fd = -1;
-#ifdef __WIN32__
-    changes_fd = msvc_posix_open(path.c_str(), O_RDONLY | O_BINARY);
-#else
-    changes_fd = open(path.c_str(), O_RDONLY | O_BINARY);
-#endif
-    fdcloser closer(changes_fd);
-
+    FD changes_fd(posixy_open(path.c_str(), O_RDONLY | O_CLOEXEC));
     if (changes_fd < 0) {
 	string message = string("Couldn't open changeset ")
 		+ path + " to read";
@@ -355,8 +348,7 @@ ChertDatabase::get_changeset_revisions(const string & path,
 
     char buf[REASONABLE_CHANGESET_SIZE];
     const char *start = buf;
-    const char *end = buf + io_read(changes_fd, buf,
-				    REASONABLE_CHANGESET_SIZE, 0);
+    const char *end = buf + io_read(changes_fd, buf, REASONABLE_CHANGESET_SIZE);
     if (size_t(end - start) < CONST_STRLEN(CHANGES_MAGIC_STRING))
 	throw Xapian::DatabaseError("Changeset too short at " + path);
     if (memcmp(start, CHANGES_MAGIC_STRING,
@@ -419,7 +411,7 @@ ChertDatabase::set_revision_number(chert_revision_number_t new_revision)
     }
 
     try {
-	fdcloser closefd(changes_fd);
+	FD closefd(changes_fd);
 	if (changes_fd >= 0) {
 	    string buf;
 	    chert_revision_number_t old_revision = get_revision_number();
@@ -480,10 +472,28 @@ ChertDatabase::set_revision_number(chert_revision_number_t new_revision)
 }
 
 void
+ChertDatabase::request_document(Xapian::docid did) const
+{
+    record_table.readahead_for_record(did);
+}
+
+void
+ChertDatabase::readahead_for_query(const Xapian::Query &query)
+{
+    Xapian::TermIterator t;
+    for (t = query.get_unique_terms_begin(); t != Xapian::TermIterator(); ++t) {
+	const string & term = *t;
+	if (!postlist_table.readahead_key(ChertPostListTable::make_key(term)))
+	    break;
+    }
+}
+
+bool
 ChertDatabase::reopen()
 {
-    LOGCALL_VOID(DB, "ChertDatabase::reopen", NO_ARGS);
-    if (readonly) open_tables_consistent();
+    LOGCALL(DB, bool, "ChertDatabase::reopen", NO_ARGS);
+    if (!readonly) RETURN(false);
+    RETURN(open_tables_consistent());
 }
 
 void
@@ -500,14 +510,15 @@ ChertDatabase::close()
 }
 
 void
-ChertDatabase::get_database_write_lock(bool creating)
+ChertDatabase::get_database_write_lock(int flags, bool creating)
 {
-    LOGCALL_VOID(DB, "ChertDatabase::get_database_write_lock", creating);
+    LOGCALL_VOID(DB, "ChertDatabase::get_database_write_lock", flags|creating);
     string explanation;
-    FlintLock::reason why = lock.lock(true, explanation);
+    bool retry = flags & Xapian::DB_RETRY_LOCK;
+    FlintLock::reason why = lock.lock(true, retry, explanation);
     if (why != FlintLock::SUCCESS) {
 	if (why == FlintLock::UNKNOWN && !creating && !database_exists()) {
-	    string msg("No chert database found at path `");
+	    string msg("No chert database found at path '");
 	    msg += db_dir;
 	    msg += '\'';
 	    throw Xapian::DatabaseOpeningError(msg);
@@ -544,13 +555,8 @@ ChertDatabase::send_whole_database(RemoteConnection & conn, double end_time)
     for (const char * p = filenames; *p; p += *p + 1) {
 	string leaf(p + 1, size_t(static_cast<unsigned char>(*p)));
 	filepath.replace(db_dir.size() + 1, string::npos, leaf);
-#ifdef __WIN32__
-	int fd = msvc_posix_open(filepath.c_str(), O_RDONLY | O_BINARY);
-#else
-	int fd = open(filepath.c_str(), O_RDONLY | O_BINARY);
-#endif
+	FD fd(posixy_open(filepath.c_str(), O_RDONLY | O_CLOEXEC));
 	if (fd >= 0) {
-	    fdcloser closefd(fd);
 	    conn.send_message(REPL_REPLY_DB_FILENAME, leaf, end_time);
 	    conn.send_file(REPL_REPLY_DB_FILEDATA, fd, end_time);
 	}
@@ -649,14 +655,8 @@ ChertDatabase::write_changesets_to_fd(int fd,
 
 	    // Look for the changeset for revision start_rev_num.
 	    string changes_name = db_dir + "/changes" + str(start_rev_num);
-#ifdef __WIN32__
-	    int fd_changes = msvc_posix_open(changes_name.c_str(), O_RDONLY | O_BINARY);
-#else
-	    int fd_changes = open(changes_name.c_str(), O_RDONLY | O_BINARY);
-#endif
+	    FD fd_changes(posixy_open(changes_name.c_str(), O_RDONLY | O_CLOEXEC));
 	    if (fd_changes >= 0) {
-		fdcloser closefd(fd_changes);
-
 		// Send it, and also update start_rev_num to the new value
 		// specified in the changeset.
 		chert_revision_number_t changeset_start_rev_num;
@@ -795,24 +795,29 @@ ChertDatabase::get_doclength(Xapian::docid did) const
 {
     LOGCALL(DB, Xapian::termcount, "ChertDatabase::get_doclength", did);
     Assert(did != 0);
-    Xapian::Internal::RefCntPtr<const ChertDatabase> ptrtothis(this);
+    intrusive_ptr<const ChertDatabase> ptrtothis(this);
     RETURN(postlist_table.get_doclength(did, ptrtothis));
 }
 
-Xapian::doccount
-ChertDatabase::get_termfreq(const string & term) const
+Xapian::termcount
+ChertDatabase::get_unique_terms(Xapian::docid did) const
 {
-    LOGCALL(DB, Xapian::doccount, "ChertDatabase::get_termfreq", term);
-    Assert(!term.empty());
-    RETURN(postlist_table.get_termfreq(term));
+    LOGCALL(DB, Xapian::termcount, "ChertDatabase::get_unique_terms", did);
+    Assert(did != 0);
+    intrusive_ptr<const ChertDatabase> ptrtothis(this);
+    ChertTermList termlist(ptrtothis, did);
+    // The "approximate" size should be exact in this case.
+    RETURN(termlist.get_approx_size());
 }
 
-Xapian::termcount
-ChertDatabase::get_collection_freq(const string & term) const
+void
+ChertDatabase::get_freqs(const string & term,
+			 Xapian::doccount * termfreq_ptr,
+			 Xapian::termcount * collfreq_ptr) const
 {
-    LOGCALL(DB, Xapian::termcount, "ChertDatabase::get_collection_freq", term);
+    LOGCALL_VOID(DB, "ChertDatabase::get_freqs", term | termfreq_ptr | collfreq_ptr);
     Assert(!term.empty());
-    RETURN(postlist_table.get_collection_freq(term));
+    postlist_table.get_freqs(term, termfreq_ptr, collfreq_ptr);
 }
 
 Xapian::doccount
@@ -851,7 +856,9 @@ ChertDatabase::get_doclength_upper_bound() const
 Xapian::termcount
 ChertDatabase::get_wdf_upper_bound(const string & term) const
 {
-    return min(get_collection_freq(term), stats.get_wdf_upper_bound());
+    Xapian::termcount cf;
+    get_freqs(term, NULL, &cf);
+    return min(cf, stats.get_wdf_upper_bound());
 }
 
 bool
@@ -872,7 +879,7 @@ LeafPostList *
 ChertDatabase::open_post_list(const string& term) const
 {
     LOGCALL(DB, LeafPostList *, "ChertDatabase::open_post_list", term);
-    Xapian::Internal::RefCntPtr<const ChertDatabase> ptrtothis(this);
+    intrusive_ptr<const ChertDatabase> ptrtothis(this);
 
     if (term.empty()) {
 	Xapian::doccount doccount = get_doccount();
@@ -889,7 +896,7 @@ ValueList *
 ChertDatabase::open_value_list(Xapian::valueno slot) const
 {
     LOGCALL(DB, ValueList *, "ChertDatabase::open_value_list", slot);
-    Xapian::Internal::RefCntPtr<const ChertDatabase> ptrtothis(this);
+    intrusive_ptr<const ChertDatabase> ptrtothis(this);
     RETURN(new ChertValueList(slot, ptrtothis));
 }
 
@@ -900,7 +907,7 @@ ChertDatabase::open_term_list(Xapian::docid did) const
     Assert(did != 0);
     if (!termlist_table.is_open())
 	throw_termlist_table_close_exception();
-    Xapian::Internal::RefCntPtr<const ChertDatabase> ptrtothis(this);
+    intrusive_ptr<const ChertDatabase> ptrtothis(this);
     RETURN(new ChertTermList(ptrtothis, did));
 }
 
@@ -914,7 +921,7 @@ ChertDatabase::open_document(Xapian::docid did, bool lazy) const
 	(void)get_doclength(did);
     }
 
-    Xapian::Internal::RefCntPtr<const Database::Internal> ptrtothis(this);
+    intrusive_ptr<const Database::Internal> ptrtothis(this);
     RETURN(new ChertDocument(ptrtothis, did, &value_manager, &record_table));
 }
 
@@ -937,7 +944,7 @@ TermList *
 ChertDatabase::open_allterms(const string & prefix) const
 {
     LOGCALL(DB, TermList *, "ChertDatabase::open_allterms", NO_ARGS);
-    RETURN(new ChertAllTermsList(Xapian::Internal::RefCntPtr<const ChertDatabase>(this),
+    RETURN(new ChertAllTermsList(intrusive_ptr<const ChertDatabase>(this),
 				 prefix));
 }
 
@@ -952,7 +959,7 @@ ChertDatabase::open_spelling_wordlist() const
 {
     ChertCursor * cursor = spelling_table.cursor_get();
     if (!cursor) return NULL;
-    return new ChertSpellingWordsList(Xapian::Internal::RefCntPtr<const ChertDatabase>(this),
+    return new ChertSpellingWordsList(intrusive_ptr<const ChertDatabase>(this),
 				      cursor);
 }
 
@@ -973,7 +980,7 @@ ChertDatabase::open_synonym_keylist(const string & prefix) const
 {
     ChertCursor * cursor = synonym_table.cursor_get();
     if (!cursor) return NULL;
-    return new ChertSynonymTermList(Xapian::Internal::RefCntPtr<const ChertDatabase>(this),
+    return new ChertSynonymTermList(intrusive_ptr<const ChertDatabase>(this),
 				    cursor, prefix);
 }
 
@@ -993,7 +1000,8 @@ ChertDatabase::open_metadata_keylist(const std::string &prefix) const
 {
     LOGCALL(DB, TermList *, "ChertDatabase::open_metadata_keylist", NO_ARGS);
     ChertCursor * cursor = postlist_table.cursor_get();
-    RETURN(new ChertMetadataTermList(Xapian::Internal::RefCntPtr<const ChertDatabase>(this),
+    if (!cursor) RETURN(NULL);
+    RETURN(new ChertMetadataTermList(intrusive_ptr<const ChertDatabase>(this),
 				     cursor, prefix));
 }
 
@@ -1023,6 +1031,19 @@ ChertDatabase::throw_termlist_table_close_exception() const
     throw Xapian::FeatureUnavailableError("Database has no termlist");
 }
 
+void
+ChertDatabase::get_used_docid_range(Xapian::docid & first,
+				    Xapian::docid & last) const
+{
+    last = stats.get_last_docid();
+    if (last == record_table.get_doccount()) {
+	// Contiguous range starting at 1.
+	first = 1;
+	return;
+    }
+    postlist_table.get_used_docid_range(first, last);
+}
+
 ///////////////////////////////////////////////////////////////////////////
 
 ChertWritableDatabase::ChertWritableDatabase(const string &dir, int action,
@@ -1047,7 +1068,7 @@ ChertWritableDatabase::ChertWritableDatabase(const string &dir, int action,
 
 ChertWritableDatabase::~ChertWritableDatabase()
 {
-    LOGCALL_DTOR(DB, "~ChertWritableDatabase");
+    LOGCALL_DTOR(DB, "ChertWritableDatabase");
     dtor_called();
 }
 
@@ -1175,8 +1196,7 @@ ChertWritableDatabase::add_document_(Xapian::docid did,
 	chert_doclen_t new_doclen = 0;
 	{
 	    Xapian::TermIterator term = document.termlist_begin();
-	    Xapian::TermIterator term_end = document.termlist_end();
-	    for ( ; term != term_end; ++term) {
+	    for ( ; term != document.termlist_end(); ++term) {
 		termcount wdf = term.get_wdf();
 		// Calculate the new document length
 		new_doclen += wdf;
@@ -1259,7 +1279,7 @@ ChertWritableDatabase::delete_document(Xapian::docid did)
 	value_manager.delete_document(did, value_stats);
 
 	// OK, now add entries to remove the postings in the underlying record.
-	Xapian::Internal::RefCntPtr<const ChertWritableDatabase> ptrtothis(this);
+	intrusive_ptr<const ChertWritableDatabase> ptrtothis(this);
 	ChertTermList termlist(ptrtothis, did);
 
 	stats.delete_document(termlist.get_doclength());
@@ -1315,7 +1335,7 @@ ChertWritableDatabase::replace_document(Xapian::docid did,
 
 	if (!termlist_table.is_open()) {
 	    // We can replace an *unused* docid <= last_docid too.
-	    Xapian::Internal::RefCntPtr<const ChertDatabase> ptrtothis(this);
+	    intrusive_ptr<const ChertDatabase> ptrtothis(this);
 	    if (!postlist_table.document_exists(did, ptrtothis)) {
 		(void)add_document_(did, document);
 		return;
@@ -1350,7 +1370,7 @@ ChertWritableDatabase::replace_document(Xapian::docid did,
 	if (!modifying || document.internal->terms_modified()) {
 	    bool pos_modified = !modifying ||
 				document.internal->term_positions_modified();
-	    Xapian::Internal::RefCntPtr<const ChertWritableDatabase> ptrtothis(this);
+	    intrusive_ptr<const ChertWritableDatabase> ptrtothis(this);
 	    ChertTermList termlist(ptrtothis, did);
 	    Xapian::TermIterator term = document.termlist_begin();
 	    chert_doclen_t old_doclen = termlist.get_doclength();
@@ -1498,28 +1518,22 @@ ChertWritableDatabase::get_doclength(Xapian::docid did) const
     RETURN(ChertDatabase::get_doclength(did));
 }
 
-Xapian::doccount
-ChertWritableDatabase::get_termfreq(const string & tname) const
+void
+ChertWritableDatabase::get_freqs(const string & term,
+				 Xapian::doccount * termfreq_ptr,
+				 Xapian::termcount * collfreq_ptr) const
 {
-    LOGCALL(DB, Xapian::doccount, "ChertWritableDatabase::get_termfreq", tname);
-    Xapian::doccount termfreq = ChertDatabase::get_termfreq(tname);
+    LOGCALL_VOID(DB, "ChertWritableDatabase::get_freqs", term | termfreq_ptr | collfreq_ptr);
+    Assert(!term.empty());
+    ChertDatabase::get_freqs(term, termfreq_ptr, collfreq_ptr);
     map<string, pair<termcount_diff, termcount_diff> >::const_iterator i;
-    i = freq_deltas.find(tname);
-    if (i != freq_deltas.end()) termfreq += i->second.first;
-    RETURN(termfreq);
-}
-
-Xapian::termcount
-ChertWritableDatabase::get_collection_freq(const string & tname) const
-{
-    LOGCALL(DB, Xapian::termcount, "ChertWritableDatabase::get_collection_freq", tname);
-    Xapian::termcount collfreq = ChertDatabase::get_collection_freq(tname);
-
-    map<string, pair<termcount_diff, termcount_diff> >::const_iterator i;
-    i = freq_deltas.find(tname);
-    if (i != freq_deltas.end()) collfreq += i->second.second;
-
-    RETURN(collfreq);
+    i = freq_deltas.find(term);
+    if (i != freq_deltas.end()) {
+	if (termfreq_ptr)
+	    *termfreq_ptr += i->second.first;
+	if (collfreq_ptr)
+	    *collfreq_ptr += i->second.second;
+    }
 }
 
 Xapian::doccount
@@ -1556,14 +1570,16 @@ bool
 ChertWritableDatabase::term_exists(const string & tname) const
 {
     LOGCALL(DB, bool, "ChertWritableDatabase::term_exists", tname);
-    RETURN(get_termfreq(tname) != 0);
+    Xapian::doccount tf;
+    get_freqs(tname, &tf, NULL);
+    RETURN(tf != 0);
 }
 
 LeafPostList *
 ChertWritableDatabase::open_post_list(const string& tname) const
 {
     LOGCALL(DB, LeafPostList *, "ChertWritableDatabase::open_post_list", tname);
-    Xapian::Internal::RefCntPtr<const ChertWritableDatabase> ptrtothis(this);
+    intrusive_ptr<const ChertWritableDatabase> ptrtothis(this);
 
     if (tname.empty()) {
 	Xapian::doccount doccount = get_doccount();
@@ -1672,7 +1688,7 @@ ChertWritableDatabase::clear_synonyms(const string & term) const
 void
 ChertWritableDatabase::set_metadata(const string & key, const string & value)
 {
-    LOGCALL(DB, string, "ChertWritableDatabase::set_metadata", key | value);
+    LOGCALL_VOID(DB, "ChertWritableDatabase::set_metadata", key | value);
     string btree_key("\x00\xc0", 2);
     btree_key += key;
     if (value.empty()) {
