@@ -2,7 +2,7 @@
  *
  * Copyright 1999,2000,2001 BrightStation PLC
  * Copyright 2002 Ananova Ltd
- * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010,2011,2013,2015 Olly Betts
+ * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2016 Olly Betts
  * Copyright 2008 Lemur Consulting Ltd
  * Copyright 2010 Richard Boulton
  *
@@ -29,9 +29,6 @@
 #include <xapian/error.h>
 
 #include "safeerrno.h"
-#ifdef __WIN32__
-# include "msvc_posix_wrapper.h"
-#endif
 
 #include "errno_to_string.h"
 #include "omassert.h"
@@ -52,31 +49,18 @@
 
 #include <sys/types.h>
 
-// Trying to include the correct headers with the correct defines set to
-// get pread() and pwrite() prototyped on every platform without breaking any
-// other platform is a real can of worms.  So instead we probe for what
-// prototypes (if any) are required in configure and put them into
-// PREAD_PROTOTYPE and PWRITE_PROTOTYPE.
-#if defined HAVE_PREAD && defined PREAD_PROTOTYPE
-PREAD_PROTOTYPE
-#endif
-#if defined HAVE_PWRITE && defined PWRITE_PROTOTYPE
-PWRITE_PROTOTYPE
-#endif
-
-#include <cstdio>    /* for rename */
 #include <cstring>   /* for memmove */
 #include <climits>   /* for CHAR_BIT */
 
 #include "chert_btreebase.h"
 #include "chert_cursor.h"
 
+#include "filetests.h"
 #include "io_utils.h"
 #include "debuglog.h"
 #include "pack.h"
 #include "str.h"
-#include "unaligned.h"
-#include "utils.h"
+#include "wordaccess.h"
 
 #include <algorithm>  // for std::min()
 #include <string>
@@ -112,9 +96,6 @@ static void report_cursor(int N, Btree * B, Cursor * C)
 static inline byte *zeroed_new(size_t size)
 {
     byte *temp = new byte[size];
-    // No need to check if temp is NULL, new throws std::bad_alloc if
-    // allocation fails.
-    Assert(temp);
     memset(temp, 0, size);
     return temp;
 }
@@ -190,49 +171,15 @@ ChertTable::read_block(uint4 n, byte * p) const
 {
     // Log the value of p, not the contents of the block it points to...
     LOGCALL_VOID(DB, "ChertTable::read_block", n | (void*)p);
+    if (rare(handle == -2))
+	ChertTable::throw_database_closed();
+
     /* Use the base bit_map_size not the bitmap's size, because
      * the latter is uninitialised in readonly mode.
      */
     Assert(n / CHAR_BIT < base.get_bit_map_size());
 
-#ifdef HAVE_PREAD
-    off_t offset = off_t(block_size) * n;
-    int m = block_size;
-    while (true) {
-	ssize_t bytes_read = pread(handle, reinterpret_cast<char *>(p), m,
-				   offset);
-	// normal case - read succeeded, so return.
-	if (bytes_read == m) break;
-	if (bytes_read == -1) {
-	    if (errno == EINTR) continue;
-	    if (errno == EBADF && handle == -2)
-		ChertTable::throw_database_closed();
-	    string message = "Error reading block " + str(n) + ": ";
-	    errno_to_string(errno, message);
-	    throw Xapian::DatabaseError(message);
-	} else if (bytes_read == 0) {
-	    string message = "Error reading block " + str(n) + ": got end of file";
-	    throw Xapian::DatabaseError(message);
-	} else if (bytes_read < m) {
-	    /* Read part of the block, which is not an error.  We should
-	     * continue reading the rest of the block.
-	     */
-	    m -= int(bytes_read);
-	    p += bytes_read;
-	    offset += bytes_read;
-	}
-    }
-#else
-    if (lseek(handle, off_t(block_size) * n, SEEK_SET) == -1) {
-	if (errno == EBADF && handle == -2)
-	    ChertTable::throw_database_closed();
-	string message = "Error seeking to block: ";
-	errno_to_string(errno, message);
-	throw Xapian::DatabaseError(message);
-    }
-
-    io_read(handle, reinterpret_cast<char *>(p), block_size, block_size);
-#endif
+    io_read_block(handle, reinterpret_cast<char *>(p), block_size, n);
 
     int dir_end = DIR_END(p);
     if (rare(dir_end < DIR_START || unsigned(dir_end) > block_size)) {
@@ -275,40 +222,7 @@ ChertTable::write_block(uint4 n, const byte * p) const
 	latest_revision_number = revision_number;
     }
 
-#ifdef HAVE_PWRITE
-    off_t offset = off_t(block_size) * n;
-    int m = block_size;
-    while (true) {
-	ssize_t bytes_written = pwrite(handle, p, m, offset);
-	if (bytes_written == m) {
-	    // normal case - write succeeded, so return.
-	    return;
-	} else if (bytes_written == -1) {
-	    if (errno == EINTR) continue;
-	    string message = "Error writing block: ";
-	    errno_to_string(errno, message);
-	    throw Xapian::DatabaseError(message);
-	} else if (bytes_written == 0) {
-	    string message = "Error writing block: wrote no data";
-	    throw Xapian::DatabaseError(message);
-	} else if (bytes_written < m) {
-	    /* Wrote part of the block, which is not an error.  We should
-	     * continue writing the rest of the block.
-	     */
-	    m -= bytes_written;
-	    p += bytes_written;
-	    offset += bytes_written;
-	}
-    }
-#else
-    if (lseek(handle, (off_t)block_size * n, SEEK_SET) == -1) {
-	string message = "Error seeking to block: ";
-	errno_to_string(errno, message);
-	throw Xapian::DatabaseError(message);
-    }
-
-    io_write(handle, reinterpret_cast<const char *>(p), block_size);
-#endif
+    io_write_block(handle, reinterpret_cast<const char *>(p), block_size, n);
 }
 
 
@@ -1073,11 +987,15 @@ ChertTable::add(const string &key, string tag, bool already_compressed)
     if (already_compressed) {
 	compressed = true;
     } else if (compress_strategy != DONT_COMPRESS && tag.size() > COMPRESS_MIN) {
-	CompileTimeAssert(DONT_COMPRESS != Z_DEFAULT_STRATEGY);
-	CompileTimeAssert(DONT_COMPRESS != Z_FILTERED);
-	CompileTimeAssert(DONT_COMPRESS != Z_HUFFMAN_ONLY);
+	static_assert(DONT_COMPRESS != Z_DEFAULT_STRATEGY,
+		      "DONT_COMPRESS clashes with zlib constant");
+	static_assert(DONT_COMPRESS != Z_FILTERED,
+		      "DONT_COMPRESS clashes with zlib constant");
+	static_assert(DONT_COMPRESS != Z_HUFFMAN_ONLY,
+		      "DONT_COMPRESS clashes with zlib constant");
 #ifdef Z_RLE
-	CompileTimeAssert(DONT_COMPRESS != Z_RLE);
+	static_assert(DONT_COMPRESS != Z_RLE,
+		      "DONT_COMPRESS clashes with zlib constant");
 #endif
 
 	lazy_alloc_deflate_zstream();
@@ -1217,6 +1135,48 @@ ChertTable::del(const string &key)
     RETURN(true);
 }
 
+bool
+ChertTable::readahead_key(const string &key) const
+{
+    LOGCALL(DB, bool, "ChertTable::readahead_key", key);
+    Assert(!key.empty());
+
+    // Two cases:
+    //
+    // handle = -1:  Lazy table which isn't yet open
+    //
+    // handle = -2:  Table has been closed.  Since the readahead is just a
+    // hint, we can safely ignore it for a closed table.
+    if (handle < 0)
+	RETURN(false);
+
+    // If the table only has one level, there are no branch blocks to preread.
+    if (level == 0)
+	RETURN(false);
+
+    form_key(key);
+    Key ktkey = kt.key();
+
+    // We'll only readahead the first level, since descending the B-tree would
+    // require actual reads that would likely hurt performance more than help.
+    const byte * p = C[level].p;
+    int c = find_in_block(p, ktkey, false, C[level].c);
+    uint4 n = Item(p, c).block_given_by();
+    // Don't preread if it's the block we last preread or already in the
+    // cursor.
+    if (n != last_readahead && n != C[level - 1].n) {
+	/* Use the base bit_map_size not the bitmap's size, because the latter
+	 * is uninitialised in readonly mode.
+	 */
+	Assert(n / CHAR_BIT < base.get_bit_map_size());
+
+	last_readahead = n;
+	if (!io_readahead_block(handle, block_size, n))
+	    RETURN(false);
+    }
+    RETURN(true);
+}
+
 bool
 ChertTable::get_exact_entry(const string &key, string & tag) const
 {
@@ -1303,7 +1263,7 @@ ChertTable::read_tag(Cursor * C_, string *tag, bool keep_compressed) const
 	if (err == Z_BUF_ERROR && inflate_zstream->avail_in == 0) {
 	    LOGLINE(DB, "Z_BUF_ERROR - faking checksum of " << inflate_zstream->adler);
 	    Bytef header2[4];
-	    setint4(header2, 0, inflate_zstream->adler);
+	    aligned_write4(header2, inflate_zstream->adler);
 	    inflate_zstream->next_in = header2;
 	    inflate_zstream->avail_in = 4;
 	    err = inflate(inflate_zstream, Z_SYNC_FLUSH);
@@ -1395,7 +1355,7 @@ ChertTable::basic_open(bool revision_supplied, chert_revision_number_t revision_
 		::close(handle);
 		handle = -1;
 	    }
-	    string message = "Error opening table `";
+	    string message = "Error opening table '";
 	    message += name;
 	    message += "':\n";
 	    message += err_msg;
@@ -1577,14 +1537,8 @@ ChertTable::do_open_to_write(bool revision_supplied,
     for (int j = 0; j <= level; j++) {
 	C[j].n = BLK_UNUSED;
 	C[j].p = new byte[block_size];
-	if (C[j].p == 0) {
-	    throw std::bad_alloc();
-	}
     }
     split_p = new byte[block_size];
-    if (split_p == 0) {
-	throw std::bad_alloc();
-    }
     read_root();
 
     buffer = zeroed_new(block_size);
@@ -1627,7 +1581,8 @@ ChertTable::ChertTable(const char * tablename_, const string & path_,
 	  compress_strategy(compress_strategy_),
 	  deflate_zstream(NULL),
 	  inflate_zstream(NULL),
-	  lazy(lazy_)
+	  lazy(lazy_),
+	  last_readahead(BLK_UNUSED)
 {
     LOGCALL_CTOR(DB, "ChertTable", tablename_ | path_ | readonly_ | compress_strategy_ | lazy_);
 }
@@ -1807,7 +1762,7 @@ ChertTable::~ChertTable() {
 }
 
 void ChertTable::close(bool permanent) {
-    LOGCALL_VOID(DB, "ChertTable::close", NO_ARGS);
+    LOGCALL_VOID(DB, "ChertTable::close", permanent);
 
     if (handle >= 0) {
 	// If an error occurs here, we just ignore it, since we're just
@@ -1919,29 +1874,14 @@ ChertTable::commit(chert_revision_number_t revision, int changes_fd,
 	if (changes_tail ? !io_full_sync(handle) : !io_sync(handle)) {
 	    (void)::close(handle);
 	    handle = -1;
-	    (void)unlink(tmp);
+	    (void)unlink(tmp.c_str());
 	    throw Xapian::DatabaseError("Can't commit new revision - failed to flush DB to disk");
 	}
 
-#if defined __WIN32__
-	if (msvc_posix_rename(tmp.c_str(), basefile.c_str()) < 0)
-#else
-	if (rename(tmp.c_str(), basefile.c_str()) < 0)
-#endif
-	{
-	    // With NFS, rename() failing may just mean that the server crashed
-	    // after successfully renaming, but before reporting this, and then
-	    // the retried operation fails.  So we need to check if the source
-	    // file still exists, which we do by calling unlink(), since we want
-	    // to remove the temporary file anyway.
-	    int saved_errno = errno;
-	    if (unlink(tmp) == 0 || errno != ENOENT) {
-		string msg("Couldn't update base file ");
-		msg += basefile;
-		msg += ": ";
-		errno_to_string(saved_errno, msg);
-		throw Xapian::DatabaseError(msg);
-	    }
+	if (!io_tmp_rename(tmp, basefile)) {
+	    string msg("Couldn't update base file ");
+	    msg += basefile;
+	    throw Xapian::DatabaseError(msg, errno);
 	}
 	base.commit();
 
@@ -2089,9 +2029,6 @@ ChertTable::do_open_to_read(bool revision_supplied, chert_revision_number_t revi
     for (int j = 0; j <= level; j++) {
 	C[j].n = BLK_UNUSED;
 	C[j].p = new byte[block_size];
-	if (C[j].p == 0) {
-	    throw std::bad_alloc();
-	}
     }
 
     read_root();
