@@ -27,14 +27,9 @@
 
 #include "chert_version.h"
 #include "io_utils.h"
-#include "omassert.h"
 #include "stringutils.h" // For STRINGIZE() and CONST_STRLEN().
 #include "str.h"
 
-#ifdef __WIN32__
-# include "msvc_posix_wrapper.h"
-#endif
-
 #include <cstring> // For memcmp() and memcpy().
 #include <string>
 
@@ -72,7 +67,7 @@ ChertVersion::create()
     uuid_generate(uuid);
     memcpy(buf + MAGIC_LEN + 4, (void*)uuid, 16);
 
-    int fd = ::open(filename.c_str(), O_WRONLY|O_CREAT|O_TRUNC|O_BINARY, 0666);
+    int fd = ::open(filename.c_str(), O_WRONLY|O_CREAT|O_TRUNC|O_BINARY|O_CLOEXEC, 0666);
 
     if (fd < 0) {
 	string msg("Failed to create chert version file: ");
@@ -98,7 +93,7 @@ ChertVersion::create()
 void
 ChertVersion::read_and_check()
 {
-    int fd = ::open(filename.c_str(), O_RDONLY|O_BINARY);
+    int fd = ::open(filename.c_str(), O_RDONLY|O_BINARY|O_CLOEXEC);
 
     if (fd < 0) {
 	string msg = filename;
@@ -110,7 +105,7 @@ ChertVersion::read_and_check()
     char buf[VERSIONFILE_SIZE + 1];
     size_t size;
     try {
-	size = io_read(fd, buf, VERSIONFILE_SIZE + 1, 0);
+	size = io_read(fd, buf, VERSIONFILE_SIZE + 1);
     } catch (...) {
 	(void)close(fd);
 	throw;
@@ -118,7 +113,8 @@ ChertVersion::read_and_check()
     (void)close(fd);
 
     if (size != VERSIONFILE_SIZE) {
-	CompileTimeAssert(VERSIONFILE_SIZE == VERSIONFILE_SIZE_LITERAL);
+	static_assert(VERSIONFILE_SIZE == VERSIONFILE_SIZE_LITERAL,
+		      "VERSIONFILE_SIZE_LITERAL needs updating");
 	string msg = filename;
 	msg += ": Chert version file should be "
 	       STRINGIZE(VERSIONFILE_SIZE_LITERAL)" bytes, actually ";
