@@ -1,7 +1,7 @@
 /** @file dbfactory.cc
  * @brief Database factories for non-remote databases.
  */
-/* Copyright 2002,2003,2004,2005,2006,2007,2008,2009 Olly Betts
+/* Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2011,2012,2013,2014,2015,2016 Olly Betts
  * Copyright 2008 Lemur Consulting Ltd
  *
  * This program is free software; you can redistribute it and/or
@@ -24,87 +24,74 @@
 
 #include "xapian/dbfactory.h"
 
+#include "xapian/constants.h"
 #include "xapian/database.h"
 #include "xapian/error.h"
 #include "xapian/version.h" // For XAPIAN_HAS_XXX_BACKEND.
 
 #include "debuglog.h"
+#include "filetests.h"
 #include "fileutils.h"
+#include "posixy_wrapper.h"
 #include "str.h"
-#include "utils.h"
 
 #include "safeerrno.h"
+#include <cstdlib> // For atoi().
 
-#ifdef XAPIAN_HAS_BRASS_BACKEND
-# include "brass/brass_database.h"
+#ifdef XAPIAN_HAS_GLASS_BACKEND
+# include "glass/glass_database.h"
 #endif
 #ifdef XAPIAN_HAS_CHERT_BACKEND
 # include "chert/chert_database.h"
 #endif
-#ifdef XAPIAN_HAS_FLINT_BACKEND
-# include "flint/flint_database.h"
-#endif
 #ifdef XAPIAN_HAS_INMEMORY_BACKEND
 # include "inmemory/inmemory_database.h"
 #endif
+// Even if none of the above get included, we still need a definition of
+// Database::Internal.
+#include "backends/database.h"
 
 #include <fstream>
 #include <string>
 
 using namespace std;
 
+static bool
+check_if_single_file_db(const struct stat & sb, const string & path,
+			int * fd_ptr = NULL)
+{
+#ifdef XAPIAN_HAS_GLASS_BACKEND
+    if (!S_ISREG(sb.st_mode)) return false;
+    // Look at the size as a clue - if it's 0 or not a multiple of 2048,
+    // then it's not a single-file glass database.  If it is, peek at the start
+    // of the file to determine which it is.
+    if (sb.st_size == 0 || sb.st_size % 2048 != 0) return false;
+    int fd = posixy_open(path.c_str(), O_RDONLY|O_BINARY);
+    if (fd != -1) {
+	char magic_buf[14];
+	// FIXME: Don't duplicate magic check here...
+	if (io_read(fd, magic_buf, 14) == 14 &&
+	    lseek(fd, 0, SEEK_SET) == 0 &&
+	    memcmp(magic_buf, "\x0f\x0dXapian Glass", 14) == 0) {
+	    if (fd_ptr) {
+		*fd_ptr = fd;
+	    } else {
+		::close(fd);
+	    }
+	    return true;
+	}
+	::close(fd);
+    }
+#else
+    (void)sb;
+    (void)path;
+    (void)fd_ptr;
+#endif
+    return false;
+}
+
 namespace Xapian {
 
-#ifdef XAPIAN_HAS_BRASS_BACKEND
-Database
-Brass::open(const string &dir) {
-    LOGCALL_STATIC(API, Database, "Brass::open", dir);
-    RETURN(Database(new BrassDatabase(dir)));
-}
-
-WritableDatabase
-Brass::open(const string &dir, int action, int block_size) {
-    LOGCALL_STATIC(API, WritableDatabase, "Brass::open", dir | action | block_size);
-    RETURN(WritableDatabase(new BrassWritableDatabase(dir, action, block_size)));
-}
-#endif
-
-#ifdef XAPIAN_HAS_CHERT_BACKEND
-Database
-Chert::open(const string &dir) {
-    LOGCALL_STATIC(API, Database, "Chert::open", dir);
-    return Database(new ChertDatabase(dir));
-}
-
-WritableDatabase
-Chert::open(const string &dir, int action, int block_size) {
-    LOGCALL_STATIC(API, WritableDatabase, "Chert::open", dir | action | block_size);
-    return WritableDatabase(new ChertWritableDatabase(dir, action, block_size));
-}
-#endif
-
-#ifdef XAPIAN_HAS_FLINT_BACKEND
-Database
-Flint::open(const string &dir) {
-    LOGCALL_STATIC(API, Database, "Flint::open", dir);
-    return Database(new FlintDatabase(dir));
-}
-
-WritableDatabase
-Flint::open(const string &dir, int action, int block_size) {
-    LOGCALL_STATIC(API, WritableDatabase, "Flint::open", dir | action | block_size);
-    return WritableDatabase(new FlintWritableDatabase(dir, action, block_size));
-}
-#endif
-
-#ifdef XAPIAN_HAS_INMEMORY_BACKEND
-WritableDatabase
-InMemory::open() {
-    LOGCALL_STATIC(API, WritableDatabase, "InMemory::open", NO_ARGS);
-    RETURN(WritableDatabase(new InMemoryDatabase));
-}
-#endif
-
 static void
 open_stub(Database &db, const string &file)
 {
@@ -142,23 +129,15 @@ open_stub(Database &db, const string &file)
 #ifdef XAPIAN_HAS_CHERT_BACKEND
 	if (type == "chert") {
 	    resolve_relative_path(line, file);
-	    db.add_database(Chert::open(line));
+	    db.add_database(Database(new ChertDatabase(line)));
 	    continue;
 	}
 #endif
 
-#ifdef XAPIAN_HAS_FLINT_BACKEND
-	if (type == "flint") {
+#ifdef XAPIAN_HAS_GLASS_BACKEND
+	if (type == "glass") {
 	    resolve_relative_path(line, file);
-	    db.add_database(Flint::open(line));
-	    continue;
-	}
-#endif
-
-#ifdef XAPIAN_HAS_BRASS_BACKEND
-	if (type == "brass") {
-	    resolve_relative_path(line, file);
-	    db.add_database(Brass::open(line));
+	    db.add_database(Database(new GlassDatabase(line)));
 	    continue;
 	}
 #endif
@@ -192,11 +171,15 @@ open_stub(Database &db, const string &file)
 
 #ifdef XAPIAN_HAS_INMEMORY_BACKEND
 	if (type == "inmemory" && line.empty()) {
-	    db.add_database(InMemory::open());
+	    db.add_database(Database(string(), DB_BACKEND_INMEMORY));
 	    continue;
 	}
 #endif
 
+	if (type == "flint") {
+	    throw FeatureUnavailableError("Flint backend no longer supported");
+	}
+
 	// Don't include the line itself - that might help an attacker
 	// by revealing part of a sensitive file's contents if they can
 	// arrange for it to be read as a stub database via infelicities in
@@ -215,7 +198,7 @@ open_stub(Database &db, const string &file)
 }
 
 static void
-open_stub(WritableDatabase &db, const string &file, int action)
+open_stub(WritableDatabase &db, const string &file, int flags)
 {
     // A stub database is a text file with one or more lines of this format:
     // <dbtype> <serialised db object>
@@ -234,10 +217,6 @@ open_stub(WritableDatabase &db, const string &file, int action)
     string line;
     unsigned int line_no = 0;
     while (true) {
-	if (db.internal.size() > 1) {
-	    throw DatabaseOpeningError(file + ": Can't open a stub database listing multiple databases as a WritableDatabase");
-	}
-
 	if (!getline(stub, line)) break;
 
 	++line_no;
@@ -251,30 +230,22 @@ open_stub(WritableDatabase &db, const string &file, int action)
 
 	if (type == "auto") {
 	    resolve_relative_path(line, file);
-	    db.add_database(WritableDatabase(line, action));
+	    db.add_database(WritableDatabase(line, flags));
 	    continue;
 	}
 
 #ifdef XAPIAN_HAS_CHERT_BACKEND
 	if (type == "chert") {
 	    resolve_relative_path(line, file);
-	    db.add_database(Chert::open(line, action));
+	    db.add_database(WritableDatabase(line, flags|DB_BACKEND_CHERT));
 	    continue;
 	}
 #endif
 
-#ifdef XAPIAN_HAS_FLINT_BACKEND
-	if (type == "flint") {
+#ifdef XAPIAN_HAS_GLASS_BACKEND
+	if (type == "glass") {
 	    resolve_relative_path(line, file);
-	    db.add_database(Flint::open(line, action));
-	    continue;
-	}
-#endif
-
-#ifdef XAPIAN_HAS_BRASS_BACKEND
-	if (type == "brass") {
-	    resolve_relative_path(line, file);
-	    db.add_database(Brass::open(line, action));
+	    db.add_database(WritableDatabase(line, flags|DB_BACKEND_GLASS));
 	    continue;
 	}
 #endif
@@ -294,13 +265,13 @@ open_stub(WritableDatabase &db, const string &file, int action)
 		} else {
 		    line.erase(0, 1);
 		}
-		db.add_database(Remote::open_writable(line, args));
+		db.add_database(Remote::open_writable(line, args, 0, flags));
 	    } else if (colon != string::npos) {
 		// tcp
 		// FIXME: timeouts
 		unsigned int port = atoi(line.c_str() + colon + 1);
 		line.erase(colon);
-		db.add_database(Remote::open_writable(line, port));
+		db.add_database(Remote::open_writable(line, port, 0, 10000, flags));
 	    }
 	    continue;
 	}
@@ -308,11 +279,15 @@ open_stub(WritableDatabase &db, const string &file, int action)
 
 #ifdef XAPIAN_HAS_INMEMORY_BACKEND
 	if (type == "inmemory" && line.empty()) {
-	    db.add_database(InMemory::open());
+	    db.add_database(WritableDatabase(string(), DB_BACKEND_INMEMORY));
 	    continue;
 	}
 #endif
 
+	if (type == "flint") {
+	    throw FeatureUnavailableError("Flint backend no longer supported");
+	}
+
 	// Don't include the line itself - that might help an attacker
 	// by revealing part of a sensitive file's contents if they can
 	// arrange for it to be read as a stub database via infelicities in
@@ -326,35 +301,56 @@ open_stub(WritableDatabase &db, const string &file, int action)
     }
 }
 
-Database
-Auto::open_stub(const string &file)
+Database::Database(const string &path, int flags)
 {
-    LOGCALL_STATIC(API, Database, "Auto::open_stub", file);
-    Database db;
-    open_stub(db, file);
-    RETURN(db);
-}
+    LOGCALL_CTOR(API, "Database", path|flags);
 
-WritableDatabase
-Auto::open_stub(const string &file, int action)
-{
-    LOGCALL_STATIC(API, WritableDatabase, "Auto::open_stub", file | action);
-    WritableDatabase db;
-    open_stub(db, file, action);
-    RETURN(db);
-}
-
-Database::Database(const string &path)
-{
-    LOGCALL_CTOR(API, "Database", path);
+    int type = flags & DB_BACKEND_MASK_;
+    switch (type) {
+	case DB_BACKEND_CHERT:
+#ifdef XAPIAN_HAS_CHERT_BACKEND
+	    internal.push_back(new ChertDatabase(path));
+	    return;
+#else
+	    throw FeatureUnavailableError("Chert backend disabled");
+#endif
+	case DB_BACKEND_GLASS:
+#ifdef XAPIAN_HAS_GLASS_BACKEND
+	    internal.push_back(new GlassDatabase(path));
+	    return;
+#else
+	    throw FeatureUnavailableError("Glass backend disabled");
+#endif
+	case DB_BACKEND_STUB:
+	    open_stub(*this, path);
+	    return;
+	case DB_BACKEND_INMEMORY:
+#ifdef XAPIAN_HAS_INMEMORY_BACKEND
+	    internal.push_back(new InMemoryDatabase());
+	    return;
+#else
+	    throw FeatureUnavailableError("Inmemory backend disabled");
+#endif
+    }
 
     struct stat statbuf;
-    if (stat(path, &statbuf) == -1) {
+    if (stat(path.c_str(), &statbuf) == -1) {
 	throw DatabaseOpeningError("Couldn't stat '" + path + "'", errno);
     }
 
     if (S_ISREG(statbuf.st_mode)) {
-	// The path is a file, so assume it is a stub database file.
+	// Could be a stub database file, or a single file glass database.
+	int fd;
+	if (check_if_single_file_db(statbuf, path, &fd)) {
+#ifdef XAPIAN_HAS_GLASS_BACKEND
+	    // Single file glass format.
+	    internal.push_back(new GlassDatabase(fd));
+	    return;
+#else
+	    throw FeatureUnavailableError("Glass backend disabled");
+#endif
+	}
+
 	open_stub(*this, path);
 	return;
     }
@@ -370,16 +366,9 @@ Database::Database(const string &path)
     }
 #endif
 
-#ifdef XAPIAN_HAS_FLINT_BACKEND
-    if (file_exists(path + "/iamflint")) {
-	internal.push_back(new FlintDatabase(path));
-	return;
-    }
-#endif
-
-#ifdef XAPIAN_HAS_BRASS_BACKEND
-    if (file_exists(path + "/iambrass")) {
-	internal.push_back(new BrassDatabase(path));
+#ifdef XAPIAN_HAS_GLASS_BACKEND
+    if (file_exists(path + "/iamglass")) {
+	internal.push_back(new GlassDatabase(path));
 	return;
     }
 #endif
@@ -387,124 +376,137 @@ Database::Database(const string &path)
     // Check for "stub directories".
     string stub_file = path;
     stub_file += "/XAPIANDB";
-    if (rare(!file_exists(stub_file))) {
-	throw DatabaseOpeningError("Couldn't detect type of database");
+    if (usual(file_exists(stub_file))) {
+	open_stub(*this, stub_file);
+	return;
     }
 
-    open_stub(*this, stub_file);
+#ifndef XAPIAN_HAS_CHERT_BACKEND
+    if (file_exists(path + "/iamchert")) {
+	throw FeatureUnavailableError("Chert backend disabled");
+    }
+#endif
+#ifndef XAPIAN_HAS_GLASS_BACKEND
+    if (file_exists(path + "/iamglass")) {
+	throw FeatureUnavailableError("Glass backend disabled");
+    }
+#endif
+    if (file_exists(path + "/iamflint")) {
+	throw FeatureUnavailableError("Flint backend no longer supported");
+    }
+
+    throw DatabaseOpeningError("Couldn't detect type of database");
 }
 
-#if defined XAPIAN_HAS_FLINT_BACKEND || \
-    defined XAPIAN_HAS_CHERT_BACKEND || \
-    defined XAPIAN_HAS_BRASS_BACKEND
+Database::Database(int fd, int flags)
+{
+    LOGCALL_CTOR(API, "Database", fd|flags);
+    if (rare(fd < 0))
+	throw InvalidArgumentError("fd < 0");
+
+#ifdef XAPIAN_HAS_GLASS_BACKEND
+    int type = flags & DB_BACKEND_MASK_;
+    switch (type) {
+	case 0: case DB_BACKEND_GLASS:
+	    internal.push_back(new GlassDatabase(fd));
+	    return;
+    }
+#else
+    (void)flags;
+#endif
+
+    (void)::close(fd);
+    throw DatabaseOpeningError("Couldn't detect type of database");
+}
+
+#if defined XAPIAN_HAS_CHERT_BACKEND || \
+    defined XAPIAN_HAS_GLASS_BACKEND
 #define HAVE_DISK_BACKEND
 #endif
 
-WritableDatabase::WritableDatabase(const std::string &path, int action)
+WritableDatabase::WritableDatabase(const std::string &path, int flags, int block_size)
     : Database()
 {
-    LOGCALL_CTOR(API, "WritableDatabase", path | action);
-#ifdef HAVE_DISK_BACKEND
-    enum {
-#ifdef XAPIAN_HAS_CHERT_BACKEND
-	CHERT,
-#endif
-#ifdef XAPIAN_HAS_FLINT_BACKEND
-	FLINT,
-#endif
-#ifdef XAPIAN_HAS_BRASS_BACKEND
-	BRASS,
-#endif
-	UNSET
-    } type = UNSET;
-#endif
-    struct stat statbuf;
-    if (stat(path, &statbuf) == -1) {
-	// ENOENT probably just means that we need to create the directory.
-	if (errno != ENOENT)
-	    throw DatabaseOpeningError("Couldn't stat '" + path + "'", errno);
-    } else {
-	// File or directory already exists.
-
-	if (S_ISREG(statbuf.st_mode)) {
-	    // The path is a file, so assume it is a stub database file.
-	    open_stub(*this, path, action);
-	    return;
-	}
-
-	if (rare(!S_ISDIR(statbuf.st_mode))) {
-	    throw DatabaseOpeningError("Not a regular file or directory: '" + path + "'");
-	}
-
-	if (file_exists(path + "/iamchert")) {
-	    // Existing chert DB.
-#ifdef XAPIAN_HAS_CHERT_BACKEND
-	    type = CHERT;
-#else
-	    throw FeatureUnavailableError("Chert backend disabled");
-#endif
-	} else if (file_exists(path + "/iamflint")) {
-	    // Existing flint DB.
-#ifdef XAPIAN_HAS_FLINT_BACKEND
-	    type = FLINT;
-#else
-	    throw FeatureUnavailableError("Flint backend disabled");
-#endif
-	} else if (file_exists(path + "/iambrass")) {
-	    // Existing brass DB.
-#ifdef XAPIAN_HAS_BRASS_BACKEND
-	    type = BRASS;
-#else
-	    throw FeatureUnavailableError("Brass backend disabled");
-#endif
+    LOGCALL_CTOR(API, "WritableDatabase", path|flags|block_size);
+    // Avoid warning if both chert and glass are disabled.
+    (void)block_size;
+    int type = flags & DB_BACKEND_MASK_;
+    // Clear the backend bits, so we just pass on other flags to open_stub, etc.
+    flags &= ~DB_BACKEND_MASK_;
+    if (type == 0) {
+	struct stat statbuf;
+	if (stat(path.c_str(), &statbuf) == -1) {
+	    // ENOENT probably just means that we need to create the directory.
+	    if (errno != ENOENT)
+		throw DatabaseOpeningError("Couldn't stat '" + path + "'", errno);
 	} else {
-	    // Check for "stub directories".
-	    string stub_file = path;
-	    stub_file += "/XAPIANDB";
-	    if (usual(file_exists(stub_file))) {
-		open_stub(*this, stub_file, action);
+	    // File or directory already exists.
+
+	    if (S_ISREG(statbuf.st_mode)) {
+		// The path is a file, so assume it is a stub database file.
+		open_stub(*this, path, flags);
 		return;
 	    }
+
+	    if (rare(!S_ISDIR(statbuf.st_mode))) {
+		throw DatabaseOpeningError("Not a regular file or directory: '" + path + "'");
+	    }
+
+	    if (file_exists(path + "/iamchert")) {
+		// Existing chert DB.
+#ifdef XAPIAN_HAS_CHERT_BACKEND
+		type = DB_BACKEND_CHERT;
+#else
+		throw FeatureUnavailableError("Chert backend disabled");
+#endif
+	    } else if (file_exists(path + "/iamglass")) {
+		// Existing glass DB.
+#ifdef XAPIAN_HAS_GLASS_BACKEND
+		type = DB_BACKEND_GLASS;
+#else
+		throw FeatureUnavailableError("Glass backend disabled");
+#endif
+	    } else if (file_exists(path + "/iamflint")) {
+		// Existing flint DB.
+		throw FeatureUnavailableError("Flint backend no longer supported");
+	    } else {
+		// Check for "stub directories".
+		string stub_file = path;
+		stub_file += "/XAPIANDB";
+		if (usual(file_exists(stub_file))) {
+		    open_stub(*this, stub_file, flags);
+		    return;
+		}
+	    }
 	}
     }
 
-#ifdef HAVE_DISK_BACKEND
     switch (type) {
-	case UNSET: {
-#ifdef XAPIAN_HAS_BRASS_BACKEND
-	    // If only brass is enabled, there's no point checking the
-	    // environmental variable.
-#if defined XAPIAN_HAS_CHERT_BACKEND || defined XAPIAN_HAS_FLINT_BACKEND
-	    // If $XAPIAN_PREFER_BRASS is set to a non-empty value, prefer brass
-	    // if there's no existing database.
-	    const char *p = getenv("XAPIAN_PREFER_BRASS");
-	    if (p && *p)
-	       	goto brass;
+	case DB_BACKEND_STUB:
+	    open_stub(*this, path, flags);
+	    return;
+	case 0:
+	    // Fall through to first enabled case, so order the remaining cases
+	    // by preference.
+#ifdef XAPIAN_HAS_GLASS_BACKEND
+	case DB_BACKEND_GLASS:
+	    internal.push_back(new GlassWritableDatabase(path, flags, block_size));
+	    return;
 #endif
-#endif
-	}
-	// Fall through to first enabled case, so order the remaining cases
-	// by preference.
 #ifdef XAPIAN_HAS_CHERT_BACKEND
-	case CHERT:
-	    internal.push_back(new ChertWritableDatabase(path, action, 8192));
-	    break;
+	case DB_BACKEND_CHERT:
+	    internal.push_back(new ChertWritableDatabase(path, flags, block_size));
+	    return;
 #endif
-#ifdef XAPIAN_HAS_FLINT_BACKEND
-	case FLINT:
-	    internal.push_back(new FlintWritableDatabase(path, action, 8192));
-	    break;
-#endif
-#ifdef XAPIAN_HAS_BRASS_BACKEND
-	case BRASS:
-#if defined XAPIAN_HAS_CHERT_BACKEND || defined XAPIAN_HAS_FLINT_BACKEND
-brass:
-#endif
-	    internal.push_back(new BrassWritableDatabase(path, action, 8192));
-	    break;
+	case DB_BACKEND_INMEMORY:
+#ifdef XAPIAN_HAS_INMEMORY_BACKEND
+	    internal.push_back(new InMemoryDatabase());
+	    return;
+#else
+	    throw FeatureUnavailableError("Inmemory backend disabled");
 #endif
     }
-#else
+#ifndef HAVE_DISK_BACKEND
     throw FeatureUnavailableError("No disk-based writable backend is enabled");
 #endif
 }
