@@ -30,7 +30,7 @@
 #include <unistd.h>
 #include <cstdlib>
 #include <sys/types.h>
-#include <sys/socket.h>
+#include "safesyssocket.h"
 #include <sys/wait.h>
 #include <signal.h>
 #include <cstring>
@@ -45,6 +45,10 @@
 # include <sys/cygwin.h>
 #endif
 
+#ifdef FLINTLOCK_USE_FLOCK
+# include <sys/file.h>
+#endif
+
 #include "xapian/error.h"
 
 using namespace std;
@@ -60,7 +64,7 @@ using namespace std;
 #endif
 
 FlintLock::reason
-FlintLock::lock(bool exclusive, string & explanation) {
+FlintLock::lock(bool exclusive, bool wait, string & explanation) {
     // Currently we only support exclusive locks.
     (void)exclusive;
     Assert(exclusive);
@@ -81,26 +85,66 @@ FlintLock::lock(bool exclusive, string & explanation) {
 #else
     const char *fnm = filename.c_str();
 #endif
+retry:
+    // FIXME: Use LockFileEx() for locking, which would allow proper blocking
+    // and also byte-range locking for when we implement MVCC.  But is there a
+    // way to interwork with the CreateFile()-based locking while doing so?
     hFile = CreateFile(fnm, GENERIC_WRITE, FILE_SHARE_READ,
 		       NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
     if (hFile != INVALID_HANDLE_VALUE) return SUCCESS;
-    if (GetLastError() == ERROR_ALREADY_EXISTS) return INUSE;
-    explanation = string();
-    return UNKNOWN;
-#elif defined __EMX__
-    APIRET rc;
-    ULONG ulAction;
-    rc = DosOpen((PCSZ)filename.c_str(), &hFile, &ulAction, 0, FILE_NORMAL,
-		 OPEN_ACTION_OPEN_IF_EXISTS | OPEN_ACTION_CREATE_IF_NEW,
-		 OPEN_SHARE_DENYWRITE | OPEN_ACCESS_WRITEONLY,
-		 NULL);
-    if (rc == NO_ERROR) return SUCCESS;
-    if (rc == ERROR_ACCESS_DENIED) return INUSE;
+    if (GetLastError() == ERROR_ALREADY_EXISTS) {
+	if (wait) {
+	    Sleep(1000);
+	    goto retry;
+	}
+	return INUSE;
+    }
     explanation = string();
     return UNKNOWN;
+#elif defined FLINTLOCK_USE_FLOCK
+    // This is much simpler than using fcntl() due to saner semantics around
+    // releasing locks when closing other descriptors on the same file (at
+    // least on platforms where flock() isn't just a compatibility wrapper
+    // around fcntl()).  We can't simply switch to this without breaking
+    // locking compatibility with previous releases, though it might be useful
+    // for porting to platforms without fcntl() locking.  Also, flock()
+    // apparently doesn't work over NFS - perhaps that's OK, but we should at
+    // least check the failure mode.
+    Assert(fd == -1);
+    int lockfd = open(filename.c_str(), O_WRONLY | O_CREAT | O_TRUNC | O_CLOEXEC, 0666);
+    if (lockfd < 0) {
+	// Couldn't open lockfile.
+	explanation.assign("Couldn't open lockfile: ");
+	errno_to_string(errno, explanation);
+	return ((errno == EMFILE || errno == ENFILE) ? FDLIMIT : UNKNOWN);
+    }
+
+    int op = LOCK_EX;
+    if (!wait) op |= LOCK_NB;
+    while (flock(lockfd, op) == -1) {
+	if (errno != EINTR) {
+	    // Lock failed - translate known errno values into a reason code.
+	    close(lockfd);
+	    switch (errno) {
+		case EWOULDBLOCK:
+		    return INUSE;
+		case ENOLCK:
+		    return UNSUPPORTED; // FIXME: what do we get for NFS?
+		default:
+		    return UNKNOWN;
+	    }
+	}
+    }
+
+    fd = lockfd;
+    return SUCCESS;
 #else
     Assert(fd == -1);
+#if defined F_SETFD && defined FD_CLOEXEC
+    int lockfd = open(filename.c_str(), O_WRONLY | O_CREAT | O_TRUNC | O_CLOEXEC, 0666);
+#else
     int lockfd = open(filename.c_str(), O_WRONLY | O_CREAT | O_TRUNC, 0666);
+#endif
     if (lockfd < 0) {
 	// Couldn't open lockfile.
 	explanation.assign("Couldn't open lockfile: ");
@@ -126,7 +170,7 @@ FlintLock::lock(bool exclusive, string & explanation) {
 	fl.l_start = 0;
 	fl.l_len = 1;
 	fl.l_pid = 0;
-	while (fcntl(lockfd, F_OFD_SETLK, &fl) == -1) {
+	while (fcntl(lockfd, wait ? F_OFD_SETLKW : F_OFD_SETLK, &fl) == -1) {
 	    if (errno != EINTR) {
 		if (errno == EINVAL) {
 		    // F_OFD_SETLK not supported by this kernel.
@@ -201,7 +245,7 @@ no_ofd_support:
     }
 
     int fds[2];
-    if (socketpair(AF_UNIX, SOCK_STREAM, PF_UNSPEC, fds) < 0) {
+    if (socketpair(AF_UNIX, SOCK_STREAM|SOCK_CLOEXEC, PF_UNSPEC, fds) < 0) {
 	// Couldn't create socketpair.
 	explanation.assign("Couldn't create socketpair: ");
 	errno_to_string(errno, explanation);
@@ -216,6 +260,20 @@ no_ofd_support:
 	// Child process.
 	close(fds[0]);
 
+#if defined F_SETFD && defined FD_CLOEXEC
+	// Clear close-on-exec flag, if we set it when we called socketpair().
+	// Clearing it here means there's no window where another thread in the
+	// parent process could fork()+exec() and end up with this fd still
+	// open (assuming close-on-exec is supported).
+	//
+	// We can't use a preprocessor check on the *value* of SOCK_CLOEXEC as
+	// on Linux SOCK_CLOEXEC is an enum, with '#define SOCK_CLOEXEC
+	// SOCK_CLOEXEC' to allow '#ifdef SOCK_CLOEXEC' to work.
+	if (SOCK_CLOEXEC != 0)
+	    (void)fcntl(fds[1], F_SETFD, 0);
+	if (O_CLOEXEC != 0)
+	    (void)fcntl(lockfd, F_SETFD, 0);
+#endif
 	// Connect pipe to stdin and stdout.
 	dup2(fds[1], 0);
 	dup2(fds[1], 1);
@@ -239,7 +297,7 @@ no_ofd_support:
 	    fl.l_whence = SEEK_SET;
 	    fl.l_start = 0;
 	    fl.l_len = 1;
-	    while (fcntl(lockfd, F_SETLK, &fl) == -1) {
+	    while (fcntl(lockfd, wait ? F_SETLKW : F_SETLK, &fl) == -1) {
 		if (errno != EINTR) {
 		    // Lock failed - translate known errno values into a reason
 		    // code.
@@ -268,7 +326,7 @@ no_ofd_support:
 	    if (why != SUCCESS) _exit(0);
 	}
 
-	// Make sure we don't block unmount() of partition holding the current
+	// Make sure we don't block unmount of partition holding the current
 	// directory.
 	if (chdir("/") < 0) {
 	    // We can't usefully do anything in response to an error, so just
@@ -342,10 +400,10 @@ FlintLock::release() {
     if (hFile == INVALID_HANDLE_VALUE) return;
     CloseHandle(hFile);
     hFile = INVALID_HANDLE_VALUE;
-#elif defined __EMX__
-    if (hFile == NULLHANDLE) return;
-    DosClose(hFile);
-    hFile = NULLHANDLE;
+#elif defined FLINTLOCK_USE_FLOCK
+    if (fd < 0) return;
+    close(fd);
+    fd = -1;
 #else
     if (fd < 0) return;
     close(fd);
