@@ -1,9 +1,9 @@
-/** \file database.h
- * \brief API for working with Xapian databases
+/** @file database.h
+ * @brief API for working with Xapian databases
  */
 /* Copyright 1999,2000,2001 BrightStation PLC
  * Copyright 2002 Ananova Ltd
- * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2011 Olly Betts
+ * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2011,2012,2013,2014,2015,2016 Olly Betts
  * Copyright 2006,2008 Lemur Consulting Ltd
  *
  * This program is free software; you can redistribute it and/or
@@ -25,11 +25,17 @@
 #ifndef XAPIAN_INCLUDED_DATABASE_H
 #define XAPIAN_INCLUDED_DATABASE_H
 
+#if !defined XAPIAN_IN_XAPIAN_H && !defined XAPIAN_LIB_BUILD
+# error "Never use <xapian/database.h> directly; include <xapian.h> instead."
+#endif
+
+#include <iosfwd>
 #include <string>
 #include <vector>
 
-#include <xapian/base.h>
-#include <xapian/document.h>
+#include <xapian/attributes.h>
+#include <xapian/deprecated.h>
+#include <xapian/intrusive_ptr.h>
 #include <xapian/types.h>
 #include <xapian/positioniterator.h>
 #include <xapian/postingiterator.h>
@@ -39,6 +45,9 @@
 
 namespace Xapian {
 
+class Compactor;
+class Document;
+
 /** This class is used to access a database, or a group of databases.
  *
  *  For searching, this class is used in conjunction with an Enquire object.
@@ -54,27 +63,21 @@ namespace Xapian {
  *  which uses an incompatible format).
  */
 class XAPIAN_VISIBILITY_DEFAULT Database {
+	/// @internal Implementation behind check() static methods.
+	static size_t check_(const std::string * path_ptr, int fd, int opts,
+			     std::ostream *out);
+
+	/// Internal helper behind public compact() methods.
+	void compact_(const std::string * output_ptr,
+		      int fd,
+		      unsigned flags,
+		      int block_size,
+		      Xapian::Compactor * compactor) const;
+
     public:
 	class Internal;
 	/// @private @internal Reference counted internals.
-	std::vector<Xapian::Internal::RefCntPtr<Internal> > internal;
-
-	/** @private @internal Get a document from the database, but doesn't
-	 *  need to check if it exists.
-	 *
-	 *  This method returns a Xapian::Document object which provides the
-	 *  information about a document.  If the document doesn't exist,
-	 *  either a NULL pointer may be returned, or the returned object will
-	 *  throw DocNotFoundError when you try to access it.
-	 *
-	 *  The caller should delete the returned object when it has finished
-	 *  with it.
-	 *
-	 *  @param did   The document id of the document to retrieve.
-	 *
-	 *  @return      Pointer to Document::Internal object.
-	 */
-	Document::Internal * get_document_lazily(Xapian::docid did) const;
+	std::vector<Xapian::Internal::intrusive_ptr<Internal> > internal;
 
 	/** Add an existing database (or group of databases) to those
 	 *  accessed by this object.
@@ -92,7 +95,20 @@ class XAPIAN_VISIBILITY_DEFAULT Database {
 	 *
 	 * @param path directory that the database is stored in.
 	 */
-	explicit Database(const std::string &path);
+	explicit Database(const std::string &path, int flags = 0);
+
+	/** Open a single-file Database.
+	 *
+	 *  This method opens a single-file Database given a file descriptor
+	 *  open on it.  Xapian looks starting at the current file offset,
+	 *  allowing a single file database to be easily embedded within
+	 *  another file.
+	 *
+	 * @param fd  file descriptor for the file.  Xapian takes ownership of
+	 *            this and will close it when the database is closed.
+	 * @param flags  Bitwise-or of Xapian::DB_* constants.
+	 */
+	explicit Database(int fd, int flags = 0);
 
 	/** @private @internal Create a Database from its internals.
 	 */
@@ -127,8 +143,14 @@ class XAPIAN_VISIBILITY_DEFAULT Database {
 	 *
 	 *  Calling reopen() on a database which has been closed (with @a
 	 *  close()) will always raise a Xapian::DatabaseError.
+	 *
+	 *  @return	true if the database might have been reopened (if false
+	 *		is returned, the database definitely hasn't been
+	 *		reopened, which applications may find useful when
+	 *		caching results, etc).  In Xapian < 1.3.0, this method
+	 *		did not return a value.
 	 */
-	void reopen();
+	bool reopen();
 
 	/** Close the database.
 	 *
@@ -183,7 +205,7 @@ class XAPIAN_VISIBILITY_DEFAULT Database {
 
 	/** Corresponding end iterator to postlist_begin().
 	 */
-	PostingIterator postlist_end(const std::string &) const {
+	PostingIterator XAPIAN_NOTHROW(postlist_end(const std::string &) const) {
 	    return PostingIterator();
 	}
 
@@ -196,7 +218,7 @@ class XAPIAN_VISIBILITY_DEFAULT Database {
 
 	/** Corresponding end iterator to termlist_begin().
 	 */
-	TermIterator termlist_end(Xapian::docid) const {
+	TermIterator XAPIAN_NOTHROW(termlist_end(Xapian::docid) const) {
 	    return TermIterator();
 	}
 
@@ -210,36 +232,20 @@ class XAPIAN_VISIBILITY_DEFAULT Database {
 
 	/** Corresponding end iterator to positionlist_begin().
 	 */
-	PositionIterator positionlist_end(Xapian::docid, const std::string &) const {
+	PositionIterator XAPIAN_NOTHROW(positionlist_end(Xapian::docid, const std::string &) const) {
 	    return PositionIterator();
 	}
 
-	/** An iterator which runs across all terms in the database.
-	 */
-	TermIterator allterms_begin() const;
-
-	/** Corresponding end iterator to allterms_begin().
-	 */
-	TermIterator allterms_end() const {
-	    return TermIterator();
-	}
-
 	/** An iterator which runs across all terms with a given prefix.
 	 *
-	 *  This is functionally similar to getting an iterator with
-	 *  allterms_begin() and then calling skip_to(prefix) on that iterator
-	 *  to move to the start of the prefix, but is more convenient (because
-	 *  it detects the end of the prefixed terms), and may be more
-	 *  efficient than simply calling skip_to() after opening the iterator,
-	 *  particularly for remote databases.
-	 *
-	 *  @param prefix The prefix to restrict the returned terms to.
+	 *  @param prefix The prefix to restrict the returned terms to (default:
+	 *		  iterate all terms)
 	 */
-	TermIterator allterms_begin(const std::string & prefix) const;
+	TermIterator allterms_begin(const std::string & prefix = std::string()) const;
 
 	/** Corresponding end iterator to allterms_begin(prefix).
 	 */
-	TermIterator allterms_end(const std::string &) const {
+	TermIterator XAPIAN_NOTHROW(allterms_end(const std::string & = std::string()) const) {
 	    return TermIterator();
 	}
 
@@ -328,13 +334,16 @@ class XAPIAN_VISIBILITY_DEFAULT Database {
 	ValueIterator valuestream_begin(Xapian::valueno slot) const;
 
 	/// Return end iterator corresponding to valuestream_begin().
-	ValueIteratorEnd_ valuestream_end(Xapian::valueno) const {
-	    return ValueIteratorEnd_();
+	ValueIterator XAPIAN_NOTHROW(valuestream_end(Xapian::valueno) const) {
+	    return ValueIterator();
 	}
 
 	/// Get the length of a document.
 	Xapian::termcount get_doclength(Xapian::docid did) const;
 
+	/// Get the number of unique terms in document.
+	Xapian::termcount get_unique_terms(Xapian::docid did) const;
+
 	/** Send a "keep-alive" to remote databases to stop them timing out.
 	 *
 	 *  Has no effect on non-remote databases.
@@ -379,7 +388,7 @@ class XAPIAN_VISIBILITY_DEFAULT Database {
 	Xapian::TermIterator spellings_begin() const;
 
 	/// Corresponding end iterator to spellings_begin().
-	Xapian::TermIterator spellings_end() const {
+	Xapian::TermIterator XAPIAN_NOTHROW(spellings_end() const) {
 	    return Xapian::TermIterator();
 	}
 
@@ -390,7 +399,7 @@ class XAPIAN_VISIBILITY_DEFAULT Database {
 	Xapian::TermIterator synonyms_begin(const std::string &term) const;
 
 	/// Corresponding end iterator to synonyms_begin(term).
-	Xapian::TermIterator synonyms_end(const std::string &) const {
+	Xapian::TermIterator XAPIAN_NOTHROW(synonyms_end(const std::string &) const) {
 	    return Xapian::TermIterator();
 	}
 
@@ -402,7 +411,7 @@ class XAPIAN_VISIBILITY_DEFAULT Database {
 	Xapian::TermIterator synonym_keys_begin(const std::string &prefix = std::string()) const;
 
 	/// Corresponding end iterator to synonym_keys_begin(prefix).
-	Xapian::TermIterator synonym_keys_end(const std::string & = std::string()) const {
+	Xapian::TermIterator XAPIAN_NOTHROW(synonym_keys_end(const std::string & = std::string()) const) {
 	    return Xapian::TermIterator();
 	}
 
@@ -453,7 +462,7 @@ class XAPIAN_VISIBILITY_DEFAULT Database {
 	Xapian::TermIterator metadata_keys_begin(const std::string &prefix = std::string()) const;
 
 	/// Corresponding end iterator to metadata_keys_begin().
-	Xapian::TermIterator metadata_keys_end(const std::string & = std::string()) const {
+	Xapian::TermIterator XAPIAN_NOTHROW(metadata_keys_end(const std::string & = std::string()) const) {
 	    return Xapian::TermIterator();
 	}
 
@@ -472,6 +481,227 @@ class XAPIAN_VISIBILITY_DEFAULT Database {
 	 *  contain the UUIDs of all the sub-databases.
 	 */
 	std::string get_uuid() const;
+
+	/** Check the integrity of a database or database table.
+	 *
+	 *  @param path	Path to database or table
+	 *  @param opts	Options to use for check
+	 *  @param out	std::ostream to write output to (NULL for no output)
+	 */
+	static size_t check(const std::string & path, int opts = 0,
+			    std::ostream *out = NULL) {
+	    return check_(&path, 0, opts, out);
+	}
+
+	/** Check the integrity of a single file database.
+	 *
+	 *  @param fd   file descriptor for the database.  The current file
+	 *              offset is used, allowing checking a single file
+	 *              database which is embedded within another file.  Xapian
+	 *              takes ownership of the file descriptor and will close
+	 *              it before returning.
+	 *  @param opts	Options to use for check
+	 *  @param out	std::ostream to write output to (NULL for no output)
+	 */
+	static size_t check(int fd, int opts = 0, std::ostream *out = NULL) {
+	    return check_(NULL, fd, opts, out);
+	}
+
+	/** Produce a compact version of this database.
+	 *
+	 *  New 1.3.4.  Various methods of the Compactor class were deprecated
+	 *  in 1.3.4.
+	 *
+	 *  @param output Path to write the compact version to.
+	 *		  This can be the same as an input if that input is a
+	 *		  stub database (in which case the database(s) listed
+	 *		  in the stub will be compacted to a new database and
+	 *		  then the stub will be atomically updated to point to
+	 *		  this new database).
+	 *
+	 *  @param flags Any of the following combined using bitwise-or (| in
+	 *		 C++):
+	 *   - Xapian::DBCOMPACT_NO_RENUMBER By default the document ids will
+	 *		be renumbered the output - currently by applying the
+	 *		same offset to all the document ids in a particular
+	 *		source database.  If this flag is specified, then this
+	 *		renumbering doesn't happen, but all the document ids
+	 *		must be unique over all source databases.  Currently
+	 *		the ranges of document ids in each source must not
+	 *		overlap either, though this restriction may be removed
+	 *		in the future.
+	 *   - Xapian::DBCOMPACT_MULTIPASS
+	 *		If merging more than 3 databases, merge the postlists
+	 *		in multiple passes, which is generally faster but
+	 *		requires more disk space for temporary files.
+	 *   - Xapian::DBCOMPACT_SINGLE_FILE
+	 *		Produce a single-file database (only supported for
+	 *		glass currently).
+	 *
+	 *  @param block_size This specifies the block size (in bytes) for
+	 *		to use for the output.  For glass, the block size must
+	 *		be a power of 2 between 2048 and 65536 (inclusive), and
+	 *		the default (also used if an invalid value is passed)
+	 *		is 8192 bytes.
+	 */
+	void compact(const std::string & output,
+		     unsigned flags = 0,
+		     int block_size = 0) {
+	    compact_(&output, 0, flags, block_size, NULL);
+	}
+
+	/** Produce a compact version of this database.
+	 *
+	 *  New 1.3.4.  Various methods of the Compactor class were deprecated
+	 *  in 1.3.4.
+	 *
+	 *  This variant writes a single-file database to the specified file
+	 *  descriptor.  Only the glass backend supports such databases, so
+	 *  this form is only supported for this backend.
+	 *
+	 *  @param fd   File descriptor to write the compact version to.  The
+	 *		descriptor needs to be readable and writable (open with
+	 *		O_RDWR) and seekable.  The current file offset is used,
+	 *		allowing compacting to a single file database embedded
+	 *		within another file.  Xapian takes ownership of the
+	 *		file descriptor and will close it before returning.
+	 *
+	 *  @param flags Any of the following combined using bitwise-or (| in
+	 *		C++):
+	 *   - Xapian::DBCOMPACT_NO_RENUMBER By default the document ids will
+	 *		be renumbered the output - currently by applying the
+	 *		same offset to all the document ids in a particular
+	 *		source database.  If this flag is specified, then this
+	 *		renumbering doesn't happen, but all the document ids
+	 *		must be unique over all source databases.  Currently
+	 *		the ranges of document ids in each source must not
+	 *		overlap either, though this restriction may be removed
+	 *		in the future.
+	 *   - Xapian::DBCOMPACT_MULTIPASS
+	 *		If merging more than 3 databases, merge the postlists
+	 *		in multiple passes, which is generally faster but
+	 *		requires more disk space for temporary files.
+	 *   - Xapian::DBCOMPACT_SINGLE_FILE
+	 *		Produce a single-file database (only supported for
+	 *		glass currently) - this flag is implied in this form
+	 *		and need not be specified explicitly.
+	 *
+	 *  @param block_size This specifies the block size (in bytes) for
+	 *		to use for the output.  For glass, the block size must
+	 *		be a power of 2 between 2048 and 65536 (inclusive), and
+	 *		the default (also used if an invalid value is passed)
+	 *		is 8192 bytes.
+	 */
+	void compact(int fd,
+		     unsigned flags = 0,
+		     int block_size = 0) {
+	    compact_(NULL, fd, flags, block_size, NULL);
+	}
+
+	/** Produce a compact version of this database.
+	 *
+	 *  New 1.3.4.  Various methods of the Compactor class were deprecated
+	 *  in 1.3.4.
+	 *
+	 *  The @a compactor functor allows handling progress output and
+	 *  specifying how user metadata is merged.
+	 *
+	 *  @param output Path to write the compact version to.
+	 *		  This can be the same as an input if that input is a
+	 *		  stub database (in which case the database(s) listed
+	 *		  in the stub will be compacted to a new database and
+	 *		  then the stub will be atomically updated to point to
+	 *		  this new database).
+	 *
+	 *  @param flags Any of the following combined using bitwise-or (| in
+	 *		 C++):
+	 *   - Xapian::DBCOMPACT_NO_RENUMBER By default the document ids will
+	 *		be renumbered the output - currently by applying the
+	 *		same offset to all the document ids in a particular
+	 *		source database.  If this flag is specified, then this
+	 *		renumbering doesn't happen, but all the document ids
+	 *		must be unique over all source databases.  Currently
+	 *		the ranges of document ids in each source must not
+	 *		overlap either, though this restriction may be removed
+	 *		in the future.
+	 *   - Xapian::DBCOMPACT_MULTIPASS
+	 *		If merging more than 3 databases, merge the postlists
+	 *		in multiple passes, which is generally faster but
+	 *		requires more disk space for temporary files.
+	 *   - Xapian::DBCOMPACT_SINGLE_FILE
+	 *		Produce a single-file database (only supported for
+	 *		glass currently).
+	 *
+	 *  @param block_size This specifies the block size (in bytes) for
+	 *		to use for the output.  For glass, the block size must
+	 *		be a power of 2 between 2048 and 65536 (inclusive), and
+	 *		the default (also used if an invalid value is passed)
+	 *		is 8192 bytes.
+	 *
+	 *  @param compactor Functor
+	 */
+	void compact(const std::string & output,
+		     unsigned flags,
+		     int block_size,
+		     Xapian::Compactor & compactor)
+	{
+	    compact_(&output, 0, flags, block_size, &compactor);
+	}
+
+	/** Produce a compact version of this database.
+	 *
+	 *  New 1.3.4.  Various methods of the Compactor class were deprecated
+	 *  in 1.3.4.
+	 *
+	 *  The @a compactor functor allows handling progress output and
+	 *  specifying how user metadata is merged.
+	 *
+	 *  This variant writes a single-file database to the specified file
+	 *  descriptor.  Only the glass backend supports such databases, so
+	 *  this form is only supported for this backend.
+	 *
+	 *  @param fd   File descriptor to write the compact version to.  The
+	 *		descriptor needs to be readable and writable (open with
+	 *		O_RDWR) and seekable.  The current file offset is used,
+	 *		allowing compacting to a single file database embedded
+	 *		within another file.  Xapian takes ownership of the
+	 *		file descriptor and will close it before returning.
+	 *
+	 *  @param flags Any of the following combined using bitwise-or (| in
+	 *		 C++):
+	 *   - Xapian::DBCOMPACT_NO_RENUMBER By default the document ids will
+	 *		be renumbered the output - currently by applying the
+	 *		same offset to all the document ids in a particular
+	 *		source database.  If this flag is specified, then this
+	 *		renumbering doesn't happen, but all the document ids
+	 *		must be unique over all source databases.  Currently
+	 *		the ranges of document ids in each source must not
+	 *		overlap either, though this restriction may be removed
+	 *		in the future.
+	 *   - Xapian::DBCOMPACT_MULTIPASS
+	 *		If merging more than 3 databases, merge the postlists
+	 *		in multiple passes, which is generally faster but
+	 *		requires more disk space for temporary files.
+	 *   - Xapian::DBCOMPACT_SINGLE_FILE
+	 *		Produce a single-file database (only supported for
+	 *		glass currently) - this flag is implied in this form
+	 *		and need not be specified explicitly.
+	 *
+	 *  @param block_size This specifies the block size (in bytes) for
+	 *		to use for the output.  For glass, the block size must
+	 *		be a power of 2 between 2048 and 65536 (inclusive), and
+	 *		the default (also used if an invalid value is passed)
+	 *		is 8192 bytes.
+	 *
+	 *  @param compactor Functor
+	 */
+	void compact(int fd,
+		     unsigned flags,
+		     int block_size,
+		     Xapian::Compactor & compactor)
+	{
+	    compact_(NULL, fd, flags, block_size, &compactor);
+	}
 };
 
 /** This class provides read/write access to a database.
@@ -492,7 +722,10 @@ class XAPIAN_VISIBILITY_DEFAULT WritableDatabase : public Database {
 	 */
 	virtual ~WritableDatabase();
 
-	/** Create an empty WritableDatabase.
+	/** Create a WritableDatabase with no subdatabases.
+	 *
+	 *  The created object isn't very useful in this state - it's intended
+	 *  as a placeholder value.
 	 */
 	WritableDatabase();
 
@@ -504,21 +737,37 @@ class XAPIAN_VISIBILITY_DEFAULT WritableDatabase : public Database {
 	 *  exist (but only the leaf directory, not recursively).
 	 *
 	 * @param path directory that the database is stored in.
-	 * @param action one of:
+	 * @param flags one of:
 	 *  - Xapian::DB_CREATE_OR_OPEN open for read/write; create if no db
-	 *    exists
+	 *    exists (the default if flags isn't specified)
 	 *  - Xapian::DB_CREATE create new database; fail if db exists
 	 *  - Xapian::DB_CREATE_OR_OVERWRITE overwrite existing db; create if
 	 *    none exists
 	 *  - Xapian::DB_OPEN open for read/write; fail if no db exists
 	 *
+	 *  Additionally, the following flags can be combined with action
+	 *  using bitwise-or (| in C++):
+	 *
+	 *   - Xapian::DB_NO_SYNC don't call fsync() or similar
+	 *   - Xapian::DB_DANGEROUS don't be crash-safe, no concurrent readers
+	 *   - Xapian::DB_RETRY_LOCK to wait to get a write lock
+	 *
+	 *  @param block_size If a new database is created, this specifies
+	 *		      the block size (in bytes) for backends which
+	 *		      have such a concept.  For chert and glass, the
+	 *		      block size must be a power of 2 between 2048 and
+	 *		      65536 (inclusive), and the default (also used if
+	 *		      an invalid value is passed) is 8192 bytes.
+	 *
 	 *  @exception Xapian::DatabaseCorruptError will be thrown if the
 	 *             database is in a corrupt state.
 	 *
 	 *  @exception Xapian::DatabaseLockError will be thrown if a lock
 	 *	       couldn't be acquired on the database.
 	 */
-	WritableDatabase(const std::string &path, int action);
+	explicit WritableDatabase(const std::string &path,
+				  int flags = 0,
+				  int block_size = 0);
 
 	/** @private @internal Create an WritableDatabase given its internals.
 	 */
@@ -584,10 +833,9 @@ class XAPIAN_VISIBILITY_DEFAULT WritableDatabase : public Database {
 
 	/** Pre-1.1.0 name for commit().
 	 *
-	 *  Use commit() instead in new code.  This alias may be deprecated in
-	 *  the future.
+	 *  Use commit() instead.
 	 */
-	void flush() { commit(); }
+	XAPIAN_DEPRECATED(void flush()) { commit(); }
 
 	/** Begin a transaction.
 	 *
@@ -924,15 +1172,6 @@ class XAPIAN_VISIBILITY_DEFAULT WritableDatabase : public Database {
 	std::string get_description() const;
 };
 
-/** Open for read/write; create if no db exists. */
-const int DB_CREATE_OR_OPEN = 1;
-/** Create a new database; fail if db exists. */
-const int DB_CREATE = 2;
-/** Overwrite existing db; create if none exists. */
-const int DB_CREATE_OR_OVERWRITE = 3;
-/** Open for read/write; fail if no db exists. */
-const int DB_OPEN = 4;
-
 }
 
 #endif /* XAPIAN_INCLUDED_DATABASE_H */
