@@ -1,7 +1,7 @@
 /** @file postingsource.h
  *  @brief External sources of posting information
  */
-/* Copyright (C) 2007,2008,2009,2010,2011,2012,2015 Olly Betts
+/* Copyright (C) 2007,2008,2009,2010,2011,2012,2013,2014,2015,2016 Olly Betts
  * Copyright (C) 2008,2009 Lemur Consulting Ltd
  *
  * This program is free software; you can redistribute it and/or modify
@@ -22,8 +22,17 @@
 #ifndef XAPIAN_INCLUDED_POSTINGSOURCE_H
 #define XAPIAN_INCLUDED_POSTINGSOURCE_H
 
+#if !defined XAPIAN_IN_XAPIAN_H && !defined XAPIAN_LIB_BUILD
+# error "Never use <xapian/postingsource.h> directly; include <xapian.h> instead."
+#endif
+
+#include <xapian/attributes.h>
 #include <xapian/database.h>
+#include <xapian/deprecated.h>
+#include <xapian/intrusive_ptr.h>
+#include <xapian/postingiterator.h>
 #include <xapian/types.h>
+#include <xapian/valueiterator.h>
 #include <xapian/visibility.h>
 
 #include <string>
@@ -31,9 +40,12 @@
 
 namespace Xapian {
 
+class Registry;
+
 /** Base class which provides an "external" source of postings.
  */
-class XAPIAN_VISIBILITY_DEFAULT PostingSource {
+class XAPIAN_VISIBILITY_DEFAULT PostingSource
+    : public Xapian::Internal::opt_intrusive_base {
     /// Don't allow assignment.
     void operator=(const PostingSource &);
 
@@ -50,39 +62,11 @@ class XAPIAN_VISIBILITY_DEFAULT PostingSource {
      */
     void * matcher_;
 
-  protected:
-#ifdef SWIG
-  public: /* So SWIG wraps this constructor correctly for PHP */
-#endif
-    /// Allow subclasses to be instantiated.
-    PostingSource() : max_weight_(0), matcher_(NULL) { }
-#ifdef SWIG
-  protected:
-#endif
-
-    /** Set an upper bound on what get_weight() can return from now on.
-     *
-     *  This upper bound is used by the matcher to perform various
-     *  optimisations, so if you can return a good bound, then matches
-     *  will generally run faster.
-     *
-     *  This method should be called after calling init(), and may be called
-     *  during iteration if the upper bound drops.
-     *
-     *  It is valid for the posting source to have returned a higher value from
-     *  get_weight() earlier in the iteration, but the posting source must not
-     *  return a higher value from get_weight() than the currently set upper
-     *  bound, and the upper bound must not be increased (until init() has been
-     *  called).
-     *
-     *  If you don't call this method, the upper bound will default to 0, for
-     *  convenience when implementing "weight-less" PostingSource subclasses.
-     *
-     *  @param max_weight	The upper bound to set.
-     */
-    void set_maxweight(Xapian::weight max_weight);
-
   public:
+    /// Allow subclasses to be instantiated.
+    XAPIAN_NOTHROW(PostingSource())
+	: max_weight_(0), matcher_(NULL) { }
+
     /** @private @internal Set the object to inform of maxweight changes.
      *
      *  This method is for internal use only - it would be private except that
@@ -119,8 +103,32 @@ class XAPIAN_VISIBILITY_DEFAULT PostingSource {
      */
     virtual Xapian::doccount get_termfreq_max() const = 0;
 
+    /** Specify an upper bound on what get_weight() will return from now on.
+     *
+     *  This upper bound is used by the matcher to perform various
+     *  optimisations, so if you can return a good bound, then matches
+     *  will generally run faster.
+     *
+     *  This method should be called after calling init(), and may be called
+     *  during iteration if the upper bound drops.  It is probably only useful
+     *  to call from subclasses (it was actually a "protected" method prior to
+     *  Xapian 1.3.4, but that makes it tricky to wrap for other languages).
+     *
+     *  It is valid for the posting source to have returned a higher value from
+     *  get_weight() earlier in the iteration, but the posting source must not
+     *  return a higher value from get_weight() than the currently set upper
+     *  bound, and the upper bound must not be increased (until init() has been
+     *  called).
+     *
+     *  If you don't call this method, the upper bound will default to 0, for
+     *  convenience when implementing "weight-less" PostingSource subclasses.
+     *
+     *  @param max_weight	The upper bound to set.
+     */
+    void set_maxweight(double max_weight);
+
     /// Return the currently set upper bound on what get_weight() can return.
-    Xapian::weight get_maxweight() const { return max_weight_; }
+    double XAPIAN_NOTHROW(get_maxweight() const) { return max_weight_; }
 
     /** Return the weight contribution for the current document.
      *
@@ -135,7 +143,7 @@ class XAPIAN_VISIBILITY_DEFAULT PostingSource {
      *  next(), skip_to() or check(), and will ensure that the PostingSource is
      *  not at the end by calling at_end()).
      */
-    virtual Xapian::weight get_weight() const;
+    virtual double get_weight() const;
 
     /** Return the current docid.
      *
@@ -160,7 +168,7 @@ class XAPIAN_VISIBILITY_DEFAULT PostingSource {
      *  @param min_wt	The minimum weight contribution that is needed (this is
      *			just a hint which subclasses may ignore).
      */
-    virtual void next(Xapian::weight min_wt) = 0;
+    virtual void next(double min_wt) = 0;
 
     /** Advance to the specified docid.
      *
@@ -189,7 +197,7 @@ class XAPIAN_VISIBILITY_DEFAULT PostingSource {
      *  @param min_wt	The minimum weight contribution that is needed (this is
      *			just a hint which subclasses may ignore).
      */
-    virtual void skip_to(Xapian::docid did, Xapian::weight min_wt);
+    virtual void skip_to(Xapian::docid did, double min_wt);
 
     /** Check if the specified docid occurs.
      *
@@ -226,7 +234,7 @@ class XAPIAN_VISIBILITY_DEFAULT PostingSource {
      *  @param min_wt	The minimum weight contribution that is needed (this is
      *			just a hint which subclasses may ignore).
      */
-    virtual bool check(Xapian::docid did, Xapian::weight min_wt);
+    virtual bool check(Xapian::docid did, double min_wt);
 
     /** Return true if the current position is past the last entry in this list.
      *
@@ -254,7 +262,7 @@ class XAPIAN_VISIBILITY_DEFAULT PostingSource {
      *  (for example when wrapping the Xapian API for use from another
      *  language) then you can define a static <code>operator delete</code>
      *  method in your subclass as shown here:
-     *  http://trac.xapian.org/ticket/554#comment:1
+     *  https://trac.xapian.org/ticket/554#comment:1
      */
     virtual PostingSource * clone() const;
 
@@ -294,14 +302,34 @@ class XAPIAN_VISIBILITY_DEFAULT PostingSource {
      *  (for example when wrapping the Xapian API for use from another
      *  language) then you can define a static <code>operator delete</code>
      *  method in your subclass as shown here:
-     *  http://trac.xapian.org/ticket/554#comment:1
+     *  https://trac.xapian.org/ticket/554#comment:1
      *
      *  If you don't want to support the remote backend, you can use the
      *  default implementation which simply throws Xapian::UnimplementedError.
      *
-     *  @param s A serialised instance of this PostingSource subclass.
+     *  @param serialised A serialised instance of this PostingSource subclass.
      */
-    virtual PostingSource * unserialise(const std::string &s) const;
+    virtual PostingSource * unserialise(const std::string &serialised) const;
+
+    /** Create object given string serialisation returned by serialise().
+     *
+     *  Note that the returned object will be deallocated by Xapian after use
+     *  with "delete".  If you want to handle the deletion in a special way
+     *  (for example when wrapping the Xapian API for use from another
+     *  language) then you can define a static <code>operator delete</code>
+     *  method in your subclass as shown here:
+     *  https://trac.xapian.org/ticket/554#comment:1
+     *
+     *  This method is supplied with a Registry object, which can be used when
+     *  unserialising objects contained within the posting source.  The default
+     *  implementation simply calls unserialise() which doesn't take the
+     *  Registry object, so you do not need to implement this method unless you
+     *  want to take advantage of the Registry object when unserialising.
+     *
+     *  @param serialised A serialised instance of this PostingSource subclass.
+     */
+    virtual PostingSource * unserialise_with_registry(const std::string &serialised,
+				      const Registry & registry) const;
 
     /** Set this PostingSource to the start of the list of postings.
      *
@@ -335,6 +363,16 @@ class XAPIAN_VISIBILITY_DEFAULT PostingSource {
      *  get_description() gives for their subclass).
      */
     virtual std::string get_description() const;
+
+    PostingSource * release() {
+	opt_intrusive_base::release();
+	return this;
+    }
+
+    const PostingSource * release() const {
+	opt_intrusive_base::release();
+	return this;
+    }
 };
 
 
@@ -349,39 +387,80 @@ class XAPIAN_VISIBILITY_DEFAULT PostingSource {
  *  ValuePostingSource::init() if they know a tighter bound on the weight.
  */
 class XAPIAN_VISIBILITY_DEFAULT ValuePostingSource : public PostingSource {
+    // We want to give a deprecation warning for uses of the members from user
+    // code, but we also want to be able to inline functions to access them,
+    // without those functions generating deprecated warnings.  To achieve
+    // this, we make the old names references to members with a "real_" prefix
+    // and then use the latter in the inlined accessor functions.  The
+    // constructor initialises all the references to point to their "real_"
+    // counterparts.
+    Xapian::Database real_db;
+
+    Xapian::valueno real_slot;
+
+    Xapian::ValueIterator real_value_it;
+
+    bool real_started;
+
+    Xapian::doccount real_termfreq_min;
+
+    Xapian::doccount real_termfreq_est;
+
+    Xapian::doccount real_termfreq_max;
+
   protected:
-    /// The database we're reading values from.
-    Xapian::Database db;
+    /** The database we're reading values from.
+     *
+     *  @deprecated Use @a get_database() in preference.
+     */
+    XAPIAN_DEPRECATED(Xapian::Database& db);
 
-    /// The slot we're reading values from.
-    Xapian::valueno slot;
+    /** The slot we're reading values from.
+     *
+     *  @deprecated Use @a get_slot() in preference.
+     */
+    XAPIAN_DEPRECATED(Xapian::valueno& slot);
 
-    /// Value stream iterator.
-    Xapian::ValueIterator value_it;
+    /** Value stream iterator.
+     *
+     *  @deprecated Use @a get_value() in preference to *value_it, and other
+     *  methods of ValuePostingSource in preference to calling methods of
+     *  value_it.
+     */
+    XAPIAN_DEPRECATED(Xapian::ValueIterator& value_it);
 
-    /// Flag indicating if we've started (true if we have).
-    bool started;
+    /** Flag indicating if we've started (true if we have).
+     *
+     *  @deprecated Use @a get_started() in preference.
+     */
+    XAPIAN_DEPRECATED(bool& started);
 
     /** A lower bound on the term frequency.
      *
      *  Subclasses should set this if they are overriding the next(), skip_to()
      *  or check() methods to return fewer documents.
+     *
+     *  @deprecated Use @a set_termfreq_min() in preference.
      */
-    Xapian::doccount termfreq_min;
+    XAPIAN_DEPRECATED(Xapian::doccount& termfreq_min);
 
     /** An estimate of the term frequency.
      *
      *  Subclasses should set this if they are overriding the next(), skip_to()
      *  or check() methods.
+     *
+     *  @deprecated Use @a set_termfreq_est() in preference.
      */
-    Xapian::doccount termfreq_est;
+    XAPIAN_DEPRECATED(Xapian::doccount& termfreq_est);
 
     /** An upper bound on the term frequency.
      *
      *  Subclasses should set this if they are overriding the next(), skip_to()
      *  or check() methods.
+     *
+     *  @deprecated Use @a set_termfreq_max() in preference.
      */
-    Xapian::doccount termfreq_max;
+    XAPIAN_DEPRECATED(Xapian::doccount& termfreq_max);
 
   public:
     /** Construct a ValuePostingSource.
@@ -394,15 +473,83 @@ class XAPIAN_VISIBILITY_DEFAULT ValuePostingSource : public PostingSource {
     Xapian::doccount get_termfreq_est() const;
     Xapian::doccount get_termfreq_max() const;
 
-    void next(Xapian::weight min_wt);
-    void skip_to(Xapian::docid min_docid, Xapian::weight min_wt);
-    bool check(Xapian::docid min_docid, Xapian::weight min_wt);
+    void next(double min_wt);
+    void skip_to(Xapian::docid min_docid, double min_wt);
+    bool check(Xapian::docid min_docid, double min_wt);
 
     bool at_end() const;
 
     Xapian::docid get_docid() const;
 
     void init(const Database & db_);
+
+    /** The database we're reading values from.
+     *
+     *  Added in 1.2.23 and 1.3.5.
+     */
+    Xapian::Database get_database() const { return real_db; }
+
+    /** The slot we're reading values from.
+     *
+     *  Added in 1.2.23 and 1.3.5.
+     */
+    Xapian::valueno get_slot() const { return real_slot; }
+
+    /** Read current value.
+     *
+     *  Added in 1.2.23 and 1.3.5.
+     */
+    std::string get_value() const { return *real_value_it; }
+
+    /** End the iteration.
+     *
+     *  Calls to at_end() will return true after calling this method.
+     *
+     *  Added in 1.2.23 and 1.3.5.
+     */
+    void done() {
+	real_value_it = real_db.valuestream_end(real_slot);
+	real_started = true;
+    }
+
+    /** Flag indicating if we've started (true if we have).
+     *
+     *  Added in 1.2.23 and 1.3.5.
+     */
+    bool get_started() const { return real_started; }
+
+    /** Set a lower bound on the term frequency.
+     *
+     *  Subclasses should set this if they are overriding the next(), skip_to()
+     *  or check() methods to return fewer documents.
+     *
+     *  Added in 1.2.23 and 1.3.5.
+     */
+    void set_termfreq_min(Xapian::doccount termfreq_min_) {
+	real_termfreq_min = termfreq_min_;
+    }
+
+    /** An estimate of the term frequency.
+     *
+     *  Subclasses should set this if they are overriding the next(), skip_to()
+     *  or check() methods.
+     *
+     *  Added in 1.2.23 and 1.3.5.
+     */
+    void set_termfreq_est(Xapian::doccount termfreq_est_) {
+	real_termfreq_est = termfreq_est_;
+    }
+
+    /** An upper bound on the term frequency.
+     *
+     *  Subclasses should set this if they are overriding the next(), skip_to()
+     *  or check() methods.
+     *
+     *  Added in 1.2.23 and 1.3.5.
+     */
+    void set_termfreq_max(Xapian::doccount termfreq_max_) {
+	real_termfreq_max = termfreq_max_;
+    }
 };
 
 
@@ -433,11 +580,11 @@ class XAPIAN_VISIBILITY_DEFAULT ValueWeightPostingSource
      */
     explicit ValueWeightPostingSource(Xapian::valueno slot_);
 
-    Xapian::weight get_weight() const;
+    double get_weight() const;
     ValueWeightPostingSource * clone() const;
     std::string name() const;
     std::string serialise() const;
-    ValueWeightPostingSource * unserialise(const std::string &s) const;
+    ValueWeightPostingSource * unserialise(const std::string &serialised) const;
     void init(const Database & db_);
 
     std::string get_description() const;
@@ -474,23 +621,23 @@ class XAPIAN_VISIBILITY_DEFAULT DecreasingValueWeightPostingSource
     bool items_at_end;
 
     /// Skip the iterator forward if in the decreasing range, and weight is low.
-    void skip_if_in_range(Xapian::weight min_wt);
+    void skip_if_in_range(double min_wt);
 
   public:
     DecreasingValueWeightPostingSource(Xapian::valueno slot_,
 				       Xapian::docid range_start_ = 0,
 				       Xapian::docid range_end_ = 0);
 
-    Xapian::weight get_weight() const;
+    double get_weight() const;
     DecreasingValueWeightPostingSource * clone() const;
     std::string name() const;
     std::string serialise() const;
-    DecreasingValueWeightPostingSource * unserialise(const std::string &s) const;
+    DecreasingValueWeightPostingSource * unserialise(const std::string &serialised) const;
     void init(const Xapian::Database & db_);
 
-    void next(Xapian::weight min_wt);
-    void skip_to(Xapian::docid min_docid, Xapian::weight min_wt);
-    bool check(Xapian::docid min_docid, Xapian::weight min_wt);
+    void next(double min_wt);
+    void skip_to(Xapian::docid min_docid, double min_wt);
+    bool check(Xapian::docid min_docid, double min_wt);
 
     std::string get_description() const;
 };
@@ -538,11 +685,11 @@ class XAPIAN_VISIBILITY_DEFAULT ValueMapPostingSource
      */
     void set_default_weight(double wt);
 
-    Xapian::weight get_weight() const;
+    double get_weight() const;
     ValueMapPostingSource * clone() const;
     std::string name() const;
     std::string serialise() const;
-    ValueMapPostingSource * unserialise(const std::string &s) const;
+    ValueMapPostingSource * unserialise(const std::string &serialised) const;
     void init(const Database & db_);
 
     std::string get_description() const;
@@ -575,17 +722,17 @@ class XAPIAN_VISIBILITY_DEFAULT FixedWeightPostingSource : public PostingSource
      *
      *  @param wt The fixed weight to return.
      */
-    explicit FixedWeightPostingSource(Xapian::weight wt);
+    explicit FixedWeightPostingSource(double wt);
 
     Xapian::doccount get_termfreq_min() const;
     Xapian::doccount get_termfreq_est() const;
     Xapian::doccount get_termfreq_max() const;
 
-    Xapian::weight get_weight() const;
+    double get_weight() const;
 
-    void next(Xapian::weight min_wt);
-    void skip_to(Xapian::docid min_docid, Xapian::weight min_wt);
-    bool check(Xapian::docid min_docid, Xapian::weight min_wt);
+    void next(double min_wt);
+    void skip_to(Xapian::docid min_docid, double min_wt);
+    bool check(Xapian::docid min_docid, double min_wt);
 
     bool at_end() const;
 
@@ -594,7 +741,7 @@ class XAPIAN_VISIBILITY_DEFAULT FixedWeightPostingSource : public PostingSource
     FixedWeightPostingSource * clone() const;
     std::string name() const;
     std::string serialise() const;
-    FixedWeightPostingSource * unserialise(const std::string &s) const;
+    FixedWeightPostingSource * unserialise(const std::string &serialised) const;
     void init(const Database & db_);
 
     std::string get_description() const;
