@@ -1,7 +1,7 @@
 /** @file localsubmatch.h
  *  @brief SubMatch class for a local database.
  */
-/* Copyright (C) 2006,2007,2009,2010 Olly Betts
+/* Copyright (C) 2006,2007,2009,2010,2011,2013,2014,2015 Olly Betts
  * Copyright (C) 2007 Lemur Consulting Ltd
  *
  * This program is free software; you can redistribute it and/or modify
@@ -22,15 +22,18 @@
 #ifndef XAPIAN_INCLUDED_LOCALSUBMATCH_H
 #define XAPIAN_INCLUDED_LOCALSUBMATCH_H
 
-#include "database.h"
+#include "backends/database.h"
 #include "debuglog.h"
-#include "omqueryinternal.h"
+#include "api/leafpostlist.h"
+#include "api/queryinternal.h"
 #include "submatch.h"
 #include "xapian/enquire.h"
 #include "xapian/weight.h"
 
 #include <map>
 
+class LeafPostList;
+
 class LocalSubMatch : public SubMatch {
     /// Don't allow assignment.
     void operator=(const LocalSubMatch &);
@@ -39,10 +42,10 @@ class LocalSubMatch : public SubMatch {
     LocalSubMatch(const LocalSubMatch &);
 
     /// The statistics for the collection.
-    const Xapian::Weight::Internal * stats;
+    Xapian::Weight::Internal * stats;
 
     /// The original query before any rearrangement.
-    const Xapian::Query::Internal * query;
+    Xapian::Query query;
 
     /// The query length (used by some weighting schemes).
     Xapian::termcount qlen;
@@ -59,19 +62,15 @@ class LocalSubMatch : public SubMatch {
     /// Weight object (used as a factory by calling create on it).
     const Xapian::Weight * wt_factory;
 
-    /// The termfreqs and weights of terms used in orig_query, or NULL.
-    std::map<std::string,
-	     Xapian::MSet::Internal::TermFreqAndWeight> * term_info;
-
   public:
     /// Constructor.
     LocalSubMatch(const Xapian::Database::Internal *db_,
-		  const Xapian::Query::Internal * query_,
+		  const Xapian::Query & query_,
 		  Xapian::termcount qlen_,
 		  const Xapian::RSet & rset_,
 		  const Xapian::Weight *wt_factory_)
 	: stats(NULL), query(query_), qlen(qlen_), db(db_), rset(rset_),
-	  wt_factory(wt_factory_), term_info(NULL)
+	  wt_factory(wt_factory_)
     {
 	LOGCALL_CTOR(MATCH, "LocalSubMatch", db_ | query_ | qlen_ | rset_ | wt_factory_);
     }
@@ -83,25 +82,23 @@ class LocalSubMatch : public SubMatch {
     void start_match(Xapian::doccount first,
 		     Xapian::doccount maxitems,
 		     Xapian::doccount check_at_least,
-		     const Xapian::Weight::Internal & total_stats);
+		     Xapian::Weight::Internal & total_stats);
 
-    /// Get PostList and term info.
-    PostList * get_postlist_and_term_info(MultiMatch *matcher,
-	std::map<std::string,
-		 Xapian::MSet::Internal::TermFreqAndWeight> *termfreqandwts,
-	Xapian::termcount * total_subqs_ptr);
+    /// Get PostList.
+    PostList * get_postlist(MultiMatch *matcher,
+			    Xapian::termcount * total_subqs_ptr);
 
     /** Convert a postlist into a synonym postlist.
      */
     PostList * make_synonym_postlist(PostList * or_pl, MultiMatch * matcher,
 				     double factor);
 
-    /** Convert an OP_LEAF query to a PostList.
-     *
-     *  This is called by QueryOptimiser when it reaches an OP_LEAF query.
-     */
-    PostList * postlist_from_op_leaf_query(const Xapian::Query::Internal *query,
-					   double factor);
+    LeafPostList * open_post_list(const std::string& term,
+				  Xapian::termcount wqf,
+				  double factor,
+				  bool need_positions,
+				  QueryOptimiser * qopt,
+				  bool lazy_weight);
 };
 
 #endif /* XAPIAN_INCLUDED_LOCALSUBMATCH_H */
