@@ -1,7 +1,7 @@
 /** @file multiandpostlist.cc
  * @brief N-way AND postlist
  */
-/* Copyright (C) 2007,2009,2012,2015 Olly Betts
+/* Copyright (C) 2007,2009,2011,2012,2015 Olly Betts
  * Copyright (C) 2009 Lemur Consulting Ltd
  *
  * This program is free software; you can redistribute it and/or
@@ -25,12 +25,14 @@
 #include "omassert.h"
 #include "debuglog.h"
 
+using namespace std;
+
 void
 MultiAndPostList::allocate_plist_and_max_wt()
 {
     plist = new PostList * [n_kids];
     try {
-	max_wt = new Xapian::weight [n_kids];
+	max_wt = new double [n_kids];
     } catch (...) {
 	delete [] plist;
 	plist = NULL;
@@ -89,8 +91,9 @@ MultiAndPostList::get_termfreq_max() const
 Xapian::doccount
 MultiAndPostList::get_termfreq_est() const
 {
+    LOGCALL(MATCH, Xapian::doccount, "MultiAndPostList::get_termfreq_est", NO_ARGS);
     if (rare(db_size == 0))
-	return 0;
+	RETURN(0);
     // We calculate the estimate assuming independence.  With this assumption,
     // the estimate is the product of the estimates for the sub-postlists
     // divided by db_size (n_kids - 1) times.
@@ -103,7 +106,7 @@ MultiAndPostList::get_termfreq_est() const
 
 TermFreqs
 MultiAndPostList::get_termfreq_est_using_stats(
-	const Xapian::Weight::Internal & stats) const 
+	const Xapian::Weight::Internal & stats) const
 {
     LOGCALL(MATCH, TermFreqs, "MultiAndPostList::get_termfreq_est_using_stats", stats);
     // We calculate the estimate assuming independence.  With this assumption,
@@ -113,6 +116,7 @@ MultiAndPostList::get_termfreq_est_using_stats(
 
     double freqest = double(freqs.termfreq);
     double relfreqest = double(freqs.reltermfreq);
+    double collfreqest = double(freqs.collfreq);
 
     // Our caller should have ensured this.
     Assert(stats.collection_size);
@@ -123,6 +127,7 @@ MultiAndPostList::get_termfreq_est_using_stats(
 	// If the collection is empty, freqest should be 0 already, so leave
 	// it alone.
 	freqest = (freqest * freqs.termfreq) / stats.collection_size;
+	collfreqest = (collfreqest * freqs.collfreq) / stats.total_term_count;
 
 	// If the rset is empty, relfreqest should be 0 already, so leave
 	// it alone.
@@ -131,10 +136,11 @@ MultiAndPostList::get_termfreq_est_using_stats(
     }
 
     RETURN(TermFreqs(static_cast<Xapian::doccount>(freqest + 0.5),
-		     static_cast<Xapian::doccount>(relfreqest + 0.5)));
+		     static_cast<Xapian::doccount>(relfreqest + 0.5),
+		     static_cast<Xapian::termcount>(collfreqest + 0.5)));
 }
 
-Xapian::weight
+double
 MultiAndPostList::get_maxweight() const
 {
     return max_total;
@@ -157,11 +163,22 @@ MultiAndPostList::get_doclength() const
     return doclength;
 }
 
-Xapian::weight
+Xapian::termcount
+MultiAndPostList::get_unique_terms() const
+{
+    Assert(did);
+    Xapian::termcount unique_terms = plist[0]->get_unique_terms();
+    for (size_t i = 1; i < n_kids; ++i) {
+	AssertEq(unique_terms, plist[i]->get_unique_terms());
+    }
+    return unique_terms;
+}
+
+double
 MultiAndPostList::get_weight() const
 {
     Assert(did);
-    Xapian::weight result = 0;
+    double result = 0;
     for (size_t i = 0; i < n_kids; ++i) {
 	result += plist[i]->get_weight();
     }
@@ -174,12 +191,12 @@ MultiAndPostList::at_end() const
     return (did == 0);
 }
 
-Xapian::weight
+double
 MultiAndPostList::recalc_maxweight()
 {
     max_total = 0.0;
     for (size_t i = 0; i < n_kids; ++i) {
-	Xapian::weight new_max = plist[i]->recalc_maxweight();
+	double new_max = plist[i]->recalc_maxweight();
 	max_wt[i] = new_max;
 	max_total += new_max;
     }
@@ -187,7 +204,7 @@ MultiAndPostList::recalc_maxweight()
 }
 
 PostList *
-MultiAndPostList::find_next_match(Xapian::weight w_min)
+MultiAndPostList::find_next_match(double w_min)
 {
 advanced_plist0:
     if (plist[0]->at_end()) {
@@ -216,14 +233,14 @@ advanced_plist0:
 }
 
 PostList *
-MultiAndPostList::next(Xapian::weight w_min)
+MultiAndPostList::next(double w_min)
 {
     next_helper(0, w_min);
     return find_next_match(w_min);
 }
 
 PostList *
-MultiAndPostList::skip_to(Xapian::docid did_min, Xapian::weight w_min)
+MultiAndPostList::skip_to(Xapian::docid did_min, double w_min)
 {
     skip_to_helper(0, did_min, w_min);
     return find_next_match(w_min);
