@@ -2,7 +2,7 @@
  *
  * Copyright 1999,2000,2001 BrightStation PLC
  * Copyright 2001,2002 Ananova Ltd
- * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010 Olly Betts
+ * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010,2011,2013,2014,2015,2016 Olly Betts
  * Copyright 2003 Orange PCS Ltd
  * Copyright 2003 Sam Liddicott
  * Copyright 2007,2008,2009 Lemur Consulting Ltd
@@ -33,27 +33,26 @@
 #include "submatch.h"
 #include "localsubmatch.h"
 #include "omassert.h"
-#include "omenquireinternal.h"
+#include "api/omenquireinternal.h"
+#include "realtime.h"
 
-#include "emptypostlist.h"
+#include "api/emptypostlist.h"
 #include "branchpostlist.h"
 #include "mergepostlist.h"
 
-#include "document.h"
-#include "omqueryinternal.h"
+#include "backends/document.h"
 
 #include "msetcmp.h"
 
 #include "valuestreamdocument.h"
-#include "weightinternal.h"
+#include "weight/weightinternal.h"
 
-#include <xapian/errorhandler.h>
 #include <xapian/matchspy.h>
 #include <xapian/version.h> // For XAPIAN_HAS_REMOTE_BACKEND
 
 #ifdef XAPIAN_HAS_REMOTE_BACKEND
 #include "remotesubmatch.h"
-#include "remote-database.h"
+#include "backends/remote/remote-database.h"
 #endif /* XAPIAN_HAS_REMOTE_BACKEND */
 
 #include <algorithm>
@@ -63,7 +62,93 @@
 #include <map>
 #include <set>
 
+#ifdef HAVE_TIMER_CREATE
+#include <signal.h>
+#include <time.h>
+#include "safeunistd.h" // For _POSIX_* feature test macros.
+
+extern "C" {
+
+static void
+set_timeout_flag(union sigval sv)
+{
+    *(reinterpret_cast<volatile bool*>(sv.sival_ptr)) = true;
+}
+
+}
+
+// The monotonic clock is the better basis for timeouts, but not always
+// available.
+
+#ifndef _POSIX_MONOTONIC_CLOCK
+const clockid_t TIMEOUT_CLOCK = CLOCK_REALTIME;
+#elif defined __sun
+// Solaris defines CLOCK_MONOTONIC, but "man timer_create" doesn't mention it
+// and when using it timer_create() fails with "EPERM" (perhaps you need to be
+// root to use it?  I can't test that).
+//
+// Solaris defines _POSIX_MONOTONIC_CLOCK so we need to special case.
+const clockid_t TIMEOUT_CLOCK = CLOCK_REALTIME;
+#elif defined __CYGWIN__
+// https://cygwin.com/cygwin-api/std-notes.html currently (2016-05-13) says:
+//
+//     clock_nanosleep currently supports only CLOCK_REALTIME and
+//     CLOCK_MONOTONIC.  clock_setres, clock_settime, and timer_create
+//     currently support only CLOCK_REALTIME.
+//
+// So CLOCK_MONOTONIC is defined, but not supported by timer_create().
+const clockid_t TIMEOUT_CLOCK = CLOCK_REALTIME;
+#else
+const clockid_t TIMEOUT_CLOCK = CLOCK_MONOTONIC;
+#endif
+
+class TimeOut {
+    struct sigevent sev;
+    timer_t timerid;
+    volatile bool expired;
+
+  public:
+    TimeOut(double limit) : expired(false) {
+	if (limit > 0) {
+	    sev.sigev_notify = SIGEV_THREAD;
+	    sev.sigev_notify_function = set_timeout_flag;
+	    sev.sigev_notify_attributes = NULL;
+	    sev.sigev_value.sival_ptr =
+		static_cast<void*>(const_cast<bool*>(&expired));
+	    if (usual(timer_create(TIMEOUT_CLOCK, &sev, &timerid) == 0)) {
+		struct itimerspec interval;
+		interval.it_interval.tv_sec = 0;
+		interval.it_interval.tv_nsec = 0;
+		RealTime::to_timespec(limit, &interval.it_value);
+		if (usual(timer_settime(timerid, 0, &interval, NULL) == 0)) {
+		    // Timeout successfully set.
+		    return;
+		}
+		timer_delete(timerid);
+	    }
+	}
+	sev.sigev_notify = SIGEV_NONE;
+    }
+
+    ~TimeOut() {
+	if (sev.sigev_notify != SIGEV_NONE) {
+	    timer_delete(timerid);
+	    sev.sigev_notify = SIGEV_NONE;
+	}
+    }
+
+    bool timed_out() const { return expired; }
+};
+#else
+class TimeOut {
+  public:
+    TimeOut(double) { }
+    bool timed_out() const { return false; }
+};
+#endif
+
 using namespace std;
+using Xapian::Internal::intrusive_ptr;
 
 const Xapian::Enquire::Internal::sort_setting REL =
 	Xapian::Enquire::Internal::REL;
@@ -71,7 +156,7 @@ const Xapian::Enquire::Internal::sort_setting REL_VAL =
 	Xapian::Enquire::Internal::REL_VAL;
 const Xapian::Enquire::Internal::sort_setting VAL =
 	Xapian::Enquire::Internal::VAL;
-#if 0 // VAL_REL isn't currently used which causes a warning with SGI CC.
+#if 0 // VAL_REL isn't currently used so avoid compiler warnings.
 const Xapian::Enquire::Internal::sort_setting VAL_REL =
 	Xapian::Enquire::Internal::VAL_REL;
 #endif
@@ -89,7 +174,7 @@ split_rset_by_db(const Xapian::RSet * rset,
 		 vector<Xapian::RSet> & subrsets)
 {
     LOGCALL_STATIC_VOID(MATCH, "split_rset_by_db", rset | number_of_subdbs | subrsets);
-    if (rset) {
+    if (rset && !rset->empty()) {
 	if (number_of_subdbs == 1) {
 	    // The common case of a single database is easy to handle.
 	    subrsets.push_back(*rset);
@@ -135,11 +220,10 @@ split_rset_by_db(const Xapian::RSet * rset,
  *  statistics arrive, we can move on to the next step.
  */
 static void
-prepare_sub_matches(vector<Xapian::Internal::RefCntPtr<SubMatch> > & leaves,
-		    Xapian::ErrorHandler * errorhandler,
+prepare_sub_matches(vector<intrusive_ptr<SubMatch> > & leaves,
 		    Xapian::Weight::Internal & stats)
 {
-    LOGCALL_STATIC_VOID(MATCH, "prepare_sub_matches", leaves | errorhandler | stats);
+    LOGCALL_STATIC_VOID(MATCH, "prepare_sub_matches", leaves | stats);
     // We use a vector<bool> to track which SubMatches we're already prepared.
     vector<bool> prepared;
     prepared.resize(leaves.size(), false);
@@ -148,19 +232,8 @@ prepare_sub_matches(vector<Xapian::Internal::RefCntPtr<SubMatch> > & leaves,
     while (unprepared) {
 	for (size_t leaf = 0; leaf < leaves.size(); ++leaf) {
 	    if (prepared[leaf]) continue;
-	    try {
-		SubMatch * submatch = leaves[leaf].get();
-		if (!submatch || submatch->prepare_match(nowait, stats)) {
-		    prepared[leaf] = true;
-		    --unprepared;
-		}
-	    } catch (Xapian::Error & e) {
-		if (!errorhandler) throw;
-
-		LOGLINE(EXCEPTION, "Calling error handler for prepare_match() on a SubMatch.");
-		(*errorhandler)(e);
-		// Continue match without this sub-match.
-		leaves[leaf] = NULL;
+	    SubMatch * submatch = leaves[leaf].get();
+	    if (!submatch || submatch->prepare_match(nowait, stats)) {
 		prepared[leaf] = true;
 		--unprepared;
 	    }
@@ -175,25 +248,24 @@ prepare_sub_matches(vector<Xapian::Internal::RefCntPtr<SubMatch> > & leaves,
 class MultipleMatchSpy : public Xapian::MatchSpy {
   private:
     /// List of match spies to call, in order.
-    const std::vector<Xapian::MatchSpy *> & spies;
+    const std::vector<Xapian::Internal::opt_intrusive_ptr<Xapian::MatchSpy>> & spies;
 
   public:
-    MultipleMatchSpy(const std::vector<Xapian::MatchSpy *> & spies_)
+    MultipleMatchSpy(const std::vector<Xapian::Internal::opt_intrusive_ptr<Xapian::MatchSpy>> & spies_)
 	    : spies(spies_) {}
 
     /** Implementation of virtual operator().
      *
      *  This implementation calls all the spies in turn.
      */
-    void operator()(const Xapian::Document &doc, Xapian::weight wt);
+    void operator()(const Xapian::Document &doc, double wt);
 };
 
 void 
-MultipleMatchSpy::operator()(const Xapian::Document &doc, Xapian::weight wt) {
+MultipleMatchSpy::operator()(const Xapian::Document &doc, double wt) {
     LOGCALL_VOID(MATCH, "MultipleMatchSpy::operator()", doc | wt);
-    vector<Xapian::MatchSpy *>::const_iterator i;
-    for (i = spies.begin(); i != spies.end(); ++i) {
-	(**i)(doc, wt);
+    for (auto i : spies) {
+	(*i)(doc, wt);
     }
 }
 
@@ -201,20 +273,20 @@ MultipleMatchSpy::operator()(const Xapian::Document &doc, Xapian::weight wt) {
 // Initialisation and cleaning up //
 ////////////////////////////////////
 MultiMatch::MultiMatch(const Xapian::Database &db_,
-		       const Xapian::Query::Internal * query_,
+		       const Xapian::Query & query_,
 		       Xapian::termcount qlen,
 		       const Xapian::RSet * omrset,
 		       Xapian::doccount collapse_max_,
 		       Xapian::valueno collapse_key_,
-		       int percent_cutoff_, Xapian::weight weight_cutoff_,
+		       int percent_cutoff_, double weight_cutoff_,
 		       Xapian::Enquire::docid_order order_,
 		       Xapian::valueno sort_key_,
 		       Xapian::Enquire::Internal::sort_setting sort_by_,
 		       bool sort_value_forward_,
-		       Xapian::ErrorHandler * errorhandler_,
+		       double time_limit_,
 		       Xapian::Weight::Internal & stats,
 		       const Xapian::Weight * weight_,
-		       const vector<Xapian::MatchSpy *> & matchspies_,
+		       const vector<Xapian::Internal::opt_intrusive_ptr<Xapian::MatchSpy>> & matchspies_,
 		       bool have_sorter, bool have_mdecider)
 	: db(db_), query(query_),
 	  collapse_max(collapse_max_), collapse_key(collapse_key_),
@@ -222,14 +294,14 @@ MultiMatch::MultiMatch(const Xapian::Database &db_,
 	  order(order_),
 	  sort_key(sort_key_), sort_by(sort_by_),
 	  sort_value_forward(sort_value_forward_),
-	  errorhandler(errorhandler_), weight(weight_),
+	  time_limit(time_limit_),
+	  weight(weight_),
 	  is_remote(db.internal.size()),
 	  matchspies(matchspies_)
 {
-    LOGCALL_CTOR(MATCH, "MultiMatch", db_ | query_ | qlen | omrset | collapse_max_ | collapse_key_ | percent_cutoff_ | weight_cutoff_ | int(order_) | sort_key_ | int(sort_by_) | sort_value_forward_ | errorhandler_ | stats | weight_ | matchspies_ | have_sorter | have_mdecider);
+    LOGCALL_CTOR(MATCH, "MultiMatch", db_ | query_ | qlen | omrset | collapse_max_ | collapse_key_ | percent_cutoff_ | weight_cutoff_ | int(order_) | sort_key_ | int(sort_by_) | sort_value_forward_ | time_limit_| stats | weight_ | matchspies_ | have_sorter | have_mdecider);
 
-    if (!query) return;
-    query->validate_query();
+    if (query.empty()) return;
 
     Xapian::doccount number_of_subdbs = db.internal.size();
     vector<Xapian::RSet> subrsets;
@@ -238,55 +310,52 @@ MultiMatch::MultiMatch(const Xapian::Database &db_,
     for (size_t i = 0; i != number_of_subdbs; ++i) {
 	Xapian::Database::Internal *subdb = db.internal[i].get();
 	Assert(subdb);
-	Xapian::Internal::RefCntPtr<SubMatch> smatch;
-	try {
-	    // There is currently only one special case, for network databases.
+	intrusive_ptr<SubMatch> smatch;
+
+	// There is currently only one special case, for network databases.
 #ifdef XAPIAN_HAS_REMOTE_BACKEND
-	    RemoteDatabase *rem_db = subdb->as_remotedatabase();
-	    if (rem_db) {
-		if (have_sorter) {
-		    throw Xapian::UnimplementedError("Xapian::KeyMaker not supported for the remote backend");
-		}
-		if (have_mdecider) {
-		    throw Xapian::UnimplementedError("Xapian::MatchDecider not supported for the remote backend");
-		}
-		rem_db->set_query(query, qlen, collapse_max, collapse_key,
-				  order, sort_key, sort_by, sort_value_forward,
-				  percent_cutoff, weight_cutoff, weight,
-				  subrsets[i], matchspies);
-		bool decreasing_relevance =
-		    (sort_by == REL || sort_by == REL_VAL);
-		smatch = new RemoteSubMatch(rem_db, decreasing_relevance, matchspies);
-		is_remote[i] = true;
-	    } else {
-		smatch = new LocalSubMatch(subdb, query, qlen, subrsets[i], weight);
+	if (subdb->get_backend_info(NULL) == BACKEND_REMOTE) {
+	    RemoteDatabase *rem_db = static_cast<RemoteDatabase*>(subdb);
+	    if (have_sorter) {
+		throw Xapian::UnimplementedError("Xapian::KeyMaker not supported for the remote backend");
 	    }
-#else
-	    // Avoid unused parameter warnings.
-	    (void)have_sorter;
-	    (void)have_mdecider;
+	    if (have_mdecider) {
+		throw Xapian::UnimplementedError("Xapian::MatchDecider not supported for the remote backend");
+	    }
+	    // FIXME: Remote handling for time_limit with multiple
+	    // databases may need some work.
+	    rem_db->set_query(query, qlen, collapse_max, collapse_key,
+			      order, sort_key, sort_by, sort_value_forward,
+			      time_limit,
+			      percent_cutoff, weight_cutoff, weight,
+			      subrsets[i], matchspies);
+	    bool decreasing_relevance =
+		(sort_by == REL || sort_by == REL_VAL);
+	    smatch = new RemoteSubMatch(rem_db, decreasing_relevance, matchspies);
+	    is_remote[i] = true;
+	} else {
 	    smatch = new LocalSubMatch(subdb, query, qlen, subrsets[i], weight);
-#endif /* XAPIAN_HAS_REMOTE_BACKEND */
-	} catch (Xapian::Error & e) {
-	    if (!errorhandler) throw;
-	    LOGLINE(EXCEPTION, "Calling error handler for creation of a SubMatch from a database and query.");
-	    (*errorhandler)(e);
-	    // Continue match without this sub-postlist.
-	    smatch = NULL;
+	    subdb->readahead_for_query(query);
 	}
+#else
+	// Avoid unused parameter warnings.
+	(void)have_sorter;
+	(void)have_mdecider;
+	smatch = new LocalSubMatch(subdb, query, qlen, subrsets[i], weight);
+#endif /* XAPIAN_HAS_REMOTE_BACKEND */
 	leaves.push_back(smatch);
     }
 
-    stats.mark_wanted_terms(*query);
-    prepare_sub_matches(leaves, errorhandler, stats);
+    stats.set_query(query);
+    prepare_sub_matches(leaves, stats);
     stats.set_bounds_from_db(db);
 }
 
-Xapian::weight
+double
 MultiMatch::getorrecalc_maxweight(PostList *pl)
 {
-    LOGCALL(MATCH, Xapian::weight, "MultiMatch::getorrecalc_maxweight", pl);
-    Xapian::weight wt;
+    LOGCALL(MATCH, double, "MultiMatch::getorrecalc_maxweight", pl);
+    double wt;
     if (recalculate_w_max) {
 	LOGLINE(MATCH, "recalculating max weight");
 	wt = pl->recalc_maxweight();
@@ -304,22 +373,23 @@ void
 MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 		     Xapian::doccount check_at_least,
 		     Xapian::MSet & mset,
-		     const Xapian::Weight::Internal & stats,
+		     Xapian::Weight::Internal & stats,
 		     const Xapian::MatchDecider *mdecider,
-		     const Xapian::MatchDecider *matchspy_legacy,
 		     const Xapian::KeyMaker *sorter)
 {
-    LOGCALL_VOID(MATCH, "MultiMatch::get_mset", first | maxitems | check_at_least | Literal("mset") | stats | Literal("mdecider") | Literal("matchspy_legacy") | Literal("sorter"));
+    LOGCALL_VOID(MATCH, "MultiMatch::get_mset", first | maxitems | check_at_least | Literal("mset") | stats | Literal("mdecider") | Literal("sorter"));
     AssertRel(check_at_least,>=,maxitems);
 
-    if (!query) {
-	mset = Xapian::MSet(new Xapian::MSet::Internal());
+    if (query.empty()) {
+	mset = Xapian::MSet();
 	mset.internal->firstitem = first;
 	return;
     }
 
     Assert(!leaves.empty());
 
+    TimeOut timeout(time_limit);
+
 #ifdef XAPIAN_HAS_REMOTE_BACKEND
     // If there's only one database and it's remote, we can just unserialise
     // its MSet and return that.
@@ -333,30 +403,12 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 #endif
 
     // Start matchers.
-    {
-	vector<Xapian::Internal::RefCntPtr<SubMatch> >::iterator leaf;
-	for (leaf = leaves.begin(); leaf != leaves.end(); ++leaf) {
-	    if (!(*leaf).get()) continue;
-	    try {
-		(*leaf)->start_match(0, first + maxitems,
-				     first + check_at_least, stats);
-	    } catch (Xapian::Error & e) {
-		if (!errorhandler) throw;
-		LOGLINE(EXCEPTION, "Calling error handler for "
-				   "start_match() on a SubMatch.");
-		(*errorhandler)(e);
-		// Continue match without this sub-match.
-		*leaf = NULL;
-	    }
-	}
+    for (auto && leaf : leaves) {
+	leaf->start_match(0, first + maxitems, first + check_at_least, stats);
     }
 
     // Get postlists and term info
     vector<PostList *> postlists;
-    map<string, Xapian::MSet::Internal::TermFreqAndWeight> termfreqandwts;
-    map<string, Xapian::MSet::Internal::TermFreqAndWeight> * termfreqandwts_ptr;
-    termfreqandwts_ptr = &termfreqandwts;
-
     Xapian::termcount total_subqs = 0;
     // Keep a count of matches which we know exist, but we won't see.  This
     // occurs when a submatch is remote, and returns a lower bound on the
@@ -364,39 +416,23 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
     // documents it returns (because it wasn't asked for more documents).
     Xapian::doccount definite_matches_not_seen = 0;
     for (size_t i = 0; i != leaves.size(); ++i) {
-	PostList *pl;
-	try {
-	    pl = leaves[i]->get_postlist_and_term_info(this,
-						       termfreqandwts_ptr,
-						       &total_subqs);
-	    if (termfreqandwts_ptr && !termfreqandwts.empty())
-		termfreqandwts_ptr = NULL;
-	    if (is_remote[i]) {
-		if (pl->get_termfreq_min() > first + maxitems) {
-		    LOGLINE(MATCH, "Found " <<
-				   pl->get_termfreq_min() - (first + maxitems)
-				   << " definite matches in remote submatch "
-				   "which aren't passed to local match");
-		    definite_matches_not_seen += pl->get_termfreq_min();
-		    definite_matches_not_seen -= first + maxitems;
-		}
+	PostList * pl = leaves[i]->get_postlist(this, &total_subqs);
+	if (is_remote[i]) {
+	    if (pl->get_termfreq_min() > first + maxitems) {
+		LOGLINE(MATCH, "Found " <<
+			       pl->get_termfreq_min() - (first + maxitems)
+			       << " definite matches in remote submatch "
+			       "which aren't passed to local match");
+		definite_matches_not_seen += pl->get_termfreq_min();
+		definite_matches_not_seen -= first + maxitems;
 	    }
-	} catch (Xapian::Error & e) {
-	    if (!errorhandler) throw;
-	    LOGLINE(EXCEPTION, "Calling error handler for "
-			       "get_term_info() on a SubMatch.");
-	    (*errorhandler)(e);
-	    // FIXME: check if *ALL* the remote servers have failed!
-	    // Continue match without this sub-match.
-	    leaves[i] = NULL;
-	    pl = new EmptyPostList;
 	}
 	postlists.push_back(pl);
     }
     Assert(!postlists.empty());
 
     ValueStreamDocument vsdoc(db);
-    ++vsdoc.ref_count;
+    ++vsdoc._refs;
     Xapian::Document doc(&vsdoc);
 
     // Get a single combined postlist
@@ -404,23 +440,14 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
     if (postlists.size() == 1) {
 	pl.reset(postlists.front());
     } else {
-	pl.reset(new MergePostList(postlists, this, vsdoc, errorhandler));
+	pl.reset(new MergePostList(postlists, this, vsdoc));
     }
 
     LOGLINE(MATCH, "pl = (" << pl->get_description() << ")");
 
-#ifdef XAPIAN_DEBUG_LOG
-    {
-	map<string, Xapian::MSet::Internal::TermFreqAndWeight>::const_iterator tfwi;
-	for (tfwi = termfreqandwts.begin(); tfwi != termfreqandwts.end(); ++tfwi) {
-	    LOGLINE(MATCH, "termfreqandwts[" << tfwi->first << "] = " << tfwi->second.termfreq << ", " << tfwi->second.termweight);
-	}
-    }
-#endif
-
     // Empty result set
     Xapian::doccount docs_matched = 0;
-    Xapian::weight greatest_wt = 0;
+    double greatest_wt = 0;
     Xapian::termcount greatest_wt_subqs_matched = 0;
 #ifdef XAPIAN_HAS_REMOTE_BACKEND
     unsigned greatest_wt_subqs_db_num = UINT_MAX;
@@ -428,7 +455,7 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
     vector<Xapian::Internal::MSetItem> items;
 
     // maximum weight a document could possibly have
-    const Xapian::weight max_possible = pl->recalc_maxweight();
+    const double max_possible = pl->recalc_maxweight();
 
     LOGLINE(MATCH, "pl = (" << pl->get_description() << ")");
     recalculate_w_max = false;
@@ -437,8 +464,8 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
     Xapian::doccount matches_lower_bound = 0;
     Xapian::doccount matches_estimated   = pl->get_termfreq_est();
 
-    if (mdecider == NULL && matchspy_legacy == NULL) {
-	// If we have a matcher decider or match spy, the lower bound must be
+    if (mdecider == NULL) {
+	// If we have a match decider, the lower bound must be
 	// set to 0 as we could discard all hits.  Otherwise set it to the
 	// minimum number of entries which the postlist could return.
 	matches_lower_bound = pl->get_termfreq_min();
@@ -449,7 +476,7 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
     MultipleMatchSpy multispy(matchspies);
     if (!matchspies.empty()) {
 	if (matchspies.size() == 1) {
-	    matchspy = matchspies[0];
+	    matchspy = matchspies[0].get();
 	} else {
 	    matchspy = &multispy;
 	}
@@ -468,7 +495,7 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 		matches_lower_bound = collapse_max;
 	}
 
-	mset = Xapian::MSet(new Xapian::MSet::Internal(
+	mset.internal = new Xapian::MSet::Internal(
 					   first,
 					   matches_upper_bound,
 					   matches_lower_bound,
@@ -477,14 +504,13 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 					   uncollapsed_lower_bound,
 					   matches_estimated,
 					   max_possible, greatest_wt, items,
-					   termfreqandwts,
-					   0));
+					   0);
 	return;
     }
 
-    // Number of documents considered by a decider or matchspy_legacy.
+    // Number of documents considered by a decider.
     Xapian::doccount decider_considered = 0;
-    // Number of documents denied by the decider or matchspy_legacy.
+    // Number of documents denied by the decider.
     Xapian::doccount decider_denied = 0;
 
     // Set max number of results that we want - this is used to decide
@@ -497,10 +523,10 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
     Xapian::Internal::MSetItem min_item(0.0, 0);
 
     // Minimum weight an item must have to be worth considering.
-    Xapian::weight min_weight = weight_cutoff;
+    double min_weight = weight_cutoff;
 
     // Factor to multiply maximum weight seen by to get the cutoff weight.
-    Xapian::weight percent_cutoff_factor = percent_cutoff / 100.0;
+    double percent_cutoff_factor = percent_cutoff / 100.0;
     // Corresponding correction to that in omenquire.cc to account for excess
     // precision on x86.
     percent_cutoff_factor -= DBL_EPSILON;
@@ -564,7 +590,7 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 	// Only calculate the weight if we need it for mcmp, or there's a
 	// percentage or weight cutoff in effect.  Otherwise we calculate it
 	// below if we haven't already rejected this candidate.
-	Xapian::weight wt = 0.0;
+	double wt = 0.0;
 	bool calculated_weight = false;
 	if (sort_by != VAL || min_weight > 0.0) {
 	    wt = pl->get_weight();
@@ -579,6 +605,10 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 	vsdoc.set_document(did);
 	LOGLINE(MATCH, "Candidate document id " << did << " wt " << wt);
 	Xapian::Internal::MSetItem new_item(wt, did);
+	if (check_at_least > maxitems && timeout.timed_out()) {
+	    check_at_least = maxitems;
+	}
+
 	if (sort_by != REL) {
 	    const string * ptr = pl->get_sort_key();
 	    if (ptr) {
@@ -594,7 +624,7 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 	    // VAL, then new_item.wt won't yet be set, but that doesn't
 	    // matter since it's not used by the sort function.
 	    if (!mcmp(new_item, min_item)) {
-		if (mdecider == NULL && !collapser && matchspy_legacy == NULL) {
+		if (mdecider == NULL && !collapser) {
 		    // Document was definitely suitable for mset - no more
 		    // processing needed.
 		    LOGLINE(MATCH, "Making note of match item which sorts lower than min_item");
@@ -614,15 +644,15 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 		    if (wt > greatest_wt) goto new_greatest_weight;
 		    continue;
 		}
-		// We can't drop the item, because we need to show it
-		// to the matchspy_legacy, test whether the mdecider would
-		// accept it, and/or test whether it would be collapsed.
+		// We can't drop the item, because we need to test whether the
+		// mdecider would accept it and/or test whether it would be
+		// collapsed.
 		LOGLINE(MATCH, "Keeping candidate which sorts lower than min_item for further investigation");
 	    }
 	}
 
 	// Use the match spy and/or decision functors (if specified).
-	if (matchspy != NULL || mdecider != NULL || matchspy_legacy != NULL) {
+	if (matchspy != NULL || mdecider != NULL) {
 	    const unsigned int multiplier = db.internal.size();
 	    Assert(multiplier != 0);
 	    Xapian::doccount n = (did - 1) % multiplier; // which actual database
@@ -630,10 +660,6 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 	    // already have been applied there so we can skip this step.
 	    if (!is_remote[n]) {
 		++decider_considered;
-		if (matchspy_legacy && !matchspy_legacy->operator()(doc)) {
-		    ++decider_denied;
-		    continue;
-		}
 		if (mdecider && !mdecider->operator()(doc)) {
 		    ++decider_denied;
 		    continue;
@@ -682,7 +708,7 @@ MultiMatch::get_mset(Xapian::doccount first, Xapian::doccount maxitems,
 		// entry with this key which it displaced might still be in the
 		// proto-MSet.  If it might be, we need to check through for
 		// it.
-		Xapian::weight old_wt = old_item.wt;
+		double old_wt = old_item.wt;
 		if (old_wt >= min_weight && mcmp(old_item, min_item)) {
 		    // Scan through (unsorted) MSet looking for entry.
 		    // FIXME: more efficient way than just scanning?
@@ -792,7 +818,7 @@ new_greatest_weight:
 #endif
 	    }
 	    if (percent_cutoff) {
-		Xapian::weight w = wt * percent_cutoff_factor;
+		double w = wt * percent_cutoff_factor;
 		if (w > min_weight) {
 		    min_weight = w;
 		    if (!is_heap) {
@@ -840,7 +866,7 @@ new_greatest_weight:
 	    // Or we could just use a linear scan here instead.
 
 	    // trim the mset to the correct answer...
-	    Xapian::weight min_wt = percent_cutoff_factor / percent_scale;
+	    double min_wt = percent_cutoff_factor / percent_scale;
 	    if (!is_heap) {
 		is_heap = true;
 		make_heap<vector<Xapian::Internal::MSetItem>::iterator,
@@ -930,7 +956,7 @@ new_greatest_weight:
 		    ", matches_upper_bound=" << matches_upper_bound);
 	}
 
-	if (mdecider || matchspy_legacy) {
+	if (mdecider) {
 	    if (!percent_cutoff) {
 		if (!collapser) {
 		    // We're not collapsing or doing a percentage cutoff, so
@@ -994,7 +1020,7 @@ new_greatest_weight:
 	       	matches_estimated = matches_lower_bound;
 	}
 
-	if (collapser || mdecider || matchspy_legacy) {
+	if (collapser || mdecider) {
 	    LOGLINE(MATCH, "Clamping estimate between bounds: "
 		    "matches_lower_bound = " << matches_lower_bound <<
 		    ", matches_estimated = " << matches_estimated <<
@@ -1010,7 +1036,7 @@ new_greatest_weight:
 		matches_estimated = docs_matched;
 	}
 
-	if (collapser && !mdecider && !percent_cutoff && !matchspy_legacy) {
+	if (collapser && !mdecider && !percent_cutoff) {
 	    AssertRel(docs_matched,<=,uncollapsed_upper_bound);
 	    if (docs_matched > uncollapsed_lower_bound)
 		uncollapsed_lower_bound = docs_matched;
@@ -1046,7 +1072,7 @@ new_greatest_weight:
 	    vector<Xapian::Internal::MSetItem>::reverse_iterator nth;
 	    nth = items.rbegin() + first;
 	    nth_element(items.rbegin(), nth, items.rend(), mcmp);
-	    // Erase the trailing ``first'' elements
+	    // Erase the trailing "first" elements
 	    items.erase(items.begin() + items.size() - first, items.end());
 	}
     }
@@ -1077,7 +1103,7 @@ new_greatest_weight:
     // is_heap so much maybe?
     if (!items.empty() && collapser && !collapser.empty()) {
 	// Nicked this formula from above.
-	Xapian::weight min_wt = 0.0;
+	double min_wt = 0.0;
 	if (percent_scale > 0.0)
 	    min_wt = percent_cutoff_factor / percent_scale;
 	Xapian::doccount entries = collapser.entries();
@@ -1095,7 +1121,7 @@ new_greatest_weight:
 	}
     }
 
-    mset = Xapian::MSet(new Xapian::MSet::Internal(
+    mset.internal = new Xapian::MSet::Internal(
 				       first,
 				       matches_upper_bound,
 				       matches_lower_bound,
@@ -1104,6 +1130,5 @@ new_greatest_weight:
 				       uncollapsed_lower_bound,
 				       uncollapsed_estimated,
 				       max_possible, greatest_wt, items,
-				       termfreqandwts,
-				       percent_scale * 100.0));
+				       percent_scale * 100.0);
 }
