@@ -1,7 +1,7 @@
 /** @file remoteserver.cc
  *  @brief Xapian remote backend server base class
  */
-/* Copyright (C) 2006,2007,2008,2009,2010,2015 Olly Betts
+/* Copyright (C) 2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2016 Olly Betts
  * Copyright (C) 2006,2007,2009,2010 Lemur Consulting Ltd
  *
  * This program is free software; you can redistribute it and/or modify
@@ -22,10 +22,12 @@
 #include <config.h>
 #include "remoteserver.h"
 
+#include "xapian/constants.h"
 #include "xapian/database.h"
 #include "xapian/enquire.h"
 #include "xapian/error.h"
 #include "xapian/matchspy.h"
+#include "xapian/query.h"
 #include "xapian/valueiterator.h"
 
 #include "safeerrno.h"
@@ -33,14 +35,24 @@
 #include <cstdlib>
 
 #include "autoptr.h"
-#include "multimatch.h"
-#include "net/length.h"
+#include "length.h"
+#include "matcher/multimatch.h"
+#include "noreturn.h"
 #include "omassert.h"
 #include "realtime.h"
 #include "serialise.h"
 #include "serialise-double.h"
+#include "serialise-error.h"
 #include "str.h"
-#include "weightinternal.h"
+#include "stringutils.h"
+#include "weight/weightinternal.h"
+
+XAPIAN_NORETURN(static void throw_read_only());
+static void
+throw_read_only()
+{
+    throw Xapian::InvalidOperationError("Server is read-only");
+}
 
 /// Class to throw when we receive the connection closing message.
 struct ConnectionClosed { };
@@ -90,21 +102,7 @@ RemoteServer::RemoteServer(const std::vector<std::string> &dbpaths,
 #endif
 
     // Send greeting message.
-    string message;
-    message += char(XAPIAN_REMOTE_PROTOCOL_MAJOR_VERSION);
-    message += char(XAPIAN_REMOTE_PROTOCOL_MINOR_VERSION);
-    message += encode_length(db->get_doccount());
-    message += encode_length(db->get_lastdocid());
-    message += encode_length(db->get_doclength_lower_bound());
-    message += encode_length(db->get_doclength_upper_bound());
-    message += (db->has_positions() ? '1' : '0');
-    // FIXME: clumsy to reverse calculate total_len like this:
-    totlen_t total_len = totlen_t(db->get_avlength() * db->get_doccount() + .5);
-    message += encode_length(total_len);
-    //message += encode_length(db->get_total_length());
-    string uuid = db->get_uuid();
-    message += uuid;
-    send_message(REPLY_GREETING, message);
+    msg_update(string());
 }
 
 RemoteServer::~RemoteServer()
@@ -192,7 +190,8 @@ RemoteServer::run()
 		0, // MSG_GETMSET - used during a conversation.
 		0, // MSG_SHUTDOWN - handled by get_message().
 		&RemoteServer::msg_openmetadatakeylist,
-		&RemoteServer::msg_query_new,
+		&RemoteServer::msg_freqs,
+		&RemoteServer::msg_uniqueterms,
 	    };
 
 	    string message;
@@ -240,13 +239,21 @@ RemoteServer::run()
 void
 RemoteServer::msg_allterms(const string &message)
 {
-    const string & prefix = message;
+    string prev = message;
+    string reply;
 
+    const string & prefix = message;
     const Xapian::TermIterator end = db->allterms_end(prefix);
     for (Xapian::TermIterator t = db->allterms_begin(prefix); t != end; ++t) {
-	string item = encode_length(t.get_termfreq());
-	item += *t;
-	send_message(REPLY_ALLTERMS, item);
+	if (rare(prev.size() > 255))
+	    prev.resize(255);
+	const string & v = *t;
+	size_t reuse = common_prefix_length(prev, v);
+	reply = encode_length(t.get_termfreq());
+	reply.append(1, char(reuse));
+	reply.append(v, reuse, string::npos);
+	send_message(REPLY_ALLTERMS, reply);
+	prev = v;
     }
 
     send_message(REPLY_DONE, string());
@@ -261,12 +268,19 @@ RemoteServer::msg_termlist(const string &message)
     decode_length(&p, p_end, did);
 
     send_message(REPLY_DOCLENGTH, encode_length(db->get_doclength(did)));
+    string prev;
     const Xapian::TermIterator end = db->termlist_end(did);
     for (Xapian::TermIterator t = db->termlist_begin(did); t != end; ++t) {
-	string item = encode_length(t.get_wdf());
-	item += encode_length(t.get_termfreq());
-	item += *t;
-	send_message(REPLY_TERMLIST, item);
+	if (rare(prev.size() > 255))
+	    prev.resize(255);
+	const string & v = *t;
+	size_t reuse = common_prefix_length(prev, v);
+	string reply = encode_length(t.get_wdf());
+	reply += encode_length(t.get_termfreq());
+	reply.append(1, char(reuse));
+	reply.append(v, reuse, string::npos);
+	send_message(REPLY_TERMLIST, reply);
+	prev = v;
     }
 
     send_message(REPLY_DONE, string());
@@ -322,9 +336,21 @@ void
 RemoteServer::msg_writeaccess(const string & msg)
 {
     if (!writable) 
-	throw Xapian::InvalidOperationError("Server is read-only");
+	throw_read_only();
 
-    wdb = new Xapian::WritableDatabase(context, Xapian::DB_OPEN);
+    int flags = Xapian::DB_OPEN;
+    const char *p = msg.c_str();
+    const char *p_end = p + msg.size();
+    if (p != p_end) {
+	unsigned flag_bits;
+	decode_length(&p, p_end, flag_bits);
+	flags |= flag_bits &~ Xapian::DB_ACTION_MASK_;
+	if (p != p_end) {
+	    throw Xapian::NetworkError("Junk at end of MSG_WRITEACCESS");
+	}
+    }
+
+    wdb = new Xapian::WritableDatabase(context, flags);
     delete db;
     db = wdb;
     msg_update(msg);
@@ -333,17 +359,27 @@ RemoteServer::msg_writeaccess(const string & msg)
 void
 RemoteServer::msg_reopen(const string & msg)
 {
-    db->reopen();
+    if (!db->reopen()) {
+	send_message(REPLY_DONE, string());
+	return;
+    }
     msg_update(msg);
 }
 
 void
 RemoteServer::msg_update(const string &)
 {
-    string message = encode_length(db->get_doccount());
-    message += encode_length(db->get_lastdocid());
-    message += encode_length(db->get_doclength_lower_bound());
-    message += encode_length(db->get_doclength_upper_bound());
+    static const char protocol[2] = {
+	char(XAPIAN_REMOTE_PROTOCOL_MAJOR_VERSION),
+	char(XAPIAN_REMOTE_PROTOCOL_MINOR_VERSION)
+    };
+    string message(protocol, 2);
+    Xapian::doccount num_docs = db->get_doccount();
+    message += encode_length(num_docs);
+    message += encode_length(db->get_lastdocid() - num_docs);
+    Xapian::termcount doclen_lb = db->get_doclength_lower_bound();
+    message += encode_length(doclen_lb);
+    message += encode_length(db->get_doclength_upper_bound() - doclen_lb);
     message += (db->has_positions() ? '1' : '0');
     // FIXME: clumsy to reverse calculate total_len like this:
     totlen_t total_len = totlen_t(db->get_avlength() * db->get_doccount() + .5);
@@ -354,22 +390,6 @@ RemoteServer::msg_update(const string &)
     send_message(REPLY_UPDATE, message);
 }
 
-/** Structure holding a list of match spies.
- *
- *  The main reason for the existence of this structure is to make it easy to
- *  ensure that the match spies are all deleted after use.
- */
-struct MatchSpyList {
-    vector<Xapian::MatchSpy *> spies;
-
-    ~MatchSpyList() {
-	vector<Xapian::MatchSpy *>::const_iterator i;
-	for (i = spies.begin(); i != spies.end(); ++i) {
-	    delete *i;
-	}
-    }
-};
-
 void
 RemoteServer::msg_query(const string &message_in)
 {
@@ -379,7 +399,7 @@ RemoteServer::msg_query(const string &message_in)
     // Unserialise the Query.
     size_t len;
     decode_length_and_check(&p, p_end, len);
-    AutoPtr<Xapian::Query::Internal> query(Xapian::Query::Internal::unserialise(string(p, len), reg));
+    Xapian::Query query(Xapian::Query::unserialise(string(p, len), reg));
     p += len;
 
     // Unserialise assorted Enquire settings.
@@ -413,12 +433,14 @@ RemoteServer::msg_query(const string &message_in)
     }
     bool sort_value_forward(*p++ != '0');
 
+    double time_limit = unserialise_double(&p, p_end);
+
     int percent_cutoff = *p++;
     if (percent_cutoff < 0 || percent_cutoff > 100) {
 	throw Xapian::NetworkError("bad message (percent_cutoff)");
     }
 
-    Xapian::weight weight_cutoff = unserialise_double(&p, p_end);
+    double weight_cutoff = unserialise_double(&p, p_end);
     if (weight_cutoff < 0) {
 	throw Xapian::NetworkError("bad message (weight_cutoff)");
     }
@@ -447,7 +469,7 @@ RemoteServer::msg_query(const string &message_in)
     p += len;
 
     // Unserialise any MatchSpy objects.
-    MatchSpyList matchspies;
+    vector<Xapian::Internal::opt_intrusive_ptr<Xapian::MatchSpy>> matchspies;
     while (p != p_end) {
 	decode_length_and_check(&p, p_end, len);
 	string spytype(p, len);
@@ -459,15 +481,15 @@ RemoteServer::msg_query(const string &message_in)
 	p += len;
 
 	decode_length_and_check(&p, p_end, len);
-	matchspies.spies.push_back(spyclass->unserialise(string(p, len), reg));
+	matchspies.push_back(spyclass->unserialise(string(p, len), reg)->release());
 	p += len;
     }
 
     Xapian::Weight::Internal local_stats;
-    MultiMatch match(*db, query.get(), qlen, &rset, collapse_max, collapse_key,
+    MultiMatch match(*db, query, qlen, &rset, collapse_max, collapse_key,
 		     percent_cutoff, weight_cutoff, order,
-		     sort_key, sort_by, sort_value_forward, NULL,
-		     local_stats, wt.get(), matchspies.spies, false, false);
+		     sort_key, sort_by, sort_value_forward, time_limit,
+		     local_stats, wt.get(), matchspies, false, false);
 
     send_message(REPLY_STATS, serialise_stats(local_stats));
 
@@ -485,16 +507,17 @@ RemoteServer::msg_query(const string &message_in)
     decode_length(&p, p_end, check_at_least);
 
     message.erase(0, message.size() - (p_end - p));
-    Xapian::Weight::Internal total_stats(unserialise_stats(message));
-    total_stats.set_bounds_from_db(*db);
+    AutoPtr<Xapian::Weight::Internal> total_stats(new Xapian::Weight::Internal);
+    unserialise_stats(message, *(total_stats.get()));
+    total_stats->set_bounds_from_db(*db);
 
     Xapian::MSet mset;
-    match.get_mset(first, maxitems, check_at_least, mset, total_stats, 0, 0, 0);
+    match.get_mset(first, maxitems, check_at_least, mset, *(total_stats.get()), 0, 0);
+    mset.internal->stats = total_stats.release();
 
     message.resize(0);
-    vector<Xapian::MatchSpy *>::const_iterator i;
-    for (i = matchspies.spies.begin(); i != matchspies.spies.end(); ++i) {
-	string spy_results = (*i)->serialise_results();
+    for (auto i : matchspies) {
+	string spy_results = i->serialise_results();
 	message += encode_length(spy_results.size());
 	message += spy_results;
     }
@@ -502,137 +525,6 @@ RemoteServer::msg_query(const string &message_in)
     send_message(REPLY_RESULTS, message);
 }
 
-void
-RemoteServer::msg_query_new(const string &message_in)
-{
-    const char *p = message_in.c_str();
-    const char *p_end = p + message_in.size();
-    size_t len;
-
-    // Unserialise the Query.
-    decode_length_and_check(&p, p_end, len);
-    AutoPtr<Xapian::Query::Internal> query(Xapian::Query::Internal::unserialise(string(p, len), reg));
-    p += len;
-
-    // Unserialise assorted Enquire settings.
-    Xapian::termcount qlen;
-    decode_length(&p, p_end, qlen);
-
-    Xapian::valueno collapse_max;
-    decode_length(&p, p_end, collapse_max);
-
-    Xapian::valueno collapse_key = Xapian::BAD_VALUENO;
-   if (collapse_max) decode_length(&p, p_end, collapse_key);
-
-    if (p_end - p < 4 || *p < '0' || *p > '2') {
-	throw Xapian::NetworkError("bad message (docid_order)");
-    }
-    Xapian::Enquire::docid_order order;
-    order = static_cast<Xapian::Enquire::docid_order>(*p++ - '0');
-
-    Xapian::valueno sort_key;
-    decode_length(&p, p_end, sort_key);
-
-    if (*p < '0' || *p > '3') {
-	throw Xapian::NetworkError("bad message (sort_by)");
-    }
-    Xapian::Enquire::Internal::sort_setting sort_by;
-    sort_by = static_cast<Xapian::Enquire::Internal::sort_setting>(*p++ - '0');
-
-    if (*p < '0' || *p > '1') {
-	throw Xapian::NetworkError("bad message (sort_value_forward)");
-    }
-    bool sort_value_forward(*p++ != '0');
-
-    int percent_cutoff = *p++;
-    if (percent_cutoff < 0 || percent_cutoff > 100) {
-	throw Xapian::NetworkError("bad message (percent_cutoff)");
-    }
-
-    Xapian::weight weight_cutoff = unserialise_double(&p, p_end);
-    if (weight_cutoff < 0) {
-	throw Xapian::NetworkError("bad message (weight_cutoff)");
-    }
-
-    // Unserialise the Weight object.
-    decode_length_and_check(&p, p_end, len);
-    string wtname(p, len);
-    p += len;
-
-    const Xapian::Weight * wttype = reg.get_weighting_scheme(wtname);
-    if (wttype == NULL) {
-	// Note: user weighting schemes should be registered by adding them to
-	// a Registry, and setting the context using
-	// RemoteServer::set_registry().
-	throw Xapian::InvalidArgumentError("Weighting scheme " +
-					   wtname + " not registered");
-    }
-
-    decode_length_and_check(&p, p_end, len);
-    AutoPtr<Xapian::Weight> wt(wttype->unserialise(string(p, len)));
-    p += len;
-
-    // Unserialise the RSet object.
-    decode_length_and_check(&p, p_end, len);
-    Xapian::RSet rset = unserialise_rset(string(p, len));
-    p += len;
-
-    // Unserialise any MatchSpy objects.
-    MatchSpyList matchspies;
-    while (p != p_end) {
-	decode_length_and_check(&p, p_end, len);
-	string spytype(p, len);
-	const Xapian::MatchSpy * spyclass = reg.get_match_spy(spytype);
-	if (spyclass == NULL) {
-	    throw Xapian::InvalidArgumentError("Match spy " + spytype +
-					       " not registered");
-	}
-	p += len;
-
-	decode_length_and_check(&p, p_end, len);
-	matchspies.spies.push_back(spyclass->unserialise(string(p, len), reg));
-	p += len;
-    }
-
-    Xapian::Weight::Internal local_stats;
-    MultiMatch match(*db, query.get(), qlen, &rset, collapse_max, collapse_key,
-		     percent_cutoff, weight_cutoff, order,
-		     sort_key, sort_by, sort_value_forward, NULL,
-		     local_stats, wt.get(), matchspies.spies, false, false);
-
-    send_message(REPLY_STATS, serialise_stats(local_stats));
-
-    string message;
-    get_message(active_timeout, message, MSG_GETMSET);
-    p = message.c_str();
-    p_end = p + message.size();
-
-    Xapian::termcount first;
-    decode_length(&p, p_end, first);
-    Xapian::termcount maxitems;
-    decode_length(&p, p_end, maxitems);
-
-    Xapian::termcount check_at_least;
-    decode_length(&p, p_end, check_at_least);
-
-    message.erase(0, message.size() - (p_end - p));
-    Xapian::Weight::Internal total_stats(unserialise_stats(message));
-    total_stats.set_bounds_from_db(*db);
-
-    Xapian::MSet mset;
-    match.get_mset(first, maxitems, check_at_least, mset, total_stats, 0, 0, 0);
-
-    message.resize(0);
-    vector<Xapian::MatchSpy *>::const_iterator i;
-    for (i = matchspies.spies.begin(); i != matchspies.spies.end(); ++i) {
-	string spy_results = (*i)->serialise_results();
-	message += encode_length(spy_results.size());
-	message += spy_results;
-    }
-    message += serialise_mset_new(mset);
-    send_message(REPLY_RESULTS_NEW, message);
-}
-
 void
 RemoteServer::msg_document(const string &message)
 {
@@ -680,6 +572,14 @@ RemoteServer::msg_termfreq(const string &term)
     send_message(REPLY_TERMFREQ, encode_length(db->get_termfreq(term)));
 }
 
+void
+RemoteServer::msg_freqs(const string &term)
+{
+    string msg = encode_length(db->get_termfreq(term));
+    msg += encode_length(db->get_collection_freq(term));
+    send_message(REPLY_FREQS, msg);
+}
+
 void
 RemoteServer::msg_valuestats(const string & message)
 {
@@ -711,11 +611,21 @@ RemoteServer::msg_doclength(const string &message)
     send_message(REPLY_DOCLENGTH, encode_length(db->get_doclength(did)));
 }
 
+void
+RemoteServer::msg_uniqueterms(const string &message)
+{
+    const char *p = message.data();
+    const char *p_end = p + message.size();
+    Xapian::docid did;
+    decode_length(&p, p_end, did);
+    send_message(REPLY_UNIQUETERMS, encode_length(db->get_unique_terms(did)));
+}
+
 void
 RemoteServer::msg_commit(const string &)
 {
     if (!wdb)
-	throw Xapian::InvalidOperationError("Server is read-only");
+	throw_read_only();
 
     wdb->commit();
 
@@ -726,7 +636,7 @@ void
 RemoteServer::msg_cancel(const string &)
 {
     if (!wdb)
-	throw Xapian::InvalidOperationError("Server is read-only");
+	throw_read_only();
 
     // We can't call cancel since that's an internal method, but this
     // has the same effect with minimal additional overhead.
@@ -738,7 +648,7 @@ void
 RemoteServer::msg_adddocument(const string & message)
 {
     if (!wdb)
-	throw Xapian::InvalidOperationError("Server is read-only");
+	throw_read_only();
 
     Xapian::docid did = wdb->add_document(unserialise_document(message));
 
@@ -749,7 +659,7 @@ void
 RemoteServer::msg_deletedocument(const string & message)
 {
     if (!wdb)
-	throw Xapian::InvalidOperationError("Server is read-only");
+	throw_read_only();
 
     const char *p = message.data();
     const char *p_end = p + message.size();
@@ -765,7 +675,7 @@ void
 RemoteServer::msg_deletedocumentterm(const string & message)
 {
     if (!wdb)
-	throw Xapian::InvalidOperationError("Server is read-only");
+	throw_read_only();
 
     wdb->delete_document(message);
 }
@@ -774,7 +684,7 @@ void
 RemoteServer::msg_replacedocument(const string & message)
 {
     if (!wdb)
-	throw Xapian::InvalidOperationError("Server is read-only");
+	throw_read_only();
 
     const char *p = message.data();
     const char *p_end = p + message.size();
@@ -788,7 +698,7 @@ void
 RemoteServer::msg_replacedocumentterm(const string & message)
 {
     if (!wdb)
-	throw Xapian::InvalidOperationError("Server is read-only");
+	throw_read_only();
 
     const char *p = message.data();
     const char *p_end = p + message.size();
@@ -811,12 +721,22 @@ RemoteServer::msg_getmetadata(const string & message)
 void
 RemoteServer::msg_openmetadatakeylist(const string & message)
 {
-    const Xapian::TermIterator end = db->metadata_keys_end(message);
-    Xapian::TermIterator t = db->metadata_keys_begin(message);
-    for (; t != end; ++t) {
-	send_message(REPLY_METADATAKEYLIST, *t);
-    }
+    string prev = message;
+    string reply;
 
+    const string & prefix = message;
+    const Xapian::TermIterator end = db->metadata_keys_end(prefix);
+    Xapian::TermIterator t = db->metadata_keys_begin(prefix);
+    for (; t != end; ++t) {
+	if (rare(prev.size() > 255))
+	    prev.resize(255);
+	const string & v = *t;
+	size_t reuse = common_prefix_length(prev, v);
+	reply.assign(1, char(reuse));
+	reply.append(v, reuse, string::npos);
+	send_message(REPLY_METADATAKEYLIST, reply);
+	prev = v;
+    }
     send_message(REPLY_DONE, string());
 }
 
@@ -824,7 +744,7 @@ void
 RemoteServer::msg_setmetadata(const string & message)
 {
     if (!wdb)
-	throw Xapian::InvalidOperationError("Server is read-only");
+	throw_read_only();
     const char *p = message.data();
     const char *p_end = p + message.size();
     size_t keylen;
@@ -839,7 +759,7 @@ void
 RemoteServer::msg_addspelling(const string & message)
 {
     if (!wdb)
-	throw Xapian::InvalidOperationError("Server is read-only");
+	throw_read_only();
     const char *p = message.data();
     const char *p_end = p + message.size();
     Xapian::termcount freqinc;
@@ -851,7 +771,7 @@ void
 RemoteServer::msg_removespelling(const string & message)
 {
     if (!wdb)
-	throw Xapian::InvalidOperationError("Server is read-only");
+	throw_read_only();
     const char *p = message.data();
     const char *p_end = p + message.size();
     Xapian::termcount freqdec;
