@@ -2,7 +2,7 @@
  *
  * Copyright 1999,2000,2001 BrightStation PLC
  * Copyright 2002 Ananova Ltd
- * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010,2011,2012 Olly Betts
+ * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010,2011,2012,2015,2016 Olly Betts
  * Copyright 2006,2008 Lemur Consulting Ltd
  * Copyright 2011 Action Without Borders
  *
@@ -29,11 +29,11 @@
 #include <algorithm>
 #include <string>
 
+#define XAPIAN_DEPRECATED(X) X
 #include <xapian.h>
 #include "backendmanager_local.h"
 #include "testsuite.h"
 #include "testutils.h"
-#include "utils.h"
 
 #include "apitest.h"
 
@@ -336,10 +336,19 @@ DEFINE_TESTCASE(expandweights1, backend) {
 
     Xapian::ESet eset = enquire.get_eset(3, myrset, enquire.USE_EXACT_TERMFREQ);
     TEST_EQUAL(eset.size(), 3);
+    TEST_REL(eset.get_ebound(), >=, eset.size());
     TEST_EQUAL_DOUBLE(eset[0].get_weight(), 6.08904001099445);
     TEST_EQUAL_DOUBLE(eset[1].get_weight(), 6.08904001099445);
     TEST_EQUAL_DOUBLE(eset[2].get_weight(), 4.73383620844021);
 
+    // Test non-default k too.
+    eset = enquire.get_eset(3, myrset, enquire.USE_EXACT_TERMFREQ, 2.0);
+    TEST_EQUAL(eset.size(), 3);
+    TEST_REL(eset.get_ebound(), >=, eset.size());
+    TEST_EQUAL_DOUBLE(eset[0].get_weight(), 5.88109547674955);
+    TEST_EQUAL_DOUBLE(eset[1].get_weight(), 5.88109547674955);
+    TEST_EQUAL_DOUBLE(eset[2].get_weight(), 5.44473599216144);
+
     return true;
 }
 
@@ -357,6 +366,7 @@ DEFINE_TESTCASE(expandweights2, backend) {
 
     Xapian::ESet eset = enquire.get_eset(3, myrset);
     TEST_EQUAL(eset.size(), 3);
+    TEST_REL(eset.get_ebound(), >=, eset.size());
     if (!startswith(get_dbtype(), "multi")) {
 	// For a single database, the weights should be the same with or
 	// without USE_EXACT_TERMFREQ.
@@ -385,8 +395,10 @@ DEFINE_TESTCASE(expandweights3, backend) {
     myrset.add_document(*i);
     myrset.add_document(*(++i));
 
-    // Set min_wt to 0
-    Xapian::ESet eset = enquire.get_eset(50, myrset, 0, 1.0, 0, 0);
+    // Set min_wt to 0.0
+    Xapian::ESet eset = enquire.get_eset(50, myrset, 0, 0, 0.0);
+    TEST_EQUAL(eset.size(), 50);
+    TEST_REL(eset.get_ebound(), >=, eset.size());
     if (!startswith(get_dbtype(), "multi")) {
 	// For a single database, the weights should be the same with or
 	// without USE_EXACT_TERMFREQ.
@@ -418,15 +430,105 @@ DEFINE_TESTCASE(expandweights4, backend) {
     myrset.add_document(*i);
     myrset.add_document(*(++i));
 
-    Xapian::ESet eset = enquire.get_eset(37, myrset, 0, 1.0, 0, -100);
+    Xapian::ESet eset = enquire.get_eset(37, myrset, 0, 0, -100);
     // Now include negative weights
     TEST_EQUAL(eset.size(), 37);
+    TEST_REL(eset.get_ebound(), >=, eset.size());
     TEST_REL(eset[36].get_weight(),<,0);
     TEST_REL(eset[36].get_weight(),>=,-100);
 
     return true;
 }
 
+// test for Bo1EWeight
+DEFINE_TESTCASE(expandweights5, backend) {
+    Xapian::Enquire enquire(get_database("apitest_simpledata"));
+    enquire.set_query(Xapian::Query("this"));
+
+    Xapian::MSet mymset = enquire.get_mset(0, 10);
+
+    Xapian::RSet myrset;
+    Xapian::MSetIterator i = mymset.begin();
+    myrset.add_document(*i);
+    myrset.add_document(*(++i));
+
+    enquire.set_expansion_scheme("bo1");
+    Xapian::ESet eset = enquire.get_eset(3, myrset);
+
+    TEST_EQUAL(eset.size(), 3);
+    TEST_REL(eset.get_ebound(), >=, eset.size());
+    TEST_EQUAL_DOUBLE(eset[0].get_weight(), 7.21765284821702);
+    TEST_EQUAL_DOUBLE(eset[1].get_weight(), 6.661623193760022);
+    TEST_EQUAL_DOUBLE(eset[2].get_weight(), 5.58090119783738);
+
+    return true;
+}
+
+// test that "trad" can be set as an expansion scheme.
+DEFINE_TESTCASE(expandweights6, backend) {
+    Xapian::Enquire enquire(get_database("apitest_simpledata"));
+    enquire.set_query(Xapian::Query("this"));
+
+    Xapian::MSet mymset = enquire.get_mset(0, 10);
+
+    Xapian::RSet myrset;
+    Xapian::MSetIterator i = mymset.begin();
+    myrset.add_document(*i);
+    myrset.add_document(*(++i));
+
+    enquire.set_expansion_scheme("trad");
+    Xapian::ESet eset = enquire.get_eset(3, myrset, enquire.USE_EXACT_TERMFREQ);
+
+    TEST_EQUAL(eset.size(), 3);
+    TEST_REL(eset.get_ebound(), >=, eset.size());
+    TEST_EQUAL_DOUBLE(eset[0].get_weight(), 6.08904001099445);
+    TEST_EQUAL_DOUBLE(eset[1].get_weight(), 6.08904001099445);
+    TEST_EQUAL_DOUBLE(eset[2].get_weight(), 4.73383620844021);
+
+    return true;
+}
+
+// test that invalid scheme names are not accepted
+DEFINE_TESTCASE(expandweights7, backend) {
+    Xapian::Enquire enquire(get_database("apitest_simpledata"));
+
+    TEST_EXCEPTION(Xapian::InvalidArgumentError,
+		   enquire.set_expansion_scheme("no_such_scheme"));
+
+    return true;
+}
+
+// test that "expand_k" can be passed as a parameter to get_eset
+DEFINE_TESTCASE(expandweights8, backend) {
+    Xapian::Enquire enquire(get_database("apitest_simpledata"));
+    enquire.set_query(Xapian::Query("this"));
+
+    Xapian::MSet mymset = enquire.get_mset(0, 10);
+
+    Xapian::RSet myrset;
+    Xapian::MSetIterator i = mymset.begin();
+    myrset.add_document(*i);
+    myrset.add_document(*(++i));
+
+    // Set expand_k to 1.0 and min_wt to 0
+    Xapian::ESet eset = enquire.get_eset(50, myrset, 0, 1.0, 0, 0);
+    if (!startswith(get_dbtype(), "multi")) {
+	// For a single database, the weights should be the same with or
+	// without USE_EXACT_TERMFREQ.
+	TEST_EQUAL_DOUBLE(eset[0].get_weight(), 6.08904001099445);
+	TEST_EQUAL_DOUBLE(eset[1].get_weight(), 6.08904001099445);
+	TEST_EQUAL_DOUBLE(eset[2].get_weight(), 4.73383620844021);
+    } else {
+	// For multiple databases, we expect that using USE_EXACT_TERMFREQ
+	// will result in different weights in some cases.
+	TEST_NOT_EQUAL_DOUBLE(eset[0].get_weight(), 6.08904001099445);
+	TEST_EQUAL_DOUBLE(eset[1].get_weight(), 6.08904001099445);
+	TEST_NOT_EQUAL_DOUBLE(eset[2].get_weight(), 4.73383620844021);
+    }
+    TEST_REL(eset.back().get_weight(),>=,0);
+
+    return true;
+}
 
 // tests that when specifying maxitems to get_eset, no more than
 // that are returned.
@@ -445,6 +547,7 @@ DEFINE_TESTCASE(expandmaxitems1, backend) {
 
     Xapian::ESet myeset = enquire.get_eset(1, myrset);
     TEST_EQUAL(myeset.size(), 1);
+    TEST_REL(myeset.get_ebound(), >=, myeset.size());
 
     return true;
 }
@@ -640,6 +743,57 @@ DEFINE_TESTCASE(expandfunctor1, backend) {
     return true;
 }
 
+DEFINE_TESTCASE(expanddeciderfilterprefix2, backend) {
+    Xapian::Enquire enquire(get_database("apitest_simpledata"));
+    enquire.set_query(Xapian::Query("this"));
+
+    Xapian::MSet mymset = enquire.get_mset(0, 10);
+    TEST(mymset.size() >= 2);
+
+    Xapian::RSet myrset;
+    Xapian::MSetIterator i = mymset.begin();
+    myrset.add_document(*i);
+    myrset.add_document(*(++i));
+
+    Xapian::ESet myeset_orig = enquire.get_eset(1000, myrset);
+    unsigned int neweset_size = 0;
+
+    // Choose the first char in the first term as prefix.
+    Xapian::ESetIterator j = myeset_orig.begin();
+    TEST(myeset_orig.size() >= 1);
+    string prefix(*j, 0, 1);
+    Xapian::ExpandDeciderFilterPrefix myfunctor(prefix);
+
+    for ( ; j != myeset_orig.end(); ++j) {
+	if (myfunctor(*j)) neweset_size++;
+    }
+    Xapian::ESet myeset = enquire.get_eset(neweset_size, myrset, &myfunctor);
+
+    Xapian::ESetIterator orig = myeset_orig.begin();
+    Xapian::ESetIterator filt = myeset.begin();
+    for (; orig != myeset_orig.end() && filt != myeset.end(); ++orig, ++filt) {
+	// skip over items that shouldn't be in myeset
+	while (orig != myeset_orig.end() && !myfunctor(*orig)) {
+	    ++orig;
+	}
+
+	TEST_AND_EXPLAIN(*orig == *filt &&
+	    orig.get_weight() == filt.get_weight(),
+	    "Mismatch in items " << *orig << " vs. " << *filt
+	    << " after filtering");
+    }
+
+    while (orig != myeset_orig.end() && !myfunctor(*orig)) {
+	++orig;
+    }
+
+    TEST_EQUAL(orig, myeset_orig.end());
+    TEST_AND_EXPLAIN(filt == myeset.end(),
+	"Extra items in the filtered eset.");
+
+    return true;
+}
+
 // tests the percent cutoff option
 DEFINE_TESTCASE(pctcutoff1, backend) {
     Xapian::Enquire enquire(get_database("apitest_simpledata"));
@@ -709,7 +863,7 @@ DEFINE_TESTCASE(pctcutoff2, backend) {
     TEST(mset.size() >= 2);
     TEST(mset[0].get_percent() - mset[1].get_percent() >= 2);
 
-    Xapian::percent cutoff = mset[0].get_percent() + mset[1].get_percent();
+    int cutoff = mset[0].get_percent() + mset[1].get_percent();
     cutoff /= 2;
 
     enquire.set_cutoff(cutoff);
@@ -767,12 +921,12 @@ DEFINE_TESTCASE(cutoff1, backend) {
     }
 
     unsigned int num_items = 0;
-    Xapian::weight my_wt = -100;
+    double my_wt = -100;
     int changes = 0;
     Xapian::MSetIterator i = mymset1.begin();
     int c = 0;
     for ( ; i != mymset1.end(); ++i, ++c) {
-        Xapian::weight new_wt = i.get_weight();
+	double new_wt = i.get_weight();
         if (new_wt != my_wt) {
 	    changes++;
 	    if (changes > 3) break;
@@ -841,7 +995,7 @@ DEFINE_TESTCASE(maxattain1, backend) {
     enquire.set_query(query("this"));
     Xapian::MSet mymset = enquire.get_mset(0, 100);
 
-    Xapian::weight mymax = 0;
+    double mymax = 0;
     Xapian::MSetIterator i = mymset.begin();
     for ( ; i != mymset.end(); ++i) {
         if (i.get_weight() > mymax) mymax = i.get_weight();
@@ -1005,30 +1159,6 @@ DEFINE_TESTCASE(getmterms2, backend) {
     return true;
 }
 
-// tests that the collapsing on termpos optimisation works
-DEFINE_TESTCASE(poscollapse1, backend) {
-    Xapian::Query myquery1(Xapian::Query::OP_OR,
-		     Xapian::Query("this", 1, 1),
-		     Xapian::Query("this", 1, 1));
-    Xapian::Query myquery2("this", 2, 1);
-
-    if (verbose) {
-	tout << myquery1.get_description() << "\n";
-	tout << myquery2.get_description() << "\n";
-    }
-
-    Xapian::Enquire enquire(get_database("apitest_simpledata"));
-    enquire.set_query(myquery1);
-    Xapian::MSet mymset1 = enquire.get_mset(0, 10);
-
-    enquire.set_query(myquery2);
-    Xapian::MSet mymset2 = enquire.get_mset(0, 10);
-
-    TEST_EQUAL(mymset1, mymset2);
-
-    return true;
-}
-
 // test that running a query twice returns the same results
 DEFINE_TESTCASE(repeatquery1, backend) {
     Xapian::Enquire enquire(get_database("apitest_simpledata"));
@@ -1192,10 +1322,6 @@ DEFINE_TESTCASE(rsetmultidb3, backend && !multi) {
 
 /// Simple test of the elite set operator.
 DEFINE_TESTCASE(eliteset1, backend) {
-    // FIXME: OP_ELITE_SET erroneously picks the best N terms separately in
-    // each sub-database!
-    SKIP_TEST_FOR_BACKEND("multi");
-
     Xapian::Database mydb(get_database("apitest_simpledata"));
     Xapian::Enquire enquire(mydb);
 
@@ -1217,10 +1343,6 @@ DEFINE_TESTCASE(eliteset1, backend) {
 /// Test that the elite set operator works if the set contains
 /// sub-expressions (regression test)
 DEFINE_TESTCASE(eliteset2, backend) {
-    // FIXME: OP_ELITE_SET erroneously picks the best N terms separately in
-    // each sub-database!
-    SKIP_TEST_FOR_BACKEND("multi");
-
     Xapian::Database mydb(get_database("apitest_simpledata"));
     Xapian::Enquire enquire(mydb);
 
@@ -1296,10 +1418,6 @@ DEFINE_TESTCASE(eliteset3, backend) {
 
 /// Test that elite set doesn't pick terms with 0 frequency
 DEFINE_TESTCASE(eliteset4, backend) {
-    // FIXME: OP_ELITE_SET erroneously picks the best N terms separately in
-    // each sub-database!
-    SKIP_TEST_FOR_BACKEND("multi");
-
     Xapian::Database mydb1(get_database("apitest_simpledata"));
     Xapian::Enquire enquire1(mydb1);
 
@@ -1325,8 +1443,6 @@ DEFINE_TESTCASE(eliteset4, backend) {
 
 /// Regression test for problem with excess precision.
 DEFINE_TESTCASE(eliteset5, backend) {
-    SKIP_TEST_FOR_BACKEND("multi");
-
     Xapian::Database mydb1(get_database("apitest_simpledata"));
     Xapian::Enquire enquire1(mydb1);
 
@@ -1373,8 +1489,8 @@ DEFINE_TESTCASE(termlisttermfreq1, backend) {
     // search for weight of term 'another'
     string theterm = stemmer("another");
 
-    Xapian::weight wt1 = 0;
-    Xapian::weight wt2 = 0;
+    double wt1 = 0;
+    double wt2 = 0;
     {
 	Xapian::ESetIterator i = eset1.begin();
 	for ( ; i != eset1.end(); i++) {
@@ -1858,7 +1974,11 @@ DEFINE_TESTCASE(emptyop1, backend) {
     enquire.set_query(query1);
     Xapian::MSet mymset = enquire.get_mset(0, 10);
     TEST_MSET_SIZE(mymset, 0);
-    TEST_EXCEPTION(Xapian::InvalidArgumentError, enquire.get_matching_terms_begin(1));
+    // In Xapian < 1.3.0, this gave InvalidArgumentError (because
+    // query1.empty()) but elsewhere we treat an empty query as just not
+    // matching any documents, so we now do the same here too.
+    TEST_EQUAL(enquire.get_matching_terms_begin(1),
+	       enquire.get_matching_terms_end(1));
 
     return true;
 }
@@ -2023,6 +2143,7 @@ DEFINE_TESTCASE(alldocspl1, writable) {
     TEST(i != db.postlist_end(""));
     TEST_EQUAL(*i, 5);
     TEST_EQUAL(i.get_doclength(), 0);
+    TEST_EQUAL(i.get_unique_terms(), 0);
     TEST_EQUAL(i.get_wdf(), 1);
     ++i;
     TEST(i == db.postlist_end(""));
@@ -2046,6 +2167,7 @@ DEFINE_TESTCASE(alldocspl2, writable) {
 	TEST(i != end);
 	TEST_EQUAL(*i, 5);
 	TEST_EQUAL(i.get_doclength(), 0);
+	TEST_EQUAL(i.get_unique_terms(), 0);
 	TEST_EQUAL(i.get_wdf(), 1);
 	++i;
 	TEST(i == end);
@@ -2058,6 +2180,7 @@ DEFINE_TESTCASE(alldocspl2, writable) {
 	TEST(i != end);
 	TEST_EQUAL(*i, 5);
 	TEST_EQUAL(i.get_doclength(), 0);
+	TEST_EQUAL(i.get_unique_terms(), 0);
 	TEST_EQUAL(i.get_wdf(), 1);
 	++i;
 	TEST(i == end);
@@ -2074,11 +2197,13 @@ DEFINE_TESTCASE(alldocspl2, writable) {
 	TEST(i != end);
 	TEST_EQUAL(*i, 5);
 	TEST_EQUAL(i.get_doclength(), 0);
+	TEST_EQUAL(i.get_unique_terms(), 0);
 	TEST_EQUAL(i.get_wdf(), 1);
 	++i;
 	TEST(i != end);
 	TEST_EQUAL(*i, 7);
 	TEST_EQUAL(i.get_doclength(), 0);
+	TEST_EQUAL(i.get_unique_terms(), 0);
 	TEST_EQUAL(i.get_wdf(), 1);
 	++i;
 	TEST(i == end);
@@ -2092,6 +2217,7 @@ DEFINE_TESTCASE(alldocspl2, writable) {
 	TEST(i != end);
 	TEST_EQUAL(*i, 7);
 	TEST_EQUAL(i.get_doclength(), 0);
+	TEST_EQUAL(i.get_unique_terms(), 0);
 	TEST_EQUAL(i.get_wdf(), 1);
 	++i;
 	TEST(i == end);
@@ -2106,6 +2232,7 @@ DEFINE_TESTCASE(alldocspl2, writable) {
     TEST(i != end);
     TEST_EQUAL(*i, 7);
     TEST_EQUAL(i.get_doclength(), 0);
+    TEST_EQUAL(i.get_unique_terms(), 0);
     TEST_EQUAL(i.get_wdf(), 1);
     ++i;
     TEST(i == end);
@@ -2319,7 +2446,7 @@ DEFINE_TESTCASE(uuid1, backend && !multi) {
 #ifdef XAPIAN_HAS_INMEMORY_BACKEND
     // This relies on InMemory databases not supporting uuids.
     // A multi-database containing a database with no uuid has no uuid.
-    db2.add_database(Xapian::InMemory::open());
+    db2.add_database(Xapian::Database(string(), Xapian::DB_BACKEND_INMEMORY));
     TEST(db2.get_uuid().empty());
 #endif
 
