@@ -1,7 +1,7 @@
 /** @file api_backend.cc
  * @brief Backend-related tests.
  */
-/* Copyright (C) 2008,2009,2010,2011,2013,2015 Olly Betts
+/* Copyright (C) 2008,2009,2010,2011,2012,2013,2014,2015,2016 Olly Betts
  * Copyright (C) 2010 Richard Boulton
  *
  * This program is free software; you can redistribute it and/or
@@ -27,24 +27,32 @@
 #define XAPIAN_DEPRECATED(X) X
 #include <xapian.h>
 
+#include "backendmanager.h"
+#include "filetests.h"
 #include "str.h"
+#include "testrunner.h"
 #include "testsuite.h"
 #include "testutils.h"
 #include "unixcmds.h"
-#include "utils.h"
 
 #include "apitest.h"
 
-#include "safeerrno.h"
 #include "safefcntl.h"
 #include "safesysstat.h"
 #include "safeunistd.h"
+#ifdef HAVE_SOCKETPAIR
+# include "safesyssocket.h"
+# include <signal.h>
+# include "safesyswait.h"
+#endif
+
+#include <fstream>
 
 using namespace std;
 
 /// Regression test - lockfile should honour umask, was only user-readable.
-DEFINE_TESTCASE(lockfileumask1, brass || chert || flint) {
-#if !defined __WIN32__ && !defined __CYGWIN__ && !defined __EMX__
+DEFINE_TESTCASE(lockfileumask1, chert || glass) {
+#if !defined __WIN32__ && !defined __CYGWIN__ && !defined __OS2__
     mode_t old_umask = umask(022);
     try {
 	Xapian::WritableDatabase db = get_named_writable_database("lockfileumask1");
@@ -53,7 +61,7 @@ DEFINE_TESTCASE(lockfileumask1, brass || chert || flint) {
 	path += "/flintlock";
 
 	struct stat statbuf;
-	TEST(stat(path, &statbuf) == 0);
+	TEST(stat(path.c_str(), &statbuf) == 0);
 	TEST_EQUAL(statbuf.st_mode & 0777, 0644);
     } catch (...) {
 	umask(old_umask);
@@ -72,12 +80,16 @@ DEFINE_TESTCASE(totaldoclen1, writable) {
     Xapian::Document doc;
     doc.add_posting("foo", 1, 2000000000);
     db.add_document(doc);
-    db.add_document(doc);
+    Xapian::Document doc2;
+    doc2.add_posting("bar", 1, 2000000000);
+    db.add_document(doc2);
     TEST_EQUAL(db.get_avlength(), 2000000000);
     db.commit();
     TEST_EQUAL(db.get_avlength(), 2000000000);
     for (int i = 0; i != 20; ++i) {
-	db.add_document(doc);
+	Xapian::Document doc3;
+	doc3.add_posting("count" + str(i), 1, 2000000000);
+	db.add_document(doc3);
     }
     TEST_EQUAL(db.get_avlength(), 2000000000);
     db.commit();
@@ -90,6 +102,47 @@ DEFINE_TESTCASE(totaldoclen1, writable) {
     return true;
 }
 
+// Check that exceeding 32bit in combined database doesn't cause a problem
+// when using 64bit docids.
+DEFINE_TESTCASE(exceed32bitcombineddb1, writable) {
+    // Test case is for 64-bit Xapian::docid.
+    // FIXME: Though we should check that the overflow is handled gracefully
+    // for 32-bit...
+    if (sizeof(Xapian::docid) == 4) return true;
+
+    // The InMemory backend uses a vector for the documents, so trying to add
+    // a document with the maximum docid is likely to fail because we can't
+    // allocate enough memory!
+    SKIP_TEST_FOR_BACKEND("inmemory");
+
+    Xapian::WritableDatabase db1 = get_writable_database();
+    Xapian::WritableDatabase db2 = get_writable_database();
+    Xapian::Document doc;
+    doc.set_data("prose");
+    doc.add_term("word");
+
+    Xapian::docid max_id = 0xffffffff;
+
+    db1.replace_document(max_id, doc);
+    db2.replace_document(max_id, doc);
+
+    Xapian::Database db;
+    db.add_database(db1);
+    db.add_database(db2);
+
+    Xapian::Enquire enquire(db);
+    enquire.set_query(Xapian::Query::MatchAll);
+    Xapian::MSet mymset = enquire.get_mset(0, 10);
+
+    TEST_EQUAL(2, mymset.size());
+
+    for (Xapian::MSetIterator i = mymset.begin(); i != mymset.end(); ++i) {
+	TEST_EQUAL("prose", i.get_document().get_data());
+    }
+
+    return true;
+}
+
 DEFINE_TESTCASE(dbstats1, backend) {
     Xapian::Database db = get_database("etext");
 
@@ -99,27 +152,29 @@ DEFINE_TESTCASE(dbstats1, backend) {
     const Xapian::termcount max_len = 532;
     const Xapian::termcount max_wdf = 22;
 
-    if (get_dbtype() != "inmemory" &&
-	get_dbtype().find("flint") == string::npos) {
+    if (get_dbtype() != "inmemory") {
 	// Should be exact as no deletions have happened.
 	TEST_EQUAL(db.get_doclength_upper_bound(), max_len);
 	TEST_EQUAL(db.get_doclength_lower_bound(), min_len);
     } else {
-	// For flint and inmemory, we usually give rather loose bounds.
+	// For inmemory, we usually give rather loose bounds.
 	TEST_REL(db.get_doclength_upper_bound(),>=,max_len);
 	TEST_REL(db.get_doclength_lower_bound(),<=,min_len);
     }
 
-    if (get_dbtype() != "inmemory" && !startswith(get_dbtype(), "remote") &&
-	get_dbtype().find("flint") == string::npos) {
+    if (get_dbtype() != "inmemory" && !startswith(get_dbtype(), "remote")) {
 	TEST_EQUAL(db.get_wdf_upper_bound("the"), max_wdf);
     } else {
-	// For flint, inmemory and remote backends, we usually give rather
-	// loose bounds (remote matches use tighter bounds, but querying the
+	// For inmemory and remote backends, we usually give rather loose
+	// bounds (remote matches use tighter bounds, but querying the
 	// wdf bound gives a looser one).
 	TEST_REL(db.get_wdf_upper_bound("the"),>=,max_wdf);
     }
 
+    // This failed with an assertion during development between 1.3.1 and
+    // 1.3.2.
+    TEST_EQUAL(db.get_wdf_upper_bound(""), 0);
+
     return true;
 }
 
@@ -134,23 +189,21 @@ DEFINE_TESTCASE(dbstats2, backend) {
     const Xapian::termcount max_len = 15;
     const Xapian::termcount max_wdf = 7;
 
-    if (get_dbtype() != "inmemory" &&
-	get_dbtype().find("flint") == string::npos) {
+    if (get_dbtype() != "inmemory") {
 	// Should be exact as no deletions have happened.
 	TEST_EQUAL(db.get_doclength_upper_bound(), max_len);
 	TEST_EQUAL(db.get_doclength_lower_bound(), min_len);
     } else {
-	// For inmemory and flint, we usually give rather loose bounds.
+	// For inmemory, we usually give rather loose bounds.
 	TEST_REL(db.get_doclength_upper_bound(),>=,max_len);
 	TEST_REL(db.get_doclength_lower_bound(),<=,min_len);
     }
 
-    if (get_dbtype() != "inmemory" && get_dbtype() != "flint" &&
-	!startswith(get_dbtype(), "remote")) {
+    if (get_dbtype() != "inmemory" && !startswith(get_dbtype(), "remote")) {
 	TEST_EQUAL(db.get_wdf_upper_bound("word"), max_wdf);
     } else {
-	// For inmemory, flint, and remote backends, we usually give rather
-	// loose bounds (remote matches use tighter bounds, but querying the
+	// For inmemory and remote backends, we usually give rather loose
+	// bounds (remote matches use tighter bounds, but querying the
 	// wdf bound gives a looser one).
 	TEST_REL(db.get_wdf_upper_bound("word"),>=,max_wdf);
     }
@@ -194,9 +247,11 @@ DEFINE_TESTCASE(modifiedpostlist1, writable) {
 DEFINE_TESTCASE(doclenaftercommit1, writable) {
     Xapian::WritableDatabase db = get_writable_database();
     TEST_EXCEPTION(Xapian::DocNotFoundError, db.get_doclength(1));
+    TEST_EXCEPTION(Xapian::DocNotFoundError, db.get_unique_terms(1));
     db.replace_document(1, Xapian::Document());
     db.commit();
     TEST_EQUAL(db.get_doclength(1), 0);;
+    TEST_EQUAL(db.get_unique_terms(1), 0);;
     return true;
 }
 
@@ -214,8 +269,8 @@ DEFINE_TESTCASE(valuesaftercommit1, writable) {
     return true;
 }
 
-DEFINE_TESTCASE(lockfilefd0or1, brass || chert || flint) {
-#if !defined __WIN32__ && !defined __CYGWIN__ && !defined __EMX__
+DEFINE_TESTCASE(lockfilefd0or1, chert || glass) {
+#if !defined __WIN32__ && !defined __CYGWIN__ && !defined __OS2__
     int old_stdin = dup(0);
     int old_stdout = dup(1);
     try {
@@ -257,7 +312,7 @@ DEFINE_TESTCASE(lockfilefd0or1, brass || chert || flint) {
 }
 
 /// Regression test for bug fixed in 1.2.13 and 1.3.1.
-DEFINE_TESTCASE(lockfilealreadyopen1, brass || chert) {
+DEFINE_TESTCASE(lockfilealreadyopen1, chert || glass) {
     string path = get_named_writable_database_path("lockfilealreadyopen1");
     int fd = ::open((path + "/flintlock").c_str(), O_RDONLY);
     try {
@@ -291,22 +346,13 @@ DEFINE_TESTCASE(matchdecider4, remote) {
     Xapian::Enquire enquire(db);
     enquire.set_query(Xapian::Query("paragraph"));
 
-    MyMatchDecider mdecider, mspyold;
+    MyMatchDecider mdecider;
     Xapian::MSet mset;
 
     TEST_EXCEPTION(Xapian::UnimplementedError,
 	mset = enquire.get_mset(0, 10, NULL, &mdecider));
     TEST(!mdecider.called);
 
-    TEST_EXCEPTION(Xapian::UnimplementedError,
-	mset = enquire.get_mset(0, 10, 0, NULL, NULL, &mspyold));
-    TEST(!mspyold.called);
-
-    TEST_EXCEPTION(Xapian::UnimplementedError,
-	mset = enquire.get_mset(0, 10, 0, NULL, &mdecider, &mspyold));
-    TEST(!mdecider.called);
-    TEST(!mspyold.called);
-
     return true;
 }
 
@@ -345,7 +391,7 @@ DEFINE_TESTCASE(replacedoc7, writable && !inmemory && !remote) {
     TEST_EQUAL(rodb.get_doccount(), 1);
 
     db.flush();
-    rodb.reopen();
+    TEST(rodb.reopen());
 
     TEST_EQUAL(rodb.get_doccount(), 2);
     return true;
@@ -429,7 +475,7 @@ DEFINE_TESTCASE(qpmemoryleak1, writable && !inmemory) {
 
     doc.add_term("foo");
     for (int i = 100; i < 120; ++i) {
-        doc.add_term(str(i));
+	doc.add_term(str(i));
     }
 
     for (int j = 0; j < 50; ++j) {
@@ -441,11 +487,12 @@ DEFINE_TESTCASE(qpmemoryleak1, writable && !inmemory) {
     Xapian::QueryParser queryparser;
     queryparser.set_database(database);
     TEST_EXCEPTION(Xapian::DatabaseModifiedError,
-	for (int k = 0; k < 3; ++k) {
+	for (int k = 0; k < 1000; ++k) {
 	    wdb.add_document(doc);
 	    wdb.commit();
 	    (void)queryparser.parse_query("1", queryparser.FLAG_PARTIAL);
 	}
+	SKIP_TEST("didn't manage to trigger DatabaseModifiedError");
     );
 
     return true;
@@ -723,7 +770,7 @@ DEFINE_TESTCASE(orcheck1, generated) {
  *
  *  We failed to mark the Btree as unmodified after cancel().
  */
-DEFINE_TESTCASE(failedreplace1, brass || chert || flint) {
+DEFINE_TESTCASE(failedreplace1, chert || glass) {
     Xapian::WritableDatabase db(get_writable_database());
     Xapian::Document doc;
     doc.add_term("foo");
@@ -739,7 +786,7 @@ DEFINE_TESTCASE(failedreplace1, brass || chert || flint) {
     return true;
 }
 
-DEFINE_TESTCASE(failedreplace2, brass || chert || flint) {
+DEFINE_TESTCASE(failedreplace2, chert || glass) {
     Xapian::WritableDatabase db(get_writable_database("apitest_simpledata"));
     db.commit();
     Xapian::doccount db_size = db.get_doccount();
@@ -803,13 +850,23 @@ DEFINE_TESTCASE(bm25weight2, backend) {
     enquire.set_weighting_scheme(Xapian::BM25Weight(0, 0, 0, 0, 1));
     Xapian::MSet mset = enquire.get_mset(0, 100);
     TEST_REL(mset.size(),>=,2);
-    Xapian::weight weight0 = mset[0].get_weight();
+    double weight0 = mset[0].get_weight();
     for (size_t i = 1; i != mset.size(); ++i) {
 	TEST_EQUAL(weight0, mset[i].get_weight());
     }
     return true;
 }
 
+DEFINE_TESTCASE(unigramlmweight2,backend) {
+    Xapian::Database db(get_database("etext"));
+    Xapian::Enquire enquire(db);
+    enquire.set_query(Xapian::Query("the"));
+    enquire.set_weighting_scheme(Xapian::LMWeight());
+    Xapian::MSet mset = enquire.get_mset(0, 100);
+    TEST_REL(mset.size(),>=,2);
+    return true;
+}
+
 DEFINE_TESTCASE(tradweight2, backend) {
     Xapian::Database db(get_database("etext"));
     Xapian::Enquire enquire(db);
@@ -817,7 +874,7 @@ DEFINE_TESTCASE(tradweight2, backend) {
     enquire.set_weighting_scheme(Xapian::TradWeight(0));
     Xapian::MSet mset = enquire.get_mset(0, 100);
     TEST_REL(mset.size(),>=,2);
-    Xapian::weight weight0 = mset[0].get_weight();
+    double weight0 = mset[0].get_weight();
     for (size_t i = 1; i != mset.size(); ++i) {
 	TEST_EQUAL(weight0, mset[i].get_weight());
     }
@@ -856,9 +913,10 @@ DEFINE_TESTCASE(emptydb1, backend) {
 // Regression test for bug fixed in 1.3.1 and 1.2.11.
 DEFINE_TESTCASE(stubdb7, !backend) {
     TEST_EXCEPTION(Xapian::DatabaseOpeningError,
-		   Xapian::Auto::open_stub("nosuchdirectory"));
+	    Xapian::Database("nosuchdirectory", Xapian::DB_BACKEND_STUB));
     TEST_EXCEPTION(Xapian::DatabaseOpeningError,
-		   Xapian::Auto::open_stub("nosuchdirectory", Xapian::DB_OPEN));
+	    Xapian::WritableDatabase("nosuchdirectory",
+		Xapian::DB_OPEN|Xapian::DB_BACKEND_STUB));
     return true;
 }
 
@@ -943,36 +1001,23 @@ DEFINE_TESTCASE(itorskiptofromend1, backend) {
 // Regression test for bug fixed in 1.2.17 and 1.3.2 - the size gets fixed
 // but the uncorrected size was passed to the base file.  Also, abort() was
 // called on 0.
-DEFINE_TESTCASE(blocksize1, brass || chert || flint) {
+DEFINE_TESTCASE(blocksize1, chert || glass) {
     string db_dir = "." + get_dbtype();
     mkdir(db_dir.c_str(), 0755);
     db_dir += "/db__blocksize1";
+    int flags;
+    if (get_dbtype() == "chert") {
+	flags = Xapian::DB_CREATE|Xapian::DB_BACKEND_CHERT;
+    } else {
+	flags = Xapian::DB_CREATE|Xapian::DB_BACKEND_GLASS;
+    }
     static const unsigned bad_sizes[] = {
 	65537, 8000, 2000, 1024, 16, 7, 3, 1, 0
     };
     for (size_t i = 0; i < sizeof(bad_sizes) / sizeof(bad_sizes[0]); ++i) {
 	size_t block_size = bad_sizes[i];
 	rm_rf(db_dir);
-	Xapian::WritableDatabase db;
-	if (get_dbtype() == "chert") {
-#ifdef XAPIAN_HAS_CHERT_BACKEND
-	    db = Xapian::Chert::open(db_dir, Xapian::DB_CREATE, block_size);
-#else
-	    SKIP_TEST("chert backend disabled");
-#endif
-	} else if (get_dbtype() == "flint") {
-#ifdef XAPIAN_HAS_FLINT_BACKEND
-	    db = Xapian::Flint::open(db_dir, Xapian::DB_CREATE, block_size);
-#else
-	    SKIP_TEST("flint backend disabled");
-#endif
-	} else {
-#ifdef XAPIAN_HAS_BRASS_BACKEND
-	    db = Xapian::Brass::open(db_dir, Xapian::DB_CREATE, block_size);
-#else
-	    SKIP_TEST("brass backend disabled");
-#endif
-	}
+	Xapian::WritableDatabase db(db_dir, flags, block_size);
 	Xapian::Document doc;
 	doc.add_term("XYZ");
 	doc.set_data("foo");
@@ -982,6 +1027,82 @@ DEFINE_TESTCASE(blocksize1, brass || chert || flint) {
     return true;
 }
 
+/// Feature test for Xapian::DB_NO_TERMLIST.
+DEFINE_TESTCASE(notermlist1, glass) {
+    string db_dir = "." + get_dbtype();
+    mkdir(db_dir.c_str(), 0755);
+    db_dir += "/db__notermlist1";
+    int flags = Xapian::DB_CREATE|Xapian::DB_NO_TERMLIST;
+    if (get_dbtype() == "chert") {
+	flags |= Xapian::DB_BACKEND_CHERT;
+    } else {
+	flags |= Xapian::DB_BACKEND_GLASS;
+    }
+    rm_rf(db_dir);
+    Xapian::WritableDatabase db(db_dir, flags);
+    Xapian::Document doc;
+    doc.add_term("hello");
+    doc.add_value(42, "answer");
+    db.add_document(doc);
+    db.commit();
+    TEST(!file_exists(db_dir + "/termlist.glass"));
+    TEST_EXCEPTION(Xapian::FeatureUnavailableError, db.termlist_begin(1));
+    return true;
+}
+
+/// Regression test for bug starting a new glass freelist block.
+DEFINE_TESTCASE(newfreelistblock1, writable) {
+    Xapian::Document doc;
+    doc.add_term("foo");
+    for (int i = 100; i < 120; ++i) {
+	doc.add_term(str(i));
+    }
+
+    Xapian::WritableDatabase wdb(get_writable_database());
+    for (int j = 0; j < 50; ++j) {
+	wdb.add_document(doc);
+    }
+    wdb.commit();
+
+    for (int k = 0; k < 1000; ++k) {
+	wdb.add_document(doc);
+	wdb.commit();
+    }
+
+    return true;
+}
+
+/** Check that the parent directory for the database doesn't need to be
+ *  writable.  Regression test for early versions on the glass new btree
+ *  branch which failed to append a "/" when generating a temporary filename
+ *  from the database directory.
+ */
+DEFINE_TESTCASE(readonlyparentdir1, chert || glass) {
+#if !defined __WIN32__ && !defined __CYGWIN__ && !defined __OS2__
+    string path = get_named_writable_database_path("readonlyparentdir1");
+    // Fix permissions if the previous test was killed.
+    (void)chmod(path.c_str(), 0700);
+    mkdir(path.c_str(), 0777);
+    mkdir((path + "/sub").c_str(), 0777);
+    Xapian::WritableDatabase db = get_named_writable_database("readonlyparentdir1/sub");
+    TEST(chmod(path.c_str(), 0500) == 0);
+    try {
+	Xapian::Document doc;
+	doc.add_term("hello");
+	doc.set_data("some text");
+	db.add_document(doc);
+	db.commit();
+    } catch (...) {
+	// Attempt to fix the permissions, otherwise things like "rm -rf" on
+	// the source tree will fail.
+	(void)chmod(path.c_str(), 0700);
+	throw;
+    }
+    TEST(chmod(path.c_str(), 0700) == 0);
+#endif
+    return true;
+}
+
 static void
 make_phrasebug1_db(Xapian::WritableDatabase &db, const string &)
 {
@@ -1011,8 +1132,124 @@ DEFINE_TESTCASE(phrasebug1, generated && positional) {
     return true;
 }
 
+/// Feature test for Xapian::DB_RETRY_LOCK
+DEFINE_TESTCASE(retrylock1, writable && !inmemory && !remote) {
+    // FIXME: Can't see an easy way to test this for remote databases - the
+    // harness doesn't seem to provide a suitable way to reopen a remote.
+#if defined HAVE_FORK && defined HAVE_SOCKETPAIR
+    int fds[2];
+    if (socketpair(AF_UNIX, SOCK_STREAM|SOCK_CLOEXEC, PF_UNSPEC, fds) < 0) {
+	FAIL_TEST("socketpair() failed");
+    }
+    if (fcntl(fds[1], F_SETFL, O_NONBLOCK) < 0)
+	FAIL_TEST("fcntl() failed to set O_NONBLOCK");
+    pid_t child = fork();
+    if (child == -1)
+	FAIL_TEST("fork() failed");
+    if (child == 0) {
+	// Wait for signal that parent has opened the database.
+	char ch;
+	while (read(fds[0], &ch, 1) < 0) { }
+
+	try {
+	    Xapian::WritableDatabase db2(get_named_writable_database_path("retrylock1"),
+					 Xapian::DB_OPEN|Xapian::DB_RETRY_LOCK);
+	    if (write(fds[0], "y", 1)) { }
+	} catch (const Xapian::DatabaseLockError &) {
+	    if (write(fds[0], "l", 1)) { }
+	} catch (const Xapian::Error &e) {
+	    const string & m = e.get_description();
+	    if (write(fds[0], m.data(), m.size())) { }
+	} catch (...) {
+	    if (write(fds[0], "o", 1)) { }
+	}
+	_exit(0);
+    }
+
+    close(fds[0]);
+
+    Xapian::WritableDatabase db = get_named_writable_database("retrylock1");
+    if (write(fds[1], "", 1) != 1)
+	FAIL_TEST("Failed to signal to child process");
+
+    char result[256];
+    int r = read(fds[1], result, sizeof(result));
+    if (r == -1) {
+	if (errno == EAGAIN) {
+	    // Good.
+	    result[0] = 'y';
+	} else {
+	    // Error.
+	    tout << "errno=" << errno << ": " << strerror(errno) << endl;
+	    result[0] = 'e';
+	}
+	r = 1;
+    } else if (r >= 1) {
+	if (result[0] == 'y') {
+	    // Child process managed to also get write lock!
+	    result[0] = '!';
+	}
+    } else {
+	// EOF.
+	result[0] = 'z';
+	r = 1;
+    }
+
+    try {
+	db.close();
+    } catch (...) {
+	kill(child, SIGKILL);
+	int status;
+	while (waitpid(child, &status, 0) < 0) {
+	    if (errno != EINTR) break;
+	}
+	throw;
+    }
+
+    if (result[0] == 'y') {
+	struct timeval tv;
+	tv.tv_sec = 3;
+	tv.tv_usec = 0;
+	fd_set f;
+	FD_ZERO(&f);
+	FD_SET(fds[1], &f);
+	int sr = select(fds[1] + 1, &f, NULL, &f, &tv);
+	if (sr == 0) {
+	    // Timed out.
+	    result[0] = 'T';
+	    r = 1;
+	} else {
+	    r = read(fds[1], result, sizeof(result));
+	    if (r == -1) {
+		// Error.
+		tout << "errno=" << errno << ": " << strerror(errno) << endl;
+		result[0] = 'E';
+		r = 1;
+	    } else if (r == 0) {
+		// EOF.
+		result[0] = 'Z';
+		r = 1;
+	    }
+	}
+    }
+
+    close(fds[1]);
+
+    kill(child, SIGKILL);
+    int status;
+    while (waitpid(child, &status, 0) < 0) {
+	if (errno != EINTR) break;
+    }
+
+    tout << string(result, r) << endl;
+    TEST_EQUAL(result[0], 'y');
+#endif
+
+    return true;
+}
+
 // Opening a WritableDatabase with low fds available - it should avoid them.
-DEFINE_TESTCASE(dbfilefd012, brass || chert) {
+DEFINE_TESTCASE(dbfilefd012, chert || glass) {
 #if !defined __WIN32__ && !defined __CYGWIN__ && !defined __OS2__
     int oldfds[3];
     for (int i = 0; i < 3; ++i) {
@@ -1069,7 +1306,7 @@ DEFINE_TESTCASE(dbfilefd012, brass || chert) {
 }
 
 /// Regression test for #675, fixed in 1.3.3 and 1.2.21.
-DEFINE_TESTCASE(cursorbug1, brass || chert || flint) {
+DEFINE_TESTCASE(cursorbug1, chert || glass) {
     Xapian::WritableDatabase wdb = get_writable_database();
     Xapian::Database db = get_writable_database_as_database();
     Xapian::Enquire enq(db);
@@ -1134,6 +1371,65 @@ DEFINE_TESTCASE(sortvalue2, backend) {
 DEFINE_TESTCASE(enquiregetquery1, backend) {
     Xapian::Database db = get_database("apitest_simpledata");
     Xapian::Enquire enq(db);
-    TEST_EQUAL(enq.get_query().get_description(), "Xapian::Query()");
+    TEST_EQUAL(enq.get_query().get_description(), "Query()");
+    return true;
+}
+
+DEFINE_TESTCASE(embedded1, singlefile) {
+    // In reality you should align the embedded database to a multiple of
+    // database block size, but any offset is meant to work.
+    off_t offset = 1234;
+
+    Xapian::Database db = get_database("apitest_simpledata");
+    const string & db_path = get_database_path("apitest_simpledata");
+    const string & tmp_path = db_path + "-embedded";
+    ofstream out(tmp_path, fstream::trunc|fstream::binary);
+    out.seekp(offset);
+    out << ifstream(db_path, fstream::binary).rdbuf();
+    out.close();
+
+    {
+	int fd = open(tmp_path.c_str(), O_RDONLY|O_BINARY);
+	lseek(fd, offset, SEEK_SET);
+	Xapian::Database db_embedded(fd);
+	TEST_EQUAL(db.get_doccount(), db_embedded.get_doccount());
+    }
+
+    {
+	int fd = open(tmp_path.c_str(), O_RDONLY|O_BINARY);
+	lseek(fd, offset, SEEK_SET);
+	size_t check_errors =
+	    Xapian::Database::check(fd, Xapian::DBCHECK_SHOW_STATS, &tout);
+	TEST_EQUAL(check_errors, 0);
+    }
+
+    return true;
+}
+
+/// Regression test for bug fixed in 1.3.7.
+DEFINE_TESTCASE(exactxor1, backend) {
+    Xapian::Database db = get_database("apitest_simpledata");
+    Xapian::Enquire enq(db);
+
+    const char * words[4] = { "blank", "test", "paragraph", "banana" };
+    Xapian::Query q(Xapian::Query::OP_XOR, words, words + 4);
+    enq.set_query(q);
+    enq.set_weighting_scheme(Xapian::BoolWeight());
+    Xapian::MSet mset = enq.get_mset(0, 0);
+    // A reversed conditional gave us 5 in this case.
+    TEST_EQUAL(mset.get_matches_upper_bound(), 6);
+    // Test improved lower bound in 1.3.7 (earlier versions gave 0).
+    TEST_EQUAL(mset.get_matches_lower_bound(), 2);
+
+    const char * words2[4] = { "queri", "test", "paragraph", "word" };
+    Xapian::Query q2(Xapian::Query::OP_XOR, words2, words2 + 4);
+    enq.set_query(q2);
+    enq.set_weighting_scheme(Xapian::BoolWeight());
+    mset = enq.get_mset(0, 0);
+    // A reversed conditional gave us 6 in this case.
+    TEST_EQUAL(mset.get_matches_upper_bound(), 5);
+    // Test improved lower bound in 1.3.7 (earlier versions gave 0).
+    TEST_EQUAL(mset.get_matches_lower_bound(), 1);
+
     return true;
 }
