@@ -1,7 +1,7 @@
 /** @file api_compact.cc
- * @brief Tests of xapian-compact.
+ * @brief Tests of Database::compact()
  */
-/* Copyright (C) 2009,2010,2011,2013 Olly Betts
+/* Copyright (C) 2009,2010,2011,2012,2013,2015,2016 Olly Betts
  * Copyright (C) 2010 Richard Boulton
  *
  * This program is free software; you can redistribute it and/or
@@ -24,18 +24,23 @@
 
 #include "api_compact.h"
 
+#include <xapian.h>
+
 #include "apitest.h"
 #include "dbcheck.h"
+#include "filetests.h"
+#include "str.h"
 #include "testsuite.h"
 #include "testutils.h"
 
-#include <xapian.h>
-
 #include <cstdlib>
 #include <fstream>
 
-#include "str.h"
-#include "utils.h"
+#include <sys/types.h>
+#include "safesysstat.h"
+#include "safefcntl.h"
+#include "safeunistd.h"
+
 #include "unixcmds.h"
 
 using namespace std;
@@ -113,21 +118,13 @@ DEFINE_TESTCASE(compactnorenumber1, generated) {
 
     rm_rf(out);
     {
-	Xapian::Compactor compact;
-	compact.set_renumber(false);
-	compact.set_destdir(out);
-	compact.add_source(a);
-	compact.compact();
+	Xapian::Database db(a);
+	db.compact(out, Xapian::DBCOMPACT_NO_RENUMBER);
     }
 
     check_sparse_uid_terms(out);
 
     {
-	if (get_dbtype() == "flint") {
-	    // Flint creates its uuid file if it doesn't already exist, so we
-	    // need a white box test to ensure xapian-compact created it.
-	    TEST(file_exists(out + "/uuid"));
-	}
 	TEST(!dir_exists(out + "/donor"));
 	Xapian::Database db(out);
 	// xapian-compact should change the UUID of the database, but didn't
@@ -143,12 +140,10 @@ DEFINE_TESTCASE(compactnorenumber1, generated) {
 
     rm_rf(out);
     {
-	Xapian::Compactor compact;
-	compact.set_renumber(false);
-	compact.set_destdir(out);
-	compact.add_source(a);
-	compact.add_source(c);
-	compact.compact();
+	Xapian::Database db;
+	db.add_database(Xapian::Database(a));
+	db.add_database(Xapian::Database(c));
+	db.compact(out, Xapian::DBCOMPACT_NO_RENUMBER);
     }
     check_sparse_uid_terms(out);
     {
@@ -161,100 +156,97 @@ DEFINE_TESTCASE(compactnorenumber1, generated) {
 
     rm_rf(out);
     {
-	Xapian::Compactor compact;
-	compact.set_renumber(false);
-	compact.set_destdir(out);
-	compact.add_source(d);
-	compact.add_source(a);
-	compact.add_source(c);
-	compact.compact();
+	Xapian::Database db;
+	db.add_database(Xapian::Database(d));
+	db.add_database(Xapian::Database(a));
+	db.add_database(Xapian::Database(c));
+	db.compact(out, Xapian::DBCOMPACT_NO_RENUMBER);
     }
     check_sparse_uid_terms(out);
 
     rm_rf(out);
     {
-	Xapian::Compactor compact;
-	compact.set_renumber(false);
-	compact.set_destdir(out);
-	compact.add_source(c);
-	compact.add_source(a);
-	compact.add_source(d);
-	compact.compact();
+	Xapian::Database db;
+	db.add_database(Xapian::Database(c));
+	db.add_database(Xapian::Database(a));
+	db.add_database(Xapian::Database(d));
+	db.compact(out, Xapian::DBCOMPACT_NO_RENUMBER);
     }
     check_sparse_uid_terms(out);
 
     // Should fail.
     rm_rf(out);
     {
-	Xapian::Compactor compact;
-	compact.set_renumber(false);
-	compact.set_destdir(out);
-	compact.add_source(a);
-	compact.add_source(b);
-	TEST_EXCEPTION(Xapian::InvalidOperationError, compact.compact());
+	Xapian::Database db;
+	db.add_database(Xapian::Database(a));
+	db.add_database(Xapian::Database(b));
+	TEST_EXCEPTION(Xapian::InvalidOperationError,
+	    db.compact(out, Xapian::DBCOMPACT_NO_RENUMBER)
+	);
     }
  
     // Should fail.
     rm_rf(out);
     {
-	Xapian::Compactor compact;
-	compact.set_renumber(false);
-	compact.set_destdir(out);
-	compact.add_source(b);
-	compact.add_source(a);
-	TEST_EXCEPTION(Xapian::InvalidOperationError, compact.compact());
+	Xapian::Database db;
+	db.add_database(Xapian::Database(b));
+	db.add_database(Xapian::Database(a));
+	TEST_EXCEPTION(Xapian::InvalidOperationError,
+	    db.compact(out, Xapian::DBCOMPACT_NO_RENUMBER)
+	);
     }
 
     // Should fail.
     rm_rf(out);
     {
-	Xapian::Compactor compact;
-	compact.set_renumber(false);
-	compact.set_destdir(out);
-	compact.add_source(a);
-	compact.add_source(b);
-	compact.add_source(d);
-	TEST_EXCEPTION(Xapian::InvalidOperationError, compact.compact());
+	Xapian::Database db;
+	db.add_database(Xapian::Database(a));
+	db.add_database(Xapian::Database(b));
+	db.add_database(Xapian::Database(d));
+	TEST_EXCEPTION(Xapian::InvalidOperationError,
+	    db.compact(out, Xapian::DBCOMPACT_NO_RENUMBER)
+	);
     }
  
     // Should fail.
     rm_rf(out);
     {
-	Xapian::Compactor compact;
-	compact.set_renumber(false);
-	compact.set_destdir(out);
-	compact.add_source(d);
-	compact.add_source(b);
-	compact.add_source(a);
-	TEST_EXCEPTION(Xapian::InvalidOperationError, compact.compact());
+	Xapian::Database db;
+	db.add_database(Xapian::Database(d));
+	db.add_database(Xapian::Database(b));
+	db.add_database(Xapian::Database(a));
+	TEST_EXCEPTION(Xapian::InvalidOperationError,
+	    db.compact(out, Xapian::DBCOMPACT_NO_RENUMBER)
+	);
     }
 
     // Should fail.
     rm_rf(out);
     {
-	Xapian::Compactor compact;
-	compact.set_renumber(false);
-	compact.set_destdir(out);
-	compact.add_source(b);
-	compact.add_source(a);
-	compact.add_source(d);
-	TEST_EXCEPTION(Xapian::InvalidOperationError, compact.compact());
+	Xapian::Database db;
+	db.add_database(Xapian::Database(b));
+	db.add_database(Xapian::Database(a));
+	db.add_database(Xapian::Database(d));
+	TEST_EXCEPTION(Xapian::InvalidOperationError,
+	    db.compact(out, Xapian::DBCOMPACT_NO_RENUMBER)
+	);
     }
 
     return true;
 }
 
 // Test use of compact to merge two databases.
-DEFINE_TESTCASE(compactmerge1, brass || chert || flint) {
+DEFINE_TESTCASE(compactmerge1, chert || glass) {
     string indbpath = get_database_path("apitest_simpledata");
     string outdbpath = get_named_writable_database_path("compactmerge1out");
     rm_rf(outdbpath);
 
-    Xapian::Compactor compact;
-    compact.set_destdir(outdbpath);
-    compact.add_source(indbpath);
-    compact.add_source(indbpath);
-    compact.compact();
+    {
+	Xapian::Database db;
+	db.add_database(Xapian::Database(indbpath));
+	db.add_database(Xapian::Database(indbpath));
+	db.compact(outdbpath);
+    }
 
     Xapian::Database indb(get_database("apitest_simpledata"));
     Xapian::Database outdb(outdbpath);
@@ -288,10 +280,10 @@ DEFINE_TESTCASE(compactmultichunks1, generated) {
     string outdbpath = get_named_writable_database_path("compactmultichunks1out");
     rm_rf(outdbpath);
 
-    Xapian::Compactor compact;
-    compact.set_destdir(outdbpath);
-    compact.add_source(indbpath);
-    compact.compact();
+    {
+	Xapian::Database db(indbpath);
+	db.compact(outdbpath);
+    }
 
     Xapian::Database indb(indbpath);
     Xapian::Database outdb(outdbpath);
@@ -303,7 +295,7 @@ DEFINE_TESTCASE(compactmultichunks1, generated) {
 }
 
 // Test compacting from a stub database directory.
-DEFINE_TESTCASE(compactstub1, brass || chert || flint) {
+DEFINE_TESTCASE(compactstub1, chert || glass) {
     const char * stubpath = ".stub/compactstub1";
     const char * stubpathfile = ".stub/compactstub1/XAPIANDB";
     mkdir(".stub", 0755);
@@ -317,10 +309,10 @@ DEFINE_TESTCASE(compactstub1, brass || chert || flint) {
     string outdbpath = get_named_writable_database_path("compactstub1out");
     rm_rf(outdbpath);
 
-    Xapian::Compactor compact;
-    compact.set_destdir(outdbpath);
-    compact.add_source(stubpath);
-    compact.compact();
+    {
+	Xapian::Database db(stubpath);
+	db.compact(outdbpath);
+    }
 
     Xapian::Database indb(stubpath);
     Xapian::Database outdb(outdbpath);
@@ -332,7 +324,7 @@ DEFINE_TESTCASE(compactstub1, brass || chert || flint) {
 }
 
 // Test compacting from a stub database file.
-DEFINE_TESTCASE(compactstub2, brass || chert || flint) {
+DEFINE_TESTCASE(compactstub2, chert || glass) {
     const char * stubpath = ".stub/compactstub2";
     mkdir(".stub", 0755);
     ofstream stub(stubpath);
@@ -344,10 +336,10 @@ DEFINE_TESTCASE(compactstub2, brass || chert || flint) {
     string outdbpath = get_named_writable_database_path("compactstub2out");
     rm_rf(outdbpath);
 
-    Xapian::Compactor compact;
-    compact.set_destdir(outdbpath);
-    compact.add_source(stubpath);
-    compact.compact();
+    {
+	Xapian::Database db(stubpath);
+	db.compact(outdbpath);
+    }
 
     Xapian::Database indb(stubpath);
     Xapian::Database outdb(outdbpath);
@@ -359,7 +351,7 @@ DEFINE_TESTCASE(compactstub2, brass || chert || flint) {
 }
 
 // Test compacting a stub database file to itself.
-DEFINE_TESTCASE(compactstub3, brass || chert || flint) {
+DEFINE_TESTCASE(compactstub3, chert || glass) {
     const char * stubpath = ".stub/compactstub3";
     mkdir(".stub", 0755);
     ofstream stub(stubpath);
@@ -372,13 +364,9 @@ DEFINE_TESTCASE(compactstub3, brass || chert || flint) {
     {
 	Xapian::Database indb(stubpath);
 	in_docs = indb.get_doccount();
+	indb.compact(stubpath);
     }
 
-    Xapian::Compactor compact;
-    compact.set_destdir(stubpath);
-    compact.add_source(stubpath);
-    compact.compact();
-
     Xapian::Database outdb(stubpath);
 
     TEST_EQUAL(in_docs, outdb.get_doccount());
@@ -388,7 +376,7 @@ DEFINE_TESTCASE(compactstub3, brass || chert || flint) {
 }
 
 // Test compacting a stub database directory to itself.
-DEFINE_TESTCASE(compactstub4, brass || chert || flint) {
+DEFINE_TESTCASE(compactstub4, chert || glass) {
     const char * stubpath = ".stub/compactstub4";
     const char * stubpathfile = ".stub/compactstub4/XAPIANDB";
     mkdir(".stub", 0755);
@@ -403,13 +391,9 @@ DEFINE_TESTCASE(compactstub4, brass || chert || flint) {
     {
 	Xapian::Database indb(stubpath);
 	in_docs = indb.get_doccount();
+	indb.compact(stubpath);
     }
 
-    Xapian::Compactor compact;
-    compact.set_destdir(stubpath);
-    compact.add_source(stubpath);
-    compact.compact();
-
     Xapian::Database outdb(stubpath);
 
     TEST_EQUAL(in_docs, outdb.get_doccount());
@@ -425,6 +409,7 @@ make_all_tables(Xapian::WritableDatabase &db, const string &)
     doc.add_term("foo");
     db.add_document(doc);
     db.add_spelling("foo");
+    db.add_synonym("bar", "pub");
     db.add_synonym("foobar", "foo");
 
     db.commit();
@@ -449,11 +434,12 @@ DEFINE_TESTCASE(compactmissingtables1, generated) {
     string out = get_named_writable_database_path("compactmissingtables1out");
     rm_rf(out);
 
-    Xapian::Compactor compact;
-    compact.set_destdir(out);
-    compact.add_source(a);
-    compact.add_source(b);
-    compact.compact();
+    {
+	Xapian::Database db;
+	db.add_database(Xapian::Database(a));
+	db.add_database(Xapian::Database(b));
+	db.compact(out);
+    }
 
     {
 	Xapian::Database db(out);
@@ -490,11 +476,12 @@ DEFINE_TESTCASE(compactmergesynonym1, generated) {
     string out = get_named_writable_database_path("compactmergesynonym1out");
     rm_rf(out);
 
-    Xapian::Compactor compact;
-    compact.set_destdir(out);
-    compact.add_source(a);
-    compact.add_source(b);
-    compact.compact();
+    {
+	Xapian::Database db;
+	db.add_database(Xapian::Database(a));
+	db.add_database(Xapian::Database(b));
+	db.compact(out);
+    }
 
     {
 	Xapian::Database db(out);
@@ -527,17 +514,16 @@ DEFINE_TESTCASE(compactmergesynonym1, generated) {
     return true;
 }
 
-DEFINE_TESTCASE(compactempty1, brass || chert) {
+DEFINE_TESTCASE(compactempty1, chert || glass) {
     string empty_dbpath = get_database_path(string());
     string outdbpath = get_named_writable_database_path("compactempty1out");
     rm_rf(outdbpath);
 
     {
 	// Compacting an empty database tried to divide by zero in 1.3.0.
-	Xapian::Compactor compact;
-	compact.set_destdir(outdbpath);
-	compact.add_source(empty_dbpath);
-	compact.compact();
+	Xapian::Database db;
+	db.add_database(Xapian::Database(empty_dbpath));
+	db.compact(outdbpath);
 
 	Xapian::Database outdb(outdbpath);
 	TEST_EQUAL(outdb.get_doccount(), 0);
@@ -546,11 +532,10 @@ DEFINE_TESTCASE(compactempty1, brass || chert) {
 
     {
 	// Check compacting two empty databases together.
-	Xapian::Compactor compact;
-	compact.set_destdir(outdbpath);
-	compact.add_source(empty_dbpath);
-	compact.add_source(empty_dbpath);
-	compact.compact();
+	Xapian::Database db;
+	db.add_database(Xapian::Database(empty_dbpath));
+	db.add_database(Xapian::Database(empty_dbpath));
+	db.compact(outdbpath);
 
 	Xapian::Database outdb(outdbpath);
 	TEST_EQUAL(outdb.get_doccount(), 0);
@@ -560,7 +545,7 @@ DEFINE_TESTCASE(compactempty1, brass || chert) {
     return true;
 }
 
-DEFINE_TESTCASE(compactmultipass1, brass || chert) {
+DEFINE_TESTCASE(compactmultipass1, chert || glass) {
     string outdbpath = get_named_writable_database_path("compactmultipass1");
     rm_rf(outdbpath);
 
@@ -573,17 +558,108 @@ DEFINE_TESTCASE(compactmultipass1, brass || chert) {
     string d = get_database_path("compactnorenumber1d", make_sparse_db,
 				 "3000 999999 !999999");
 
-    Xapian::Compactor compact;
-    compact.set_destdir(outdbpath);
-    compact.add_source(a);
-    compact.add_source(b);
-    compact.add_source(c);
-    compact.add_source(d);
-    compact.set_multipass(true);
-    compact.compact();
+    {
+	Xapian::Database db;
+	db.add_database(Xapian::Database(a));
+	db.add_database(Xapian::Database(b));
+	db.add_database(Xapian::Database(c));
+	db.add_database(Xapian::Database(d));
+	db.compact(outdbpath, Xapian::DBCOMPACT_MULTIPASS);
+    }
 
     Xapian::Database outdb(outdbpath);
     dbcheck(outdb, 29, 1041);
 
     return true;
 }
+
+// Test compacting to an fd.
+DEFINE_TESTCASE(compacttofd1, glass) {
+    Xapian::Database indb(get_database("apitest_simpledata"));
+    string outdbpath = get_named_writable_database_path("compacttofd1out");
+    rm_rf(outdbpath);
+
+    int fd = open(outdbpath.c_str(), O_CREAT|O_RDWR, 0666);
+    TEST(fd != -1);
+    indb.compact(fd);
+
+    // Confirm that the fd was closed by Xapian.
+    TEST(close(fd) == -1);
+    TEST(errno == EBADF);
+
+    Xapian::Database outdb(outdbpath);
+
+    TEST_EQUAL(indb.get_doccount(), outdb.get_doccount());
+    dbcheck(outdb, outdb.get_doccount(), outdb.get_doccount());
+
+    return true;
+}
+
+// Test compacting to an fd at at offset.
+DEFINE_TESTCASE(compacttofd2, glass) {
+    Xapian::Database indb(get_database("apitest_simpledata"));
+    string outdbpath = get_named_writable_database_path("compacttofd2out");
+    rm_rf(outdbpath);
+
+    int fd = open(outdbpath.c_str(), O_CREAT|O_RDWR, 0666);
+    TEST(fd != -1);
+    TEST(lseek(fd, 8192, SEEK_SET) == 8192);
+    indb.compact(fd);
+
+    // Confirm that the fd was closed by Xapian.
+    TEST(close(fd) == -1);
+    TEST(errno == EBADF);
+
+    fd = open(outdbpath.c_str(), O_RDONLY, 0666);
+    TEST(fd != -1);
+
+    // Test that the database wasn't just written to the start of the file.
+    char buf[8192];
+    size_t n = sizeof(buf);
+    while (n) {
+	ssize_t c = read(fd, buf, n);
+	TEST(c > 0);
+	for (const char * p = buf; p != buf + c; ++p) {
+	    TEST(*p == 0);
+	}
+	n -= c;
+    }
+
+    TEST(lseek(fd, 8192, SEEK_SET) == 8192);
+    Xapian::Database outdb(fd);
+
+    TEST_EQUAL(indb.get_doccount(), outdb.get_doccount());
+    dbcheck(outdb, outdb.get_doccount(), outdb.get_doccount());
+
+    return true;
+}
+
+// Regression test for bug fixed in 1.3.5.  If you commit a WritableDatabase
+// with uncommitted changes, you get an inconsistent output.
+DEFINE_TESTCASE(compactsingle1, glass) {
+    Xapian::WritableDatabase db = get_writable_database();
+    Xapian::Document doc;
+    doc.add_term("foo");
+    doc.add_term("bar");
+    doc.add_term("baz");
+    db.add_document(doc);
+
+    string output = ".glass/db__compactsingle1-out";
+    // In 1.3.4, we would hang if the output file already existed, so check
+    // that works.
+    touch(output);
+
+    TEST_EXCEPTION(Xapian::InvalidOperationError,
+	db.compact(output, Xapian::DBCOMPACT_SINGLE_FILE));
+
+    // Check the file wasn't removed by the failed attempt.
+    TEST(file_exists(output));
+
+    db.commit();
+    db.compact(output, Xapian::DBCOMPACT_SINGLE_FILE);
+    db.close();
+
+    TEST_EQUAL(Xapian::Database::check(output, 0, &tout), 0);
+
+    return true;
+}
