@@ -2,7 +2,7 @@
  *
  * Copyright 1999,2000,2001 BrightStation PLC
  * Copyright 2002 Ananova Ltd
- * Copyright 2002,2003,2004,2005,2006,2007,2009 Olly Betts
+ * Copyright 2002,2003,2004,2005,2006,2007,2009,2016 Olly Betts
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -156,7 +156,7 @@ DEFINE_TESTCASE(near1, positional) {
     q = Xapian::Query(Xapian::Query::OP_NEAR, subqs.begin(), subqs.end(), 8);
     enquire.set_query(q);
 
-    // retrieve the top ten results
+    // retrieve the top twenty results
     mymset = enquire.get_mset(0, 20);
     mset_expect_order(mymset, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14);
 
@@ -168,7 +168,7 @@ DEFINE_TESTCASE(near1, positional) {
     q = Xapian::Query(Xapian::Query::OP_NEAR, subqs.begin(), subqs.end(), 999999999);
     enquire.set_query(q);
 
-    // retrieve the top ten results
+    // retrieve the top twenty results
     mymset = enquire.get_mset(0, 20);
     mset_expect_order(mymset, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14);
 
@@ -181,6 +181,7 @@ DEFINE_TESTCASE(near2, positional) {
     Xapian::Enquire enquire(mydb);
     Xapian::Stem stemmer("english");
     enquire.set_weighting_scheme(Xapian::BoolWeight());
+    Xapian::MSet mymset;
 
     // make a query
     vector<Xapian::Query> subqs;
@@ -189,11 +190,14 @@ DEFINE_TESTCASE(near2, positional) {
 			    Xapian::Query(stemmer("phrase")),
 			    Xapian::Query(stemmer("near"))));
     subqs.push_back(Xapian::Query(stemmer("and")));
-    q = Xapian::Query(Xapian::Query::OP_NEAR, subqs.begin(), subqs.end(), 2);
-    enquire.set_query(q);
+    TEST_EXCEPTION(Xapian::UnimplementedError,
+	q = Xapian::Query(q.OP_NEAR, subqs.begin(), subqs.end(), 2);
+	enquire.set_query(q);
 
-    // retrieve the top ten results
-    Xapian::MSet mymset = enquire.get_mset(0, 10);
+	// retrieve the top ten results
+	mymset = enquire.get_mset(0, 10)
+    );
+#if 0 // Disable until we reimplement this.
     mset_expect_order(mymset, 1);
 
     subqs.clear();
@@ -219,6 +223,7 @@ DEFINE_TESTCASE(near2, positional) {
     // retrieve the top ten results
     mymset = enquire.get_mset(0, 10);
     mset_expect_order(mymset, 2);
+#endif
 
     return true;
 }
@@ -417,6 +422,7 @@ DEFINE_TESTCASE(phrase2, positional) {
     Xapian::Enquire enquire(mydb);
     Xapian::Stem stemmer("english");
     enquire.set_weighting_scheme(Xapian::BoolWeight());
+    Xapian::MSet mymset;
 
     // make a query
     vector<Xapian::Query> subqs;
@@ -425,11 +431,14 @@ DEFINE_TESTCASE(phrase2, positional) {
 			    Xapian::Query(stemmer("phrase")),
 			    Xapian::Query(stemmer("near"))));
     subqs.push_back(Xapian::Query(stemmer("and")));
-    q = Xapian::Query(Xapian::Query::OP_PHRASE, subqs.begin(), subqs.end(), 2);
-    enquire.set_query(q);
+    TEST_EXCEPTION(Xapian::UnimplementedError,
+	q = Xapian::Query(q.OP_PHRASE, subqs.begin(), subqs.end(), 2);
+	enquire.set_query(q);
 
-    // retrieve the top ten results
-    Xapian::MSet mymset = enquire.get_mset(0, 10);
+	// retrieve the top ten results
+	mymset = enquire.get_mset(0, 10)
+    );
+#if 0 // Disable until we reimplement this.
     mset_expect_order(mymset);
 
     subqs.clear();
@@ -455,6 +464,7 @@ DEFINE_TESTCASE(phrase2, positional) {
     // retrieve the top ten results
     mymset = enquire.get_mset(0, 10);
     mset_expect_order(mymset);
+#endif
 
     return true;
 }
