@@ -1,7 +1,7 @@
 /** @file api_unicode.cc
  * @brief Test the Unicode and UTF-8 classes and functions.
  */
-/* Copyright (C) 2006,2007,2008,2009,2010,2013 Olly Betts
+/* Copyright (C) 2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2016 Olly Betts
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -152,6 +152,26 @@ DEFINE_TESTCASE(unicode1,!backend) {
     TEST_EQUAL(Unicode::get_category(0xa3), Unicode::CURRENCY_SYMBOL);
     // U+0242 was added in Unicode 5.0.0.
     TEST_EQUAL(Unicode::get_category(0x242), Unicode::LOWERCASE_LETTER);
+    // U+0526 was added in Unicode 6.0.0.
+    TEST_EQUAL(Unicode::get_category(0x0526), Unicode::UPPERCASE_LETTER);
+    // U+0527 was added in Unicode 6.0.0.
+    TEST_EQUAL(Unicode::get_category(0x0527), Unicode::LOWERCASE_LETTER);
+    // U+0620 was added in Unicode 6.0.0.
+    TEST_EQUAL(Unicode::get_category(0x0620), Unicode::OTHER_LETTER);
+    // U+065F was added in Unicode 6.0.0.
+    TEST_EQUAL(Unicode::get_category(0x065F), Unicode::NON_SPACING_MARK);
+    // U+06DE changed category in Unicode 6.0.0.
+    TEST_EQUAL(Unicode::get_category(0x06DE), Unicode::OTHER_SYMBOL);
+    // U+0840 was added in Unicode 6.0.0.
+    TEST_EQUAL(Unicode::get_category(0x0840), Unicode::OTHER_LETTER);
+    // U+093A was added in Unicode 6.0.0.
+    TEST_EQUAL(Unicode::get_category(0x093A), Unicode::NON_SPACING_MARK);
+    // U+093B was added in Unicode 6.0.0.
+    TEST_EQUAL(Unicode::get_category(0x093B), Unicode::COMBINING_SPACING_MARK);
+    // U+0CF1 changed category in Unicode 6.0.0.
+    TEST_EQUAL(Unicode::get_category(0x0CF1), Unicode::OTHER_LETTER);
+    // U+0CF2 changed category in Unicode 6.0.0.
+    TEST_EQUAL(Unicode::get_category(0x0CF2), Unicode::OTHER_LETTER);
     // U+11A7 was added in Unicode 5.2.0.
     TEST_EQUAL(Unicode::get_category(0x11A7), Unicode::OTHER_LETTER);
     // U+9FCB was added in Unicode 5.2.0.
@@ -166,48 +186,96 @@ DEFINE_TESTCASE(unicode1,!backend) {
     // U+1109A was added in Unicode 5.2.0.
     TEST_EQUAL(Unicode::get_category(0x1109a), Unicode::OTHER_LETTER);
     // U+1F773 was added in Unicode 6.0.0.
-    TEST_EQUAL(Unicode::get_category(0x1F773), Unicode::UNASSIGNED);
+    TEST_EQUAL(Unicode::get_category(0x1F773), Unicode::OTHER_SYMBOL);
     // U+2B740 was added in Unicode 6.0.0.
-    TEST_EQUAL(Unicode::get_category(0x2B740), Unicode::UNASSIGNED);
+    TEST_EQUAL(Unicode::get_category(0x2B740), Unicode::OTHER_LETTER);
     // U+2B81D was added in Unicode 6.0.0.
-    TEST_EQUAL(Unicode::get_category(0x2B81D), Unicode::UNASSIGNED);
+    TEST_EQUAL(Unicode::get_category(0x2B81D), Unicode::OTHER_LETTER);
     // U+00A7 changed category in Unicode 6.1.0 (was OTHER_SYMBOL).
-    TEST_EQUAL(Unicode::get_category(0xA7), Unicode::OTHER_SYMBOL);
+    TEST_EQUAL(Unicode::get_category(0xA7), Unicode::OTHER_PUNCTUATION);
     // U+00AA changed category in Unicode 6.1.0 (was LOWERCASE_LETTER).
-    TEST_EQUAL(Unicode::get_category(0xAA), Unicode::LOWERCASE_LETTER);
+    TEST_EQUAL(Unicode::get_category(0xAA), Unicode::OTHER_LETTER);
     // U+00B6 changed category in Unicode 6.1.0 (was OTHER_SYMBOL).
-    TEST_EQUAL(Unicode::get_category(0xB6), Unicode::OTHER_SYMBOL);
+    TEST_EQUAL(Unicode::get_category(0xB6), Unicode::OTHER_PUNCTUATION);
     // U+00BA changed category in Unicode 6.1.0 (was LOWERCASE_LETTER).
-    TEST_EQUAL(Unicode::get_category(0xBA), Unicode::LOWERCASE_LETTER);
+    TEST_EQUAL(Unicode::get_category(0xBA), Unicode::OTHER_LETTER);
     // U+058F was added in Unicode 6.1.0.
-    TEST_EQUAL(Unicode::get_category(0x058F), Unicode::UNASSIGNED);
+    TEST_EQUAL(Unicode::get_category(0x058F), Unicode::CURRENCY_SYMBOL);
     // U+0604 was added in Unicode 6.1.0.
-    TEST_EQUAL(Unicode::get_category(0x0604), Unicode::UNASSIGNED);
+    TEST_EQUAL(Unicode::get_category(0x0604), Unicode::FORMAT);
     // U+08A0 was added in Unicode 6.1.0.
-    TEST_EQUAL(Unicode::get_category(0x08A0), Unicode::UNASSIGNED);
+    TEST_EQUAL(Unicode::get_category(0x08A0), Unicode::OTHER_LETTER);
     // U+08E4 was added in Unicode 6.1.0.
-    TEST_EQUAL(Unicode::get_category(0x08E4), Unicode::UNASSIGNED);
+    TEST_EQUAL(Unicode::get_category(0x08E4), Unicode::NON_SPACING_MARK);
     // U+0AF0 was added in Unicode 6.1.0.
-    TEST_EQUAL(Unicode::get_category(0x0AF0), Unicode::UNASSIGNED);
+    TEST_EQUAL(Unicode::get_category(0x0AF0), Unicode::OTHER_PUNCTUATION);
     // U+9FCC was added in Unicode 6.1.0.
-    TEST_EQUAL(Unicode::get_category(0x9FCC), Unicode::UNASSIGNED);
+    TEST_EQUAL(Unicode::get_category(0x9FCC), Unicode::OTHER_LETTER);
     // U+A7F9 was added in Unicode 6.1.0.
-    TEST_EQUAL(Unicode::get_category(0xA7F9), Unicode::UNASSIGNED);
+    TEST_EQUAL(Unicode::get_category(0xA7F9), Unicode::MODIFIER_LETTER);
     // U+110F0 was added in Unicode 6.1.0.
-    TEST_EQUAL(Unicode::get_category(0x110F0), Unicode::UNASSIGNED);
+    TEST_EQUAL(Unicode::get_category(0x110F0), Unicode::DECIMAL_DIGIT_NUMBER);
     // U+11100 was added in Unicode 6.1.0.
-    TEST_EQUAL(Unicode::get_category(0x11100), Unicode::UNASSIGNED);
+    TEST_EQUAL(Unicode::get_category(0x11100), Unicode::NON_SPACING_MARK);
     // U+1EEF0 was added in Unicode 6.1.0.
-    TEST_EQUAL(Unicode::get_category(0x1EEF0), Unicode::UNASSIGNED);
+    TEST_EQUAL(Unicode::get_category(0x1EEF0), Unicode::MATH_SYMBOL);
     // U+1F634 was added in Unicode 6.1.0.
-    TEST_EQUAL(Unicode::get_category(0x1F634), Unicode::UNASSIGNED);
+    TEST_EQUAL(Unicode::get_category(0x1F634), Unicode::OTHER_SYMBOL);
     // U+20BA was added in Unicode 6.2.0.
-    TEST_EQUAL(Unicode::get_category(0x20BA), Unicode::UNASSIGNED);
+    TEST_EQUAL(Unicode::get_category(0x20BA), Unicode::CURRENCY_SYMBOL);
     // U+061C was added in Unicode 6.3.0.
-    TEST_EQUAL(Unicode::get_category(0x61C), Unicode::UNASSIGNED);
-    // U+037F is due to become "GREEK CAPITAL LETTER YOT" in Unicode 7.0.0, but
-    // currently should be unassigned.
-    TEST_EQUAL(Unicode::get_category(0x37F), Unicode::UNASSIGNED);
+    TEST_EQUAL(Unicode::get_category(0x61C), Unicode::FORMAT);
+    // U+037F "GREEK CAPITAL LETTER YOT" was added in Unicode 7.0.0.
+    TEST_EQUAL(Unicode::get_category(0x37F), Unicode::UPPERCASE_LETTER);
+
+    // Added or changed in Unicode 8.0.0:
+    // U+08B3 "ARABIC LETTER AIN WITH THREE DOTS BELOW".
+    TEST_EQUAL(Unicode::get_category(0x8B3), Unicode::OTHER_LETTER);
+    // U+0AF9 "GUJARATI LETTER ZHA".
+    TEST_EQUAL(Unicode::get_category(0xAF9), Unicode::OTHER_LETTER);
+    // U+0C5A "TELUGU LETTER RRRA".
+    TEST_EQUAL(Unicode::get_category(0xC5A), Unicode::OTHER_LETTER);
+    // U+0D5F "MALAYALAM LETTER ARCHAIC II".
+    TEST_EQUAL(Unicode::get_category(0xD5F), Unicode::OTHER_LETTER);
+    // U+13F5 "CHEROKEE LETTER MV".
+    TEST_EQUAL(Unicode::get_category(0x13F5), Unicode::UPPERCASE_LETTER);
+    // U+13F8 "CHEROKEE SMALL LETTER YE".
+    TEST_EQUAL(Unicode::get_category(0x13F8), Unicode::LOWERCASE_LETTER);
+    // U+19B7 "NEW TAI LUE VOWEL SIGN O" changed to be OTHER_LETTER in 8.0.0.
+    TEST_EQUAL(Unicode::get_category(0x19B7), Unicode::OTHER_LETTER);
+    // U+20BE "LARI SIGN".
+    TEST_EQUAL(Unicode::get_category(0x20BE), Unicode::CURRENCY_SYMBOL);
+    // U+218A "TURNED DIGIT TWO".
+    TEST_EQUAL(Unicode::get_category(0x218A), Unicode::OTHER_SYMBOL);
+    // U+10C9C "OLD HUNGARIAN CAPITAL LETTER OO".
+    TEST_EQUAL(Unicode::get_category(0x10C9C), Unicode::UPPERCASE_LETTER);
+    // U+12399 "CUNEIFORM SIGN U U".
+    TEST_EQUAL(Unicode::get_category(0x12399), Unicode::OTHER_LETTER);
+    // U+1D800 "SIGNWRITING HAND-FIST INDEX".
+    TEST_EQUAL(Unicode::get_category(0x1D800), Unicode::OTHER_SYMBOL);
+
+    // Added or changed in Unicode 9.0.0:
+    // U+08B6 "ARABIC LETTER BEH WITH SMALL MEEM ABOVE"
+    TEST_EQUAL(Unicode::get_category(0x8B6), Unicode::OTHER_LETTER);
+    // U+08E2 "ARABIC DISPUTED END OF AYAH"
+    TEST_EQUAL(Unicode::get_category(0x8E2), Unicode::FORMAT);
+    // U+0C80 "KANNADA SIGN SPACING CANDRABINDU"
+    TEST_EQUAL(Unicode::get_category(0xC80), Unicode::OTHER_LETTER);
+    // U+0D56 "MALAYALAM LETTER CHILLU LLL"
+    TEST_EQUAL(Unicode::get_category(0xD56), Unicode::OTHER_LETTER);
+    // U+0D58 "MALAYALAM FRACTION ONE ONE-HUNDRED-AND-SIXTIETH"
+    TEST_EQUAL(Unicode::get_category(0xD58), Unicode::OTHER_NUMBER);
+    // U+1885 "MONGOLIAN LETTER ALI GALI BALUDA"
+    TEST_EQUAL(Unicode::get_category(0x1885), Unicode::NON_SPACING_MARK);
+    // U+1886 "MONGOLIAN LETTER ALI GALI THREE BALUDA"
+    TEST_EQUAL(Unicode::get_category(0x1886), Unicode::NON_SPACING_MARK);
+    // U+104FB "OSAGE SMALL LETTER ZHA"
+    TEST_EQUAL(Unicode::get_category(0x104FB), Unicode::LOWERCASE_LETTER);
+    // U+1141F "NEWA LETTER TA"
+    TEST_EQUAL(Unicode::get_category(0x1141F), Unicode::OTHER_LETTER);
+    // U+1F989 "OWL"
+    TEST_EQUAL(Unicode::get_category(0x1F989), Unicode::OTHER_SYMBOL);
+
     // Test some invalid Unicode values.
     TEST_EQUAL(Unicode::get_category(0x110000), Unicode::UNASSIGNED);
     TEST_EQUAL(Unicode::get_category(0xFFFFFFFF), Unicode::UNASSIGNED);
@@ -263,7 +331,7 @@ DEFINE_TESTCASE(caseconvert1,!backend) {
     return true;
 }
 
-/// Test Unicode 5.1 support.
+/// Test Unicode 5.1 and later support.
 DEFINE_TESTCASE(caseconvert2,!backend) {
     using namespace Xapian;
 
@@ -277,6 +345,18 @@ DEFINE_TESTCASE(caseconvert2,!backend) {
     TEST_EQUAL(Unicode::get_category(0x5be), Unicode::DASH_PUNCTUATION);
     TEST_EQUAL(Unicode::get_category(0x1f093), Unicode::OTHER_SYMBOL);
 
+    // U+0526, U+0527 and U+A78D were added in Unicode 6.0.0:
+    TEST_EQUAL(Unicode::toupper(0x265), 0xa78d);
+    TEST_EQUAL(Unicode::tolower(0xa78d), 0x265);
+    TEST_EQUAL(Unicode::tolower(0x526), 0x527);
+    TEST_EQUAL(Unicode::toupper(0x527), 0x526);
+
+    // U+A7AA was added in Unicode 6.1.0:
+    TEST_EQUAL(Unicode::toupper(0x266), 0xa7aa);
+    TEST_EQUAL(Unicode::tolower(0xa7aa), 0x266);
+    TEST_EQUAL(Unicode::tolower(0x526), 0x527);
+    TEST_EQUAL(Unicode::toupper(0x527), 0x526);
+
     TEST_EQUAL(Unicode::tolower(0x370), 0x371);
     TEST_EQUAL(Unicode::toupper(0x371), 0x370);
     TEST_EQUAL(Unicode::tolower(0x372), 0x373);
@@ -286,6 +366,14 @@ DEFINE_TESTCASE(caseconvert2,!backend) {
     TEST_EQUAL(Unicode::tolower(0x3cf), 0x3d7);
     TEST_EQUAL(Unicode::toupper(0x3d7), 0x3cf);
 
+    // U+20BA was added in Unicode 6.2.0:
+    TEST_EQUAL(Unicode::toupper(0x20ba), 0x20ba);
+    TEST_EQUAL(Unicode::tolower(0x20ba), 0x20ba);
+
+    // U+061C was added in Unicode 6.3.0:
+    TEST_EQUAL(Unicode::toupper(0x61c), 0x61c);
+    TEST_EQUAL(Unicode::tolower(0x61c), 0x61c);
+
     unsigned u;
     for (u = 0x514; u < 0x524; u += 2) {
 	TEST_EQUAL(Unicode::get_category(u), Unicode::UPPERCASE_LETTER);
@@ -293,7 +381,25 @@ DEFINE_TESTCASE(caseconvert2,!backend) {
 	TEST_EQUAL(Unicode::tolower(u), u + 1);
 	TEST_EQUAL(Unicode::toupper(u + 1), u);
     }
-	
+
+    // U+A7B1 was added in Unicode 8.0.0 as an uppercase form of U+0287.
+    TEST_EQUAL(Unicode::tolower(0xA7B1), 0x0287);
+    TEST_EQUAL(Unicode::toupper(0xA7B1), 0xA7B1);
+    TEST_EQUAL(Unicode::tolower(0x0287), 0x0287);
+    TEST_EQUAL(Unicode::toupper(0x0287), 0xA7B1);
+
+    // U+A7B4 (capital) and U+A7B5 (small) added in Unicode 8.0.0
+    TEST_EQUAL(Unicode::tolower(0xA7B4), 0xA7B5);
+    TEST_EQUAL(Unicode::toupper(0xA7B4), 0xA7B4);
+    TEST_EQUAL(Unicode::tolower(0xA7B5), 0xA7B5);
+    TEST_EQUAL(Unicode::toupper(0xA7B5), 0xA7B4);
+
+    // U+A7AE was added in Unicode 9.0.0 as an uppercase form of U+026A.
+    TEST_EQUAL(Unicode::tolower(0xA7AE), 0x026A);
+    TEST_EQUAL(Unicode::toupper(0xA7AE), 0xA7AE);
+    TEST_EQUAL(Unicode::tolower(0x026A), 0x026A);
+    TEST_EQUAL(Unicode::toupper(0x026A), 0xA7AE);
+
     return true;
 }
 
@@ -338,17 +444,24 @@ DEFINE_TESTCASE(unicodepredicates1,!backend) {
 	0x370, 0x372, 0x376, 0x3cf, 0x515, 0x517, 0x519, 0x51b, 0x51d, 0x51f,
 	0x521, 0x523, 0x2c6d, 0x2c6e, 0x2c6f,
 	// OTHER_LETTER
+	0x8bb, // Added in Unicode 9.0.0
+	0xc80, // Added in Unicode 9.0.0
 	0x10345,
 	// MODIFIER_LETTER (added in Unicode 5.1.0)
 	0x2ec, 0x374,
 	// NON_SPACING_MARK (added to is_wordchar() in 1.1.0)
 	0x651,
 	0x487, // Added in Unicode 5.1.0
+	0x8db, // Added in Unicode 9.0.0
 	0
     };
     const unsigned currency[] = {
 	// CURRENCY_SYMBOL
 	'$', 0xa3,
+	// CURRENCY_SYMBOL (added in Unicode 6.2.0)
+	0x20ba,
+	// CURRENCY_SYMBOL (added in Unicode 8.0.0)
+	0x20be,
 	0
     };
     const unsigned whitespace[] = {
@@ -361,8 +474,12 @@ DEFINE_TESTCASE(unicodepredicates1,!backend) {
     const unsigned other[] = {
 	// DASH_PUNCTUATION (added in Unicode 5.1.0)
 	0x5be,
-	// OTHER_SYMBOL (added in Unicode 5.1.0)
-	0x1f093,
+	// OTHER_SYMBOL
+	0xd4f, // Added in Unicode 9.0.0
+	0x1f093, // Added in Unicode 5.1.0
+	// FORMAT
+	0x61c, // Added in Unicode 6.3.0
+	0x8e2, // Added in Unicode 9.0.0
 	// UNASSIGNED
 	0xffff, 0x10ffff, 0x110000, 0xFFFFFFFF,
 	// PRIVATE_USE
