@@ -58,7 +58,7 @@ CPUTimer::get_current_cputime() const
 #ifdef HAVE_GETRUSAGE
     struct rusage r;
     if (getrusage(RUSAGE_SELF, &r) == -1) {
-	FAIL_TEST(string("Couldn't measure CPU for self: ") + strerror(errno));
+	FAIL_TEST("Couldn't measure CPU for self: " << strerror(errno));
     }
 
     t = r.ru_utime.tv_sec + r.ru_stime.tv_sec;
@@ -66,7 +66,7 @@ CPUTimer::get_current_cputime() const
 #elif defined HAVE_TIMES
     struct tms b;
     if (times(&b) == (clock_t)-1) {
-	FAIL_TEST(string("Couldn't measure CPU: ") + strerror(errno));
+	FAIL_TEST("Couldn't measure CPU: " << strerror(errno));
     }
     t = (double)(b.tms_utime + b.tms_stime);
 # ifdef HAVE_SYSCONF
